
/* $Id: munge_material.c,v 3.2 2000/06/12 21:38:34 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/*
			Copyright (c) 1989 by
			Advanced Visual Systems Inc.
			All Rights Reserved
	
	This software comprises unpublished confidential information of
	Advanced Visual Systems Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.
	
	This file is under sccs control at AVS in:
	/nfs/caffeine/root/src1/sccs/avs/examples/s.threshold.c
	
*/

#include <stdio.h>
#include <avs/avs.h>
#include <avs/field.h>

/*********************************************************************/

/* 
 * This is a C example to compute the threshold of a 3D scalar field of
 * floating point numbers. 
 *
 * The threshold function examines each element of a field to see
 * whether it falls within the range specified by a minimum and maximum 
 * parameter (controlled by dials).  Elements in the range are passed 
 * unchanged to the output field, elements outside the range are 
 * set to zero in the output field.
 *
 * There are two different versions in this file.
 *	1. Normal method of separate memory for input and output.
 *	2. Modifying data in place to save memory.  (new AVS 5 feature)
 *
 * The function AVSinit_modules is called from the main() routine
 * supplied by AVS.  In it, we call AVSmodule_from_desc with the name
 * of our description routine.
 */

#define INPUT_FIELD_DEFN    "field 3D scalar float rectilinear"

#define SPIN_FIELD_DEFN    "field 3D 3-vector float"

#define RADIUS_FIELD_DEFN    "field 3D scalar float"


static int threshold();

static int thresh_compute();


static int ((*mod_list[])()) = { threshold };
#define NMODS (sizeof(mod_list) / sizeof(char *))

void
AVSinit_modules()
{
	AVSinit_from_module_list (mod_list, NMODS);
}


/*  The routine "threshold" is the description routine.  */

static int threshold()
{
	/* Set the module name and type */
	AVSset_module_name("Munge Material", MODULE_FILTER);

	/* Create an input port for the required field input */
	AVScreate_input_port("Input Field", INPUT_FIELD_DEFN, REQUIRED);

	/* Create an output port for the spin vector */
	AVScreate_output_port("Spin Vector", SPIN_FIELD_DEFN );

	/* Create an output port for the radius */
	AVScreate_output_port("Atom Radius", RADIUS_FIELD_DEFN );

	/* Tell avs what subroutine to call to do the compute */
	AVSset_compute_proc(thresh_compute);

	return( 1 );
}


/*
 * thresh_compute is the compute routine.  It is called whenever AVS
 * wants to compute new threshold results.  The arguments are: the
 * value of the input field, the new output field (doubly indirected),
 * the minimum parameter value and the maximum parameter value.  Note
 * the order is always inputs, outputs, parameters.  The min comes
 * before the max because in the description routine above, the min
 * is declared before the max.
 */

static int thresh_compute(input, spin_output, radius_output)
AVSfield_float *input, **spin_output, **radius_output;
{
	register int i, j;

	float *radius_xcoords;
	float *radius_ycoords;
	float *radius_zcoords;

	float *spin_xcoords;
	float *spin_ycoords;
	float *spin_zcoords;

	int dims[3];

	/*
	 * We use a triply nested loop to traverse the field.  The macros
	 * MAXX, MAXY, and MAXZ determine the maximum extent of the field
	 * in each of the three dimensions.  We know this will be a
	 * 3-dimensional field because of the declaration in the description
	 * routine, so we don't need to check.  When we want to reference
	 * an element of the field we use the I3D macro which picks an
	 * element of a 3D field.  Note that the first index (i) varies the
	 * fastest in memory, so we make that the innermost loop.
	 */

	/* Allocate the Output Field Frames */

	dims[0] = MAXY(input);

	dims[1] = dims[2] = 1;

	if ( *spin_output == NULL )
	{
		printf( "Allocating Spin Vector Output Field\n" );

		*spin_output = (AVSfield_float *) AVSdata_alloc(
			SPIN_FIELD_DEFN, dims );
	}

	if ( *radius_output == NULL )
	{
		printf( "Allocating Radius Output Field\n" );

		*radius_output = (AVSfield_float *) AVSdata_alloc(
			RADIUS_FIELD_DEFN, dims );
	}

	/* Get Coordinate Pointers */

	spin_xcoords = RECT_X( *spin_output );
	spin_ycoords = RECT_Y( *spin_output );
	spin_zcoords = RECT_Z( *spin_output );

	radius_xcoords = RECT_X( *radius_output );
	radius_ycoords = RECT_Y( *radius_output );
	radius_zcoords = RECT_Z( *radius_output );

	/* Fill In Output Fields */

	for ( j=0 ; j < MAXY(input) ; j++ )
	{
		for ( i=0 ; i < MAXX(input) ; i++ )
		{
			switch ( i )
			{
				/* Get Coords */

				case 0:
				{
					spin_xcoords[j] = radius_xcoords[j] =
						I3D( input, i, j, 0 );

					break;
				}

				case 1:
				{
					spin_ycoords[j] = radius_ycoords[j] =
						I3D( input, i, j, 0 );

					break;
				}

				case 2:
				{
					spin_zcoords[j] = radius_zcoords[j] =
						I3D( input, i, j, 0 );

					break;
				}

				/* Get Spin Vector */

				case 3:
				case 4:
				case 5:
				{
					I3DV( *spin_output, j, 0, 0 )[ i - 3 ] =
						I3D( input, i, j, 0 );

					break;
				}

				/* Radius */

				case 6:
				{
					I3D( *radius_output, j, 0, 0 ) =
						I3D( input, i, j, 0 );

					break;
				}
			}
		}
	}

	/* When we're done, we return 1 to indicate success */
	return(1);
}

