
/* $Id: cpd_ulib.c,v 3.50 2000/08/03 18:20:57 kohl Exp $ */

/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* File: cpd_ulib.c - Higher level user functions for Checkpointing. */


/* System Header Files */

#include <stdio.h>


/* STV Header Files */

#include "stv.h"
#include "stvmac.h"
#include "cpd_proto.h"
#include "stv_glob.h"


/* Private Checkpointing Globals */

static int CPTIMESTAMP = 0;
static int CPLOADCOUNT = -1;
static int ISRESTART = stvFalse;
static int CPDOKTOSEND = stvTrue;
static int DORESTART = stvFalse;
static int CPRECEIVED = 0;
static int CPRELEASED = 0;
static int CPFINISHED = 0;

static int CPD_BEEN_THERE_DONE_THAT = 0;


/* Public Checkpointing Globals */

int CPD_CONFIG_TID = -1;


/*  -------------  stv_cpInit() -------------- */
/* This initializes the checkpoint information
*   o  stv_isCpRestart() is called to get restart flag.
*   o  return value of new NPROCS value in user ptr.
*   o  set up cpd client notify.
*   o  set executable file name for restarts.
*/

int
stv_cpInit( aout_name, notifytag, ntasks )
char *aout_name;
int notifytag;
int *ntasks;
{
	int cp_restart;
	int systag;
	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_cpInit",
			stvSystemInitd | stvSystemActive )) != stvStatusOk )
		return( cc );

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_ULIB | STV_DEBUG_CKPT ) )
	{
		printf( "stv_cpInit():  aout_name = %s,  notifytag = %d\n",
				( aout_name != NULL ) ? aout_name : "(null)",
				notifytag );
	}

	/* Better Make Sure We Have the System Contexts First... :-) */

	while ( (cc = stv_verify_system_contexts( stvFalse, stvFalse ))
			< 0 )
	{
		printf( "Waiting to Obtain CUMULVS System Contexts...\n" );

		stvSleep( 1 );
	}

	if ( cc > 0 )
		stv_define_message_handlers();

	/* Create Checkpointing Message Handlers */

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoRestart, stvRestart,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoCheckpointDataUpdate, stvCheckpointDataUpdate,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoCheckpointRelease, stvCheckpointRelease,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoAppRelease, stvAppRelease,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoCpdHalt, stvCpdHalt,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoConfigCpd, stvConfigCpd,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoCpdNproc, stvCpdNproc,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoCpdXon, stvCpdXon,
			stvAny, STV_USERTAG, STV_APP_CTX );

	/* Checkpointing User Library Notify Handlers */

	stv_add_handler( &STVNOTIFYHANDLERS, &nSTVNOTIFYHANDLERS,
			stv_protoCpdUlibNotify, stvCpdHalt,
			stvAny, STV_USERTAG, STV_APP_CTX );

	/* Add Message Pass-Thrus */

	stv_add_msg_pass_thru( stvRestart );

	stv_add_msg_pass_thru( stvCpdHalt );

	/* Check for Restart */

	cp_restart = stv_isCpRestart();

	if ( cp_restart )
	{
		if ( ntasks != NULL )
			*ntasks = cp_restart;
	}

	/* Set Up Client Notify */

	stv_client_notify( notifytag, stvMsgOp( stvRestart ) );

	/* Set Executable File Name */

	if ( aout_name != NULL && strcmp( aout_name, "" ) )
		stv_aout_restart( aout_name );
	
	else
	{
		STV_ERROR( "stv_cpInit", stvStatusBadParam,
			"Missing or Invalid Executable Name - No Restarts" );
	}

	/* Return Restart Flag */

	return( cp_restart );
}


/* ----------- stv_checkpoint ---------- */
/* send data to the local checkpointing daemon. Start a local
daemon if it is not found.  Return stvRestart, if a restart is
needed from a checkpoint
*/

int
stv_checkpoint()
{
	STV_BUFPTR msg;

	int cc, info;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_checkpoint",
			stvSystemInitd | stvSystemActive )) != stvStatusOk )
		return( cc );

	/* Debug Entry Message */

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_ULIB | STV_DEBUG_CKPT ) )
	{
		printf(
			"%s: [%d]  nfields=%d nparams=%d cpdtid=%x OK=%d DO=%d\n",
			"stv_checkpoint()", CPTIMESTAMP, nCPFIELDS, nCPPARAMS,
			CPDTID, CPDOKTOSEND, DORESTART );
	}

	/* find my local checkpoint daemon -- start it if needed */

	if ( (CPDTID = stv_find_cpd()) < 0 )
	{
		STV_ERROR( "stv_checkpoint", stvStatusSysFail,
				"Couldn't Find or (Re)Start Local Cpd." );

		return( stvStatusSysFail );
	}

	/* we're ready to checkpoint all the fields and parameters that
		have been flagged for checkpointing */

	while ( CPDOKTOSEND == stvFalse
			&& DORESTART == stvFalse
			&& CPDTID != -1 )
	{
		if ( (cc = stv_handleMessages( stvAny,
				stvAny, STV_USERTAG, STV_APP_CTX, stvMsgBlock )) <= 0 )
		{
			STV_ERROR( "stv_checkpoint", stvStatusSysFail,
					"Bad Return Status from stv_handleMessages()." );

			return( stvStatusSysFail );
		}
	}

	/* if a restart has been requested, then we don't do any */
	/* checkpointing, just flag that a restart has been requested */
	/* and return */

	if ( DORESTART == stvFalse && CPDTID >= 0 && CPDOKTOSEND )
	{
		if ( STV_OPT_VERBOSE & ( STV_DEBUG_ULIB | STV_DEBUG_CKPT ) )
			printf( "Sending Checkpoint Data.\n" );

		/* increment the checkpoint before sending */

		CPTIMESTAMP++;

		stv_pack_and_send_checkpoint( CPDTID, STV_NTASKS,
				STV_MYNODEID, STV_TIMESTAMP, CPTIMESTAMP,
				STVPARAMS.next, nCPPARAMS,
				STVFIELDS.next, nCPFIELDS );

		CPDOKTOSEND = stvFalse;
	}

	if ( DORESTART )
		return( stvRestart );

	else
		return( stvStatusOk );
}


/* This routine will load checkpointed data into the user's
 *	data structures.  stv_init() should have been called with
 *  STV_MYNODEID set. If it hasn't been inited, an error is returned.
 *
 *  This routine can be called several times and data incrementally
 *	loaded into the user's memory. That is stv_defineField and
 *	stv_defineParams may be interspersed with the stv_loadfromcp.
 *	The first stv_loadfromcp will obtain checkpoint information from
 *	the the cpd server. Subsequent loadfromcp's will use this
 *	information unless stv_restartFromCp has been called.
 *
 *	returns the number of data items left to load.
 *  -1 on error.
 */

int
stv_loadFromCP()
{
	int systag;
	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_loadFromCP",
			stvSystemInitd | stvSystemActive )) != stvStatusOk )
		return( cc );

	/* Special Check for Group Membership */

	if ( STV_MYNODEID == -1 )
	{
		STV_ERROR( "stv_loadFromCP", stvStatusBadParam,
				"STV Not Initialized - Call stv_init()." );

		return( stvStatusBadParam );
	}

	/* Debug Entry Message */

	if ( STV_OPT_VERBOSE
			& ( STV_DEBUG_ULIB | STV_DEBUG_CKPT | STV_DEBUG_RESTART ) )
	{
		printf( "stv_loadFromCP -- application %s -- node %d\n",
				STVUSERAPPNAME, STV_MYNODEID );
	}

	/* find my local checkpoint daemon -- start it if needed */

	if ( (CPDTID = stv_find_cpd()) < 0 )
	{
		STV_ERROR( "stv_loadFromCP", stvStatusSysFail,
				"Couldn't Find or (Re)Start Local Cpd." );

		return( stvStatusSysFail );
	}

	/* Check the load count. -1 means that we need to get the */
	/* checkpoint from the daemon */

	if ( CPLOADCOUNT == -1 )
	{
		/* Whoa...  We In Deep Now - Better Tell Viewers to GO AWAY! */
		STV_CP_RECOVERY_HOLD = stvTrue;

		/* The user may have called stv_loadFromCP() without getting
			the stvRestart message */

		if ( DORESTART == stvFalse && ISRESTART == stvFalse )
		{
			while ( DORESTART == stvFalse && CPDTID != -1 )
			{
				if ( STV_OPT_VERBOSE
						& ( STV_DEBUG_ULIB | STV_DEBUG_RESTART ) )
				{
					printf( "%s:  waiting for stvRestart message...\n",
						"stv_loadFromCP()" );
				}

				if ( (cc = stv_handleMessages( stvAny,
						stvAny, STV_USERTAG, STV_APP_CTX, stvMsgBlock ))
						<= 0 )
				{
					sprintf( STVERRTXT,
							"Bad Return Status from %s.",
							"stv_handleMessages()" );
					STV_ERROR( "stv_loadFromCP", stvStatusSysFail,
							STVERRTXT );

					return( stvStatusSysFail );
				}
			}
		}

		/* wait for cpd to release us before sending next checkpoint */

		CPDOKTOSEND = stvFalse;

		/* wait until the new guys are spawned before
			forwarding init0 messages */

		STV_FULLGROUP = 0;

		/* Get Checkpoint */

		if ( (cc = stv_get_checkpoint()) < 0 )
		{
			STV_ERROR( "stv_loadFromCP", cc,
					"Couldn't get checkpoint data." );

			return( cc );
		}

		ISRESTART = stvTrue;
	}

	CPLOADCOUNT = stv_load_checkpoint();

	if ( CPLOADCOUNT == 0 )
	{
		/* We've loaded all the checkpointed data */

		ISRESTART = stvFalse;

		CPLOADCOUNT = -1;

		/* Wait for Checkpoint Recovery Release */

		CPRELEASED = 0;

		while ( !CPRELEASED )
		{
			printf( "Waiting for Checkpoint Release...\n" );

			if ( (cc = stv_handleMessages( stvAny,
					stvAny, STV_USERTAG, STV_APP_CTX, stvMsgBlock ))
					<= 0 )
			{
				sprintf( STVERRTXT, "Bad Return Status from %s.",
						"stv_handleMessages()" );
				STV_ERROR( "stv_loadFromCP", stvStatusSysFail,
						STVERRTXT );

				return( stvStatusSysFail ); 	
			}

			/* check for re-restart...  D-Oh! */

			if ( DORESTART )
			{
				/* send extra request, to let Cpd know you're ready */
				cpd_requestCheckpoint();

				if ( STV_OPT_VERBOSE
						& ( STV_DEBUG_ULIB | STV_DEBUG_RESTART ) )
					printf( "%s: DORESTART -> 0\n",
							"stv_loadFromCP()" );

				DORESTART = stvFalse;
			}
		}

		/* O.K., We MUST Be Done By Now, Let The Viewers Rip! */
		STV_CP_RECOVERY_HOLD = stvFalse;

		return( 0 );
	}

	return( CPLOADCOUNT );
}


/* ----------- stv_isCpRestart ---------- */
/* return number of tasks if doing a restart.
*/

int
stv_isCpRestart()
{
	int parent;
	int systag;
	int tmp;
	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_isCpRestart",
			stvSystemInitd | stvSystemActive )) != stvStatusOk )
		return( cc );

	/* find my local checkpoint daemon -- start it if needed */

	if ( CPDTID < 0 )
	{
		if ( (CPDTID = stv_find_cpd()) < 0 )
		{
			STV_ERROR( "stv_isCpRestart", stvStatusSysFail,
					"Couldn't Find or (Re)Start Local Cpd." );

			return( stvStatusSysFail );
		}

		if ( stv_parent( &parent ) && parent == CPDTID )
		{
			/* Get Ntasks */

			if ( stv_handleMessages( stvCpdNproc,
					CPDTID, STV_USERTAG, STV_APP_CTX, stvMsgBlock )
					<= 0 )
			{
				STV_ERROR( "stv_isCpRestart", stvStatusSysFail,
						"Cpd Failed -- Bailing..." );

				exit( -1 );
			}

			DORESTART = stvTrue;  /* so that loadFromCP() will work */

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_ULIB | STV_DEBUG_RESTART ) )
			{
				printf( "%s:  STV_NTASKS=%d, setting DORESTART -> %d\n",
					"stv_isCpRestart()", STV_NTASKS, DORESTART );
			}

			return( STV_NTASKS );
		}
	}

	/* Handle Other Restart Cases Here ?! */
	/* XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX */

	return( 0 );
}


/* ----------- stv_restartFromCp ---------- */
/* Indicate that the task wants to restart from the most-recent
	Checkpoint
*/

int
stv_restartFromCp()
{
	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_restartFromCp",
			stvSystemInitd | stvSystemActive )) != stvStatusOk )
		return( cc );

	/* Special Check for Group Membership */

	if ( STV_MYNODEID == -1 )
	{
		STV_ERROR( "stv_restartFromCp", stvStatusBadParam,
				"STV Not Initialized - Call stv_init()." );

		return( stvStatusBadParam );
	}

	ISRESTART = stvTrue;

	CPLOADCOUNT = -1;

	return( ISRESTART );
}


/* ----------- stv_find_cpd ---------- */
/* find my local cpd -- tries to start one if one isn't running.
* Checks in with the local cpd daemon and gets a restart flag
*/

int
stv_find_cpd()
{
	STV_BUFPTR msg;

	char *spargv[4];

	char strtag[10];
	char flags[25];

	int *tids = 0;

	int spflag;
	int myhost;
	int ntids;
	int info;
	int i;

	myhost = stv_id_to_host( STV_MYID );

	/* don't have the CPDTID, first check the stvCpdWorkGroup for one */

	if ( CPDTID < 0
		&& stv_get_group_tids( stvCpdWorkGroup, tids, &ntids ) >= 0 )
	{
		for( i=0 ; i < ntids ; i++ )
		{
			if ( myhost == stv_id_to_host( tids[i] ) )
			{
				CPDTID = tids[i];

				break;
			}
		}

		if ( CPDTID >= 0 )
		{
			if ( stv_notify_event( stvTaskExit, STV_USERTAG,
					STV_APP_CTX, 1, &CPDTID ) == stvFalse )
			{
				STV_ERROR( "stv_find_cpd", CPDTID,
						"CPD Notify Request Failed!" );

				return( CPDTID );
			}
		}
	}

	/* spawn a cpd server on this host */

	if ( CPDTID < 0 )
	{
		spargv[0] = flags;
		spargv[1] = strtag;
		spargv[2] = ( char * ) NULL;

		strcpy( spargv[0], "-reply_user" ); /* reply flag */

		sprintf( spargv[1], "%d", STV_USERTAG );

		spflag = stvTaskHost + stvMppFront;

		if ( STV_OPT_VERBOSE & STV_DEBUG_SPAWN )
			spflag += stvTaskDebug;

		if ( !stv_spawn( "cpd", spargv, spflag,
				".", 1, &CPDTID, (int *) NULL ) )
		{
			sprintf( STVERRTXT, "%s: Couldn't Spawn Cpd Program!",
					"stv_find_cpd" );
			stv_mp_err( STVERRTXT );

			CPDTID = -1;

			return( stvStatusSysFail );
		}

		/* setup the notify */

		if ( stv_notify_event( stvTaskExit, STV_USERTAG, STV_APP_CTX,
				1, &CPDTID ) == stvFalse )
		{
			sprintf( STVERRTXT, "%s: CPD Notify Request Failed!",
					"stv_find_cpd" );
			stv_mp_err( STVERRTXT );

			return( CPDTID );
		}

		/* wait for this tid to send us the real cpd tid */

		if ( stv_handleMessages( stvConfigCpd,
				CPDTID, STV_USERTAG, STV_APP_CTX, stvMsgBlock ) <= 0 )
		{
			sprintf( STVERRTXT, "%s: CPD Config Reply Failed!",
					"stv_find_cpd" );
			stv_mp_err( STVERRTXT );

			return( stvStatusSysFail );
		}

		else
		{
			CPDTID = CPD_CONFIG_TID;

			/* Reset It */
			CPD_CONFIG_TID = -1;
		}
	}

	return( CPDTID );
}


/* ---------------- stv_protoCpdUlibNotify --------------- */
/* Handle a Notify for the Checkpointing User Library
*/

int
stv_protoCpdUlibNotify( msg )
STV_BUFPTR msg;
{
	int cc;

	if ( STV_OPT_VERBOSE
			& ( STV_DEBUG_ULIB | STV_DEBUG_NOTIFY
				| STV_DEBUG_RESTART ) )
	{
		printf( "Entry stv_protoCpdUlibNotify().\n" );
	}

	if ( msg == (STV_BUFPTR) NULL )
		return( stvStatusOk );

	if ( msg->src != CPDTID )
		return( stvStatusOk );

	cc = stv_vector_handler( STVHANDLERS, nSTVHANDLERS, msg );

	return( cc );
}


/* ---------------- stv_protoCpdHalt --------------- */
/* handle a CPD failing
*/

int
stv_protoCpdHalt( msg )
STV_BUFPTR msg;
{
	int vtid;
	int myid;

	if ( msg == (STV_BUFPTR) NULL )
		return( stvStatusOk );

	if ( msg->op != stvCpdHalt )
		return( stvStatusOk );

	if ( msg->src == CPDTID )
			CPDTID = -1;

	/* leave a suicide note */

	stv_myid( &myid );

	sprintf( STVERRTXT,
			"[%x] Our Cpd Died -- We're Outta Here\n", myid );
	STV_ERROR( "stv_protoCpdHalt", stvStatusSysFail, STVERRTXT );

	stv_exit();

	exit( -1 );

	return( stvStatusOk );
}


void
stv_pack_and_send_checkpoint( cpdtid, ntasks, nodeid, timestamp,
		cptimestamp, params, nparams, fields, nfields )
int cpdtid;
int ntasks;
int nodeid;
int timestamp;
int cptimestamp;
STV_PARAMPTR params;
int nparams;
STV_FIELDPTR fields;
int nfields;
{
	CPD_HOLE Hout;

	int bufid;
	int cc;

	if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT )
	{
		printf( "stv_pack_and_send -- application %s -- node %d \n",
				STVUSERAPPNAME, nodeid );
	}

	stv_init_msg_buf( stvDataDefault, &bufid );

	cc = stvMsgOp( stvCheckpointDataSubmit );
	stv_pack_int( bufid, &cc, 1, 1 );

	/* dummy fgid, for message header consistency */
	cc = -1;
	stv_pack_int( bufid, &cc, 1, 1 );

	stv_pack_str( bufid, STVUSERAPPNAME );
	stv_pack_int( bufid, &STV_USERTAG, 1, 1 );
	stv_pack_int( bufid, &nodeid, 1, 1 );
	stv_pack_int( bufid, &cptimestamp, 1, 1 );
	stv_pack_int( bufid, &timestamp, 1, 1 );
	stv_pack_int( bufid, &STV_NRESTARTS, 1, 1 );

	stv_pack_app_info( bufid, ntasks, STVPARAMS.next, nSTVPARAMS,
		STVFIELDS.next, nSTVFIELDS, stvCpOnly );

	Hout.type = stvCpdHoleTypePack;
	Hout.parm = bufid;

	cpd_pack_checkpoint_data( &Hout, 1, STVPARAMS.next, nSTVPARAMS,
			STVPARTICLES.next, nSTVPARTICLES,
			STVFIELDS.next, nSTVFIELDS );

	stv_send_msg( bufid, cpdtid, stvCpdTag, STV_CPD_CTX );

	stv_free_msg_buf( bufid );
}


/* -------- stv_load_checkpoint  -------- */
/* load data from a downloaded checkpoint into the users memory
* return the number of data items left to load.
*/

int
stv_load_checkpoint()
{
	int count;
	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_load_checkpoint",
			stvSystemInitd | stvSystemActive )) != stvStatusOk )
		return( cc );

	/* Debug Entry Message */

	if ( STV_OPT_VERBOSE
			& ( STV_DEBUG_ULIB | STV_DEBUG_CKPT | STV_DEBUG_RESTART ) )
		printf( "stv_load_checkpoint() called\n" );

	count = cpd_download_data( &CPDATA, STVPARAMS.next, nSTVPARAMS,
			STVPARTICLES.next, nSTVPARTICLES,
			STVFIELDS.next, nSTVFIELDS );

	return( count );
}


/* ------------- stv_get_checkpoint ----------- */
/* get the most recent checkpoint from the local daemon
*/

int
stv_get_checkpoint()
{
	STV_BUFPTR msg;

	int bufid;
	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_get_checkpoint",
			stvSystemInitd | stvSystemActive )) != stvStatusOk )
		return( cc );

	/* Debug Entry Message */

	if ( STV_OPT_VERBOSE
			& ( STV_DEBUG_ULIB | STV_DEBUG_CKPT | STV_DEBUG_RESTART ) )
		printf( "stv_get_checkpoint called\n" );

	if ( (CPDTID = stv_find_cpd()) < 0 )
	{
		STV_ERROR( "stv_get_checkpoint", stvStatusSysFail,
				"Couldn't Find or (Re)Start Local Cpd." );

		return( stvStatusSysFail );
	}

	/* request the latest checkpoint (until we actually get it :-) */

	do
	{
		/* send checkpoint request */

		cpd_requestCheckpoint();

		/* wait for the checkpoint reply */

		if ( STV_OPT_VERBOSE
				& ( STV_DEBUG_ULIB | STV_DEBUG_RESTART ) )
			printf( "%s: DORESTART -> 0\n", "stv_get_checkpoint()" );

		DORESTART = stvFalse;

		CPRECEIVED = 0;

		while ( !CPRECEIVED && DORESTART == stvFalse )
		{
			if ( (cc = stv_handleMessages( stvAny,
					stvAny, STV_USERTAG, STV_APP_CTX, stvMsgBlock ))
					<= 0 )
			{
				sprintf( STVERRTXT, "Bad Return Status from %s.",
						"stv_handleMessages()" );
				STV_ERROR( "stv_get_checkpoint", stvStatusSysFail,
						STVERRTXT );

				return( stvStatusSysFail ); 	
			}
		}

		if ( CPRECEIVED < 0 )
		{
			STV_ERROR( "stv_get_checkpoint", stvStatusSysFail,
					"Error Receiving Checkpoint Data." );

			return( stvStatusSysFail );
		}
	}
	while ( !CPRECEIVED );

	return( stvStatusOk );
}


/* ------------- cpd_requestCheckpoint ----------- */

void
cpd_requestCheckpoint()
{
	int bufid;
	int cc;

	if ( STV_OPT_VERBOSE
			& ( STV_DEBUG_ULIB | STV_DEBUG_CKPT
					| STV_DEBUG_RESTART ) )
	{
		printf( "%s: sending stvRequestCheckpoint to Cpd tid %x.\n",
				"cpd_requestCheckpoint()", CPDTID );
	}

	stv_init_msg_buf( stvDataDefault, &bufid );

	cc = stvMsgOp( stvRequestCheckpoint );
	stv_pack_int( bufid, &cc, 1, 1 );

	/* dummy fgid, for message header consistency */
	cc = -1;
	stv_pack_int( bufid, &cc, 1, 1 );

	stv_pack_str( bufid, STVUSERAPPNAME );
	stv_pack_int( bufid, &STV_USERTAG, 1, 1 );
	stv_pack_int( bufid, &STV_MYNODEID, 1, 1 );

	stv_pack_app_info( bufid, STV_NTASKS,
		STVPARAMS.next, nSTVPARAMS,
		STVFIELDS.next, nSTVFIELDS, stvCpOnly );

	stv_send_msg( bufid, CPDTID, stvCpdTag, STV_CPD_CTX );

	stv_free_msg_buf( bufid );
}


/* ------------- stv_protoRestart ----------- */

int
stv_protoRestart( msg )
STV_BUFPTR msg;
{
	STV_BUFPTR msg1;

	/* called in error */

	if ( msg->op != stvRestart )
		return( stvStatusOk );

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_ULIB | STV_DEBUG_RESTART ) )
	{
		printf(
		"%s:  CPDOKTOSEND:%d->0, CPLOADCOUNT:%d->-1, DORESTART:%d->1\n",
			"stv_protoRestart()", CPDOKTOSEND, CPLOADCOUNT,
			DORESTART );
	}

	CPDOKTOSEND = stvFalse;

	/* force stv_get_checkpoint to be called */

	CPLOADCOUNT = -1;

	DORESTART = stvTrue;

	return( stvStatusOk );
}


/* ----------- stv_protoCheckpointDataUpdate ----------- */

int
stv_protoCheckpointDataUpdate( msg )
STV_BUFPTR msg;
{
	STV_BUFPTR msg1;

	int cc;

	/* called in error */

	if ( msg->op != stvCheckpointDataUpdate )
		return( stvStatusOk );

	stv_unpack_int( msg->bufid, &CPTIMESTAMP, 1, 1 );

	if ( CPTIMESTAMP < 0 )
	{
		STV_ERROR( "stv_protoCheckpointDataUpdate", stvStatusBadParam,
				"Error Receiving Checkpoint." );

		CPRECEIVED = -1;

		return( stvStatusBadParam );
	}

	stv_unpack_int( msg->bufid, &STV_TIMESTAMP, 1, 1 );
	stv_unpack_int( msg->bufid, &STV_NTASKS, 1, 1 );
	stv_unpack_int( msg->bufid, &STV_MYNODEID, 1, 1 );
	stv_unpack_int( msg->bufid, &STV_NRESTARTS, 1, 1 );

	if ( (cc = cpd_receive_checkpoint_data( msg->bufid, &CPDATA ))
			!= stvStatusOk )
	{
		STV_ERROR( "stv_protoCheckpointDataUpdate", cc,
				"Error Receiving Checkpoint Data." );

		CPRECEIVED = -1;

		return( cc );
	}

	CPRECEIVED = 1;

	return( stvStatusOk );
}


/* ----------- stv_protoCheckpointRelease ----------- */

int
stv_protoCheckpointRelease( msg )
STV_BUFPTR msg;
{
	STV_BUFPTR msg1;

	int tschk;
	int cc;

	/* called in error */

	if ( msg->op != stvCheckpointRelease )
		return( stvStatusOk );

	stv_unpack_int( msg->bufid, &tschk, 1, 1 );

	if ( tschk != CPTIMESTAMP )
	{
		STV_ERROR( "stv_protoCheckpointRelease", stvStatusSysFail,
				"Checkpoint Timestamp Mismatch...  Bailing Out." );

		exit( -1 );
	}

	stv_unpack_int( msg->bufid, &STV_NRESTARTS, 1, 1 );

	CPRELEASED = 1;

	return( stvStatusOk );
}


/* ----------- stv_protoAppRelease ----------- */

int
stv_protoAppRelease( msg )
STV_BUFPTR msg;
{
	STV_BUFPTR msg1;

	int tschk;
	int cc;

	/* called in error */

	if ( msg->op != stvAppRelease )
		return( stvStatusOk );

	stv_unpack_int( msg->bufid, &tschk, 1, 1 );

	if ( tschk != CPTIMESTAMP )
	{
		STV_ERROR( "stv_protoAppRelease", stvStatusSysFail,
				"Checkpoint Timestamp Mismatch...  Bailing Out." );

		exit( -1 );
	}

	CPFINISHED = 1;

	return( stvStatusOk );
}


/* ------------- stv_protoConfigCpd ----------- */

int
stv_protoConfigCpd( msg )
STV_BUFPTR msg;
{
	/* called in error */

	if ( msg->op != stvConfigCpd )
		return( stvStatusOk );

	stv_unpack_int( msg->bufid, &CPD_CONFIG_TID, 1, 1 );

	return( stvStatusOk );
}


/* ------------- stv_protoCpdNproc ----------- */

int
stv_protoCpdNproc( msg )
STV_BUFPTR msg;
{
	/* called in error */

	if ( msg->op != stvCpdNproc )
		return( stvStatusOk );

	stv_unpack_int( msg->bufid, &STV_NTASKS, 1, 1 );

	return( stvStatusOk );
}


/* ------------- stv_protoCpdXon ----------- */

int
stv_protoCpdXon( msg )
STV_BUFPTR msg;
{
	int gen;

	/* called in error */

	if ( msg->op != stvCpdXon )
		return( stvStatusOk );

	stv_unpack_int( msg->bufid, &gen, 1, 1 );

	if ( gen >= stv_generation() )
		CPDOKTOSEND = stvTrue;	

	return( stvStatusOk );
}


/* ---- stv_restart_aout( name ) -----*/
/* tell the checkpoint daemon what executable to use for restart */

int
stv_aout_restart( name )
char *name;
{
	int bufid;
	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_aout_restart",
			stvSystemInitd | stvSystemActive )) != stvStatusOk )
		return( cc );

	/* Verify CPD */

	if ( (CPDTID = stv_find_cpd()) < 0 )
	{
		STV_ERROR( "stv_aout_restart", stvStatusSysFail,
				"Couldn't Find or (Re)Start Local Cpd." );

		return( stvStatusSysFail );
	}

	stv_init_msg_buf( stvDataDefault, &bufid );

	cc = stvMsgOp( stvCpdAout );
	stv_pack_int( bufid, &cc, 1, 1 );

	/* dummy fgid, for message header consistency */
	cc = -1;
	stv_pack_int( bufid, &cc, 1, 1 );

	stv_pack_str( bufid, STVUSERAPPNAME );
	stv_pack_int( bufid, &STV_USERTAG, 1, 1 );
	stv_pack_int( bufid, &STV_USERTAG, 1, 1 );
	stv_pack_str( bufid, name );

	stv_send_msg( bufid, CPDTID, stvCpdTag, STV_CPD_CTX );

	stv_free_msg_buf( bufid );

	return( stvStatusOk );
}


/* ------- stv_client_notify ------ */
/* tell the cpd to use the tag, value combination to indicate that
	the task should restart */

int
stv_client_notify( tag, value )
int tag, value;
{
	int bufid;
	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_client_notify",
			stvSystemInitd | stvSystemActive )) != stvStatusOk )
		return( cc );

	/* Verify CPD */

	if ( (CPDTID = stv_find_cpd()) < 0 )
	{
		STV_ERROR( "stv_client_notify", stvStatusSysFail,
				"Couldn't Find or (Re)Start Local Cpd." );

		return( stvStatusSysFail );
	}

	stv_init_msg_buf( stvDataDefault, &bufid );

	cc = stvMsgOp( stvCpdUserNotifyInfo );
	stv_pack_int( bufid, &cc, 1, 1 );

	/* dummy fgid, for message header consistency */
	cc = -1;
	stv_pack_int( bufid, &cc, 1, 1 );

	stv_pack_str( bufid, STVUSERAPPNAME );
	stv_pack_int( bufid, &STV_USERTAG, 1, 1 );
	stv_pack_int( bufid, &STV_NTASKS, 1, 1 );
	stv_pack_int( bufid, &tag, 1, 1 );
	stv_pack_int( bufid, &value, 1, 1 );
	stv_pack_int( bufid, &STV_OPT_CP_RECOVERY, 1, 1 );

	stv_send_msg( bufid, CPDTID, stvCpdTag, STV_CPD_CTX );

	stv_free_msg_buf( bufid );

	stv_notify_event( stvTaskExit, tag, STV_BASECTX, 1, &CPDTID );

	return( stvStatusOk );
}


/* ------- stv_cpFinished ------ */
/* tell the cpd that we're done... */

int
stv_cpFinished()
{
	int bufid;
	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_cpFinished",
			stvSystemInitd | stvSystemActive )) != stvStatusOk )
		return( cc );

	/* Verify CPD */

	if ( (CPDTID = stv_find_cpd()) < 0 )
	{
		STV_ERROR( "stv_cpFinished", stvStatusSysFail,
				"Couldn't Find or (Re)Start Local Cpd." );

		return( stvStatusSysFail );
	}

	stv_init_msg_buf( stvDataDefault, &bufid );

	cc = stvMsgOp( stvAppFinished );
	stv_pack_int( bufid, &cc, 1, 1 );

	/* dummy fgid, for message header consistency */
	cc = -1;
	stv_pack_int( bufid, &cc, 1, 1 );

	stv_pack_str( bufid, STVUSERAPPNAME );
	stv_pack_int( bufid, &STV_USERTAG, 1, 1 );
	stv_pack_int( bufid, &STV_NTASKS, 1, 1 );

	stv_send_msg( bufid, CPDTID, stvCpdTag, STV_CPD_CTX );

	stv_free_msg_buf( bufid );

	/* Wait for Release */

	CPFINISHED = 0;

	while ( !CPFINISHED )
	{
		printf( "Waiting for Finished Release...\n" );

		if ( (cc = stv_handleMessages( stvAny,
				stvAny, STV_USERTAG, STV_APP_CTX, stvMsgBlock )) <= 0 )
		{
			sprintf( STVERRTXT, "Bad Return Status from %s.",
					"stv_handleMessages()" );
			STV_ERROR( "stv_cpFinished", stvStatusSysFail,
					STVERRTXT );

			return( stvStatusSysFail ); 	
		}

		/* check for re-restart...  D-Oh! */

		if ( DORESTART )
		{
			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_ULIB | STV_DEBUG_CKPT
							| STV_DEBUG_RESTART ) )
				printf( "%s: DORESTART -> 1\n",
						"stv_cpFinished()" );

			return( stvRestart );
		}
	}

	return( stvStatusOk );
}


/* ------ stv_generation ------ */
/* return how many times this running app has restarted */

int
stv_generation()
{
	return( STV_NRESTARTS );
}


/* ------ stv_cpdtid ------ */
/* return the current cpdtid*/

int
stv_cpdtid()
{
	return( CPDTID );
}

