
/* $Id: stv_msg.c,v 3.108 2000/08/03 16:50:13 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* System Header Files */

#include <stdio.h>
#include <malloc.h>


/* STV Header Files */

#include "stv.h"
#include "stvmac.h"
#include "stv_glob.h"


/*** Main Viewer Initialization Routine ***/


int
stv_viewer_init( V, mytid, appname, block )
STV_VIEWER *V;
int mytid;
char *appname;
int block;
{
	static int first = stvTrue;

	STV_VIEWER Vptr;

	char *debug_env;

	int no_flush;
	int rbuf;
	int src;
	int tid;
	int tmp;
	int cc;

	/* Dump Banner */

	if ( first )
	{
		stv_banner();

		first = stvFalse;

		/* Check for Debug Flags in Env */

		if ( (debug_env = getenv( "STV_DEBUG" )) != NULL )
		{
			cc = atoi( debug_env );

			if ( cc )
			{
				printf( "Setting STV_DEBUG from environment: " );
				stv_dump_debug_flag( cc, stvTrue );

				stv_setopt( stvOptNull, stvOptVerbose, cc );
			}
		}

		/* Verify CUMULVS System Contexts */

		if ( stv_verify_system_contexts( stvTrue, stvTrue ) < 0 )
		{
			STV_ERROR( "stv_viewer_init", stvStatusSysFail,
				"Verifying CUMULVS System Contexts.  Bailing Out..." );

			return( stvStatusSysFail );
		}

		/* Add Message Handlers */

		stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
				stv_protoInit1, stvInit1,
				stvAny, stvAny, STV_VIEWER_CTX );
		
		stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
				stv_protoFieldTime, stvFieldTime,
				stvAny, stvAny, STV_VIEWER_CTX );
		
		stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
				stv_protoFieldAck, stvFieldAck,
				stvAny, stvAny, STV_VIEWER_CTX );
		
		stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
				stv_protoRegion, stvRegion,
				stvAny, stvAny, STV_VIEWER_CTX );
		
		/* Add Viewer Library Notify Handlers */

		stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
				stv_protoNotify, stvNotify,
				stvAny, stvAny, STV_VIEWER_CTX );

		stv_add_handler( &STVNOTIFYHANDLERS, &nSTVNOTIFYHANDLERS,
				stv_protoViewerNotify, stvNotify,
				stvAny, stvAny, STV_VIEWER_CTX );

		/* Add Message Pass-Thrus */

		stv_add_msg_pass_thru( stvNotify );
	}

	/* Create New Viewer */

	if ( V == NULL || *V == NULL )
	{
		Vptr = stv_create_viewer();

		STV_NO_MEM_return( Vptr, "stv_viewer_init" );

		/* Add to Master List */

		Vptr->next = STVVIEWERS;

		STVVIEWERS = Vptr;

		nSTVVIEWERS++;

		/* Initialize Counters */

		Vptr->nvfgs = 0;

		Vptr->nvfields = 0;
		Vptr->nvparams = 0;
		Vptr->nvtasks = 0;

		Vptr->nparticles = 0;
		Vptr->ndecomps = 0;
		Vptr->nfields = 0;
		Vptr->nparams = 0;

		/* Return New Viewer in Ptr (if exists) */

		if ( V != NULL )
			*V = Vptr;
	}

	/* Reset Existing Viewer */

	else
	{
		Vptr = *V;

		stv_viewer_reset( Vptr );
	}

	/* Grab Next Viewer ID */

	Vptr->id = STV_VIEWER_ID++;

	/* Save Viewer Task ID */

	Vptr->mytid = mytid;

	/* Increment Initialization Sequence */

	Vptr->init_seq = STV_INIT_ID++;

	no_flush = 0;

	do
	{
		/* Look for Application Group */
	
		cc = stv_viewer_lookup_app_group( appname, &tid, block );
	
		if ( !cc )
			return( stvStatusBadParam );
	
		if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
		{
			printf(
				"Group Found, Instance 0 is Task ID=[ 0x%x / %d ]\n",
				tid, tid );
			fflush( stdout );
		}
	
		/* Obtain Application Message Context */
	
		stv_viewer_lookup_app_context( appname, Vptr );
	
		if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
		{
			printf( "Message Tag is %d.\n", Vptr->msgtag );
			fflush( stdout );
		}
	
		/* Flush Any Queued Messages */
	
		while ( !no_flush
			&& stv_recv_msg_noblock( -1, -1, STV_VIEWER_CTX, &rbuf ) )
		{
			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER | STV_DEBUG_FLUSH ) )
			{
				stv_msg_buf_info( rbuf, (int *) NULL, &tmp, &src );
	
				printf(
					"Cleaned Out Message:  SRC=[ 0x%x / %d ] tag=%d\n",
					src, src, tmp );
				fflush( stdout );
			}
	
			stv_free_msg_buf( rbuf );
		}
	
		/* Set Up Notify for Main Task */
	
		cc = stv_notify_event( stvTaskExit, Vptr->msgtag,
				STV_VIEWER_CTX, 1, &tid );
	
		if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_NOTIFY ) )
		{
			printf( "Set Up Notify for Main Task, cc=%d\n", cc );
			fflush( stdout );
		}
	
		/* Send First Initialization Packet INIT0 */
	
		stv_viewer_send_Init0( Vptr, tid );
	
		Vptr->init_gen = -1;

		stv_free_view_task_list( &(Vptr->vtasks) );

		Vptr->nvtasks = 0;

		if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
		{
			printf( "INIT0 Message Sent.\n" );
			fflush( stdout );
		}
	
		/* Wait for INIT1 Replies */
	
		cc = stv_viewer_recv_Init1( Vptr );
	
		if ( cc == stvStatusOk )
		{
			if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
			{
				printf( "All INIT1 Replies Received\n" );
				fflush( stdout );
			}
		}
	
		else if ( cc != stvStatusBadParam )
		{
			/* Bail */
	
			STV_ERROR( "stv_viewer_init", cc, "Bailing Out..." );
		}

		no_flush++;
	}
	while ( cc == stvStatusBadParam );

	return( cc );
}


/*** Group Routines ***/


/* Viewer Lookup Application Group */

int
stv_viewer_lookup_app_group( appname, tid, block )
char *appname;
int *tid;
int block;
{
	char appgroup[1024];
	char tmp[1024];

	char *ptr;

	int cc;

	if ( appname == NULL || !strcmp( appname, "" ) )
		ptr = stvDefaultClient;
	
	else
		ptr = appname;

	sprintf( appgroup, "%s%s", stvPrefix, ptr );

	do
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
		{
			printf( "Looking up Group \"%s\".\n", appgroup );
			fflush( stdout );
		}

		cc = stv_get_id_from_instance( appgroup, 0, tid );

		if ( !cc || *tid < 0 )
		{
			sprintf( tmp, "Getting Instance 0 ID for %s", ptr );
				
			stv_mp_err( tmp );

			stvSleep( 1 );
		}
	}
	while ( ( !cc || *tid < 0 ) && block );

	if ( !cc || *tid < 0 )
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
		{
			printf( "Non-blocking, retry later...\n" );
			fflush( stdout );
		}

		return( stvFalse );
	}
	
	else
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
		{
			printf( "Found Tid=%x in Group \"%s\".\n", *tid, appgroup );
			fflush( stdout );
		}

		return( stvTrue );
	}
}


/* Lookup Application Message Tag / Context */

void
stv_viewer_lookup_app_context( appname, V )
char *appname;
STV_VIEWER V;
{
	char appgroup[1024];

	int index;
	int cc;

	if ( appname == NULL || !strcmp( appname, "" ) )
		sprintf( appgroup, "%s%s", stvPrefix, stvDefaultClient );

	else
		sprintf( appgroup, "%s%s", stvPrefix, appname );

	do
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
		{
			printf( "Looking up Application Context\n" );
			fflush( stdout );
		}

		cc = stv_name_lookup( appgroup, 0, &(V->msgtag), &index );

		if ( !cc || index < 0 )
		{
			stv_mp_err( "Looking Up Msg Tag" );

			stvSleep( 1 );
		}
	}
	while ( !cc || index < 0 );

	if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "Found Application Context tag=%d at index=%d\n",
				V->msgtag, index );
		fflush( stdout );
	}
}


/*** Viewer Field Request Routines ***/


STV_VIEW_FIELD_GROUP
stv_viewer_request_static_field( V, R, C, freq, no_synch, restart )
STV_VIEWER V;
STV_REGION R;
STV_CELL C;
int freq;
int no_synch;
int *restart;
{
	int frflag;

	frflag = stvFieldRqstDefault | stvFieldRqstStatic;

	if ( no_synch )
		frflag |= stvFieldRqstNoSynch;

	return( stv_viewer_submit_field_request( V, R, C, freq,
			frflag, restart ) );
}


STV_VIEW_FIELD_GROUP
stv_viewer_request_field( V, R, C, freq, no_synch, restart )
STV_VIEWER V;
STV_REGION R;
STV_CELL C;
int freq;
int no_synch;
int *restart;
{
	int frflag;

	frflag = stvFieldRqstDefault;

	if ( no_synch )
		frflag |= stvFieldRqstNoSynch;

	return( stv_viewer_submit_field_request( V, R, C, freq,
			frflag, restart ) );
}


STV_VIEW_FIELD_GROUP
stv_viewer_submit_field_request( V, R, C, freq, flag, restart )
STV_VIEWER V;
STV_REGION R;
STV_CELL C;
int freq;
int flag;
int *restart;
{
	STV_VIEW_FIELD_GROUP VFGptr;
	STV_VIEW_FIELD_GROUP VFG;

	int is_static;
	int no_synch;
	int cc;

	/* Create View Field Group */

	VFG = stv_make_view_field_group( V->vfields, V->nvfields );

	if ( VFG == NULL )
		return( (STV_VIEW_FIELD_GROUP) NULL );

	/* Add to View Field Group List */

	if ( V->vfglist == NULL )
	{
		V->vfglist = VFG;

		VFGptr = (STV_VIEW_FIELD_GROUP) NULL;
	}

	else
	{
		VFGptr = V->vfglist;

		while ( VFGptr->next != NULL )
			VFGptr = VFGptr->next;

		VFGptr->next = VFG;
	}

	(V->nvfgs)++;

	stv_copy_region( VFG->visregion, R );

	stv_copy_cell( VFG->viscell, C );

	VFG->freq = freq;

	VFG->eff_freq = stv_calculate_eff_freq( VFG->freq,
			VFG->view_field_list[0]->field->time_scale );

	VFG->flag = flag;

	no_synch = flag & stvFieldRqstNoSynch;

	is_static = flag & stvFieldRqstStatic;

	cc = stv_viewer_FieldRqst( VFG, V, no_synch, is_static, restart );

	/* Return Result */

	if ( cc != stvStatusOk )
	{
		/* Free It Up... Something Died */

		if ( VFGptr != NULL )
			stv_free_view_field_group( &(VFGptr->next) );
		
		else
			stv_free_view_field_group( &(V->vfglist) );

		return( (STV_VIEW_FIELD_GROUP) NULL );
	}

	else
		return( VFG );
}


int
stv_calculate_eff_freq( freq, time_scale )
int freq;
int time_scale;
{
	float ratio;

	int eff_freq;

	/* Requested Frequency is a Time Value */

	if ( stv_is_time_value( freq ) )
	{
		/* Field's Time Scale is Also a Time Value */

		if ( stv_is_time_value( time_scale ) )
		{
			/* Calculate Ratio of 2 Time Values */

			ratio = stv_time_ratio( freq, time_scale );

			/* Effective Frequency is Truncation of Ratio */

			eff_freq = (int) ratio;

			eff_freq = ( eff_freq < 1 ) ? 1 : eff_freq;
		}

		/* Field's Time Scale is Not Set - Default to Every Frame */

		else
			eff_freq = 1;
	}

	/* Requested Frequency is an Iteration Count - Use As Is */

	else
		eff_freq = freq;

	/* Return Effective Frequency */

	return( eff_freq );
}


/*** Field Request Protocol Routine ***/

int
stv_viewer_FieldRqst( VFG, V, no_synch, is_static, restart )
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
int no_synch;
int is_static;
int *restart;
{
	int cc;
	int i;

	/* Set Up Notifies for View Field Group Tasks */

	cc = stv_notify_event( stvTaskExit, V->msgtag, STV_VIEWER_CTX,
		VFG->ntids, VFG->tid_list );

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_NOTIFY ) )
	{
		printf( "Set Up Notify for View Field Group Tasks, cc=%d\n",
			cc );
		fflush( stdout );
	}

	/* Send Field Request to Application */

	if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "\nSending FieldRqst...\n" );

		printf( "\teff_freq=%d, freq = ", VFG->eff_freq );
		stv_dump_time_str( stdout, VFG->freq, stvFalse );
		printf( ", no_synch=%d, is_static=%d\n", no_synch, is_static );

		for ( i=0 ; i < VFG->ntids ; i++ )
		{
			printf( "\tTask ID=[ 0x%x / %d ]\n",
				VFG->tid_list[i], VFG->tid_list[i] );
		}

		for ( i=0 ; i < VFG->nvfields ; i++ )
		{
			printf( "\tfield=\"%s\"\n",
				(VFG->view_field_list[i])->field->name );
		}

		fflush( stdout );
	}

	stv_viewer_send_FieldRqst( VFG, V, no_synch, is_static );

	/* If Synchronizing, Handle Field Times, etc. */

	if ( !no_synch )
	{
		/* Receive Back Field Times */

		cc = stv_viewer_recv_FieldTime( VFG, V );

		if ( cc != stvStatusOk )
		{
			/* Bail */

			STV_ERROR( "stv_viewer_FieldRqst", cc, "Bailing Out...\n" );

			if ( cc == stvStatusBail )
			{
				STV_ERROR( "stv_viewer_FieldRqst", cc,
					"Sending FieldHaltAll" );

				stv_viewer_send_FieldHaltAll( VFG, V );
			}

			else if ( cc == stvStatusReconfigure )
				(*restart)++;

			return( cc );
		}

		if ( STV_OPT_VERBOSE
				& ( STV_DEBUG_VIEWER | STV_DEBUG_TIME
						| STV_DEBUG_FREQ ) )
		{
			printf( "All FieldTime Replies Received, Max Time is %d.\n",
				VFG->timemax );
			fflush( stdout );
		}

		/* Determine the Starting T */

		if ( VFG->locked == stvAppUnlocked )
		{
			VFG->time_offset = VFG->timemax + 2;

			/* Send FieldTstart to Tasks */

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER | STV_DEBUG_TIME
							| STV_DEBUG_FREQ ) )
			{
				printf( "\nSending FieldTstart, To = %d.\n",
					VFG->time_offset );
				fflush( stdout );
			}

			stv_viewer_send_FieldTstart( VFG, V, VFG->time_offset );
		}

		else
			VFG->time_offset = VFG->timemax;
	}

	else if ( is_static )
	{
		/* Receive Back Field Request Acks & Nacks */

		cc = stv_viewer_recv_FieldAcks( VFG, V );

		if ( cc != stvStatusOk )
		{
			/* Bail */

			STV_ERROR( "stv_viewer_FieldRqst", cc, "Bailing Out...\n" );

			if ( cc == stvStatusBail )
			{
				STV_ERROR( "stv_viewer_FieldRqst", cc,
					"Sending FieldHaltAll" );

				stv_viewer_send_FieldHaltAll( VFG, V );
			}

			else if ( cc == stvStatusReconfigure )
				(*restart)++;

			return( cc );
		}

		if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
		{
			printf(
				"All Field Request Ack Replies Recvd, %d Tasks Left.\n",
				VFG->ntids );
			fflush( stdout );
		}
	}

	/* Initialize View Field Group for Region Frames */

	if ( !no_synch )
		VFG->region_time = VFG->time_offset;

	else
		VFG->region_time = 0;

	if ( VFG->locked == stvAppUnlocked )
		VFG->region_time += VFG->eff_freq;

	VFG->region_cnt = 0;

	VFG->frame_ok = stvTrue;

	return( stvStatusOk );
}


/*** Viewer Frame Collection Routine ***/

int
stv_viewer_receive_frame( VFG, V, restart, block )
STV_VIEW_FIELD_GROUP *VFG;
STV_VIEWER V;
int *restart;
int block;
{
	STV_FIELDPTR F;

	int do_parts;
	int cc;
	int i;

	/* Receive Region Frame Pieces from Tasks */

	cc = stv_viewer_recv_Region( VFG, V, block );

	if ( cc == stvStatusIncomplete )
		return( cc );

	else if ( cc == stvStatusOk )
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
		{
			printf( "All Region Replies Received for fgid=%d.\n",
				(*VFG)->fgid );
			fflush( stdout );
		}

		if ( STV_OPT_DUMP || STV_OPT_DUMP_ALL )
		{
			do_parts = 0;

			for ( i=0 ; i < (*VFG)->nvfields ; i++ )
			{
				F = ((*VFG)->view_field_list[i])->field;

				if ( F->decomp != NULL )
				{
					printf(
						"\nGlobal Field Frame Data for \"%s\":\n\tat ",
						F->name );
					stv_dump_base_time_str( stdout,
						F->time_base, F->time_units, stvFalse );
					printf( "\n\n" );
					fflush( stdout );

					stv_dump_vfg_field_array( *VFG, i );
				}

				else if ( F->particle != NULL )
					do_parts++;
			}

			if ( do_parts )
			{
				printf( "\nGlobal Particle Field Frame Data:\n\tat " );
				stv_dump_base_time_str( stdout,
					F->time_base, F->time_units, stvFalse );
				printf( "\n\n" );
				fflush( stdout );

				stv_dump_particle_field( (*VFG)->particle_list );
			}
		}
	}

	else if ( cc == stvStatusBadFrame )
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
		{
			printf(
				"All Regions Received fgid=%d - Ignore Bad Frame.\n",
				(*VFG)->fgid );
			fflush( stdout );
		}
	}

	else
	{
		/* Bail */

		STV_ERROR( "stv_viewer_receive_frame", cc, "Bailing Out..." );

		if ( cc == stvStatusBail || cc == stvStatusBadParam )
		{
			STV_ERROR( "stv_viewer_receive_frame", cc,
				"Sending FieldHaltAll" );

			if ( *VFG != NULL )
				stv_viewer_send_FieldHaltAll( *VFG, V );
			
			else
				stv_viewer_send_FieldHaltAll( V->vfglist, V );
		}

		else if ( cc == stvStatusReconfigure )
			(*restart)++;
	}

	return( cc );
}


/* Receive Frame Pieces from Tasks */

int
stv_viewer_recv_Region( VFGptr, V, block )
STV_VIEW_FIELD_GROUP *VFGptr;
STV_VIEWER V;
int block;
{
	STV_VIEW_FIELD_GROUP VFG;

	int ret;
	int cc;

	/* Clear VFGptr in case we bail... */

	*VFGptr = (STV_VIEW_FIELD_GROUP) NULL;

	/* Check for Bailed View Field Group */

	VFG = V->vfglist;

	while ( VFG != NULL && !(VFG->bail) )
		VFG = VFG->next;

	/* Check for Already Finished View Field Group */

	if ( VFG == NULL )
	{
		VFG = V->vfglist;

		while ( VFG != NULL && VFG->region_cnt < VFG->ntids )
			VFG = VFG->next;
	}

	/* Handle Region Messages */

	if ( VFG == NULL )
	{
		do
		{
			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER | STV_DEBUG_MSG ) )
			{
				printf( "Blocking for Region Messages.\n" );
				fflush( stdout );
			}

			if ( (cc = stv_handleMessages( stvRegion,
					stvAny, V->msgtag, STV_VIEWER_CTX,
					block )) < 0 )
			{
				STV_ERROR( "stv_viewer_recv_Region", cc,
						"Receiving Region Message From Application." );
			
				return( cc );
			}

			if ( !cc && block == stvMsgBlock )
			{
				STV_ERROR( "stv_viewer_recv_Region", stvStatusBadParam,
						"Receiving Region Message From Application." );
			
				return( stvStatusBadParam );
			}

			/* Check for Bailed View Field Group */

			VFG = V->vfglist;

			while ( VFG != NULL && !(VFG->bail) )
				VFG = VFG->next;

			/* Check for Finished View Field Group */

			if ( VFG == NULL )
			{
				VFG = V->vfglist;

				while ( VFG != NULL && VFG->region_cnt < VFG->ntids )
					VFG = VFG->next;
			}
		}
		while ( block == stvMsgBlock && VFG == NULL );
	}

	/* Set Return Code */

	if ( block != stvMsgBlock && VFG == NULL )
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
		{
			printf( "Incomplete Frame\n" );
			fflush( stdout );
		}

		return( stvStatusIncomplete );
	}

	else if ( VFG->bail )
	{
		if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_NOTIFY ) )
		{
			printf( "View Field Group Lost a Task - Bail!\n" );
			fflush( stdout );
		}

		ret = stvStatusBail;
	}

	else if ( VFG->frame_ok )
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
		{
			printf( "Frame is O.K.\n" );
			fflush( stdout );
		}

		ret = stvStatusOk;
	}

	else
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
		{
			printf( "Bad Frame  :-(\n" );
			fflush( stdout );
		}

		ret = stvStatusBadFrame;
	}

	/* Reset View Field Group for Next Region */

	VFG->region_cnt = 0;

	VFG->frame_ok = stvTrue;

	/* Pass Back Pointer to View Field Group */

	*VFGptr = VFG;

	return( ret );
}


/* Receive and Process One Frame Region Message */

int
stv_protoRegion( msg )
STV_BUFPTR msg;
{
	STV_VIEW_FIELD_GROUP VFG;

	STV_FIELDPTR F;

	STV_VIEWER V;

	int waslocked;
	int tmp;
	int vid;
	int cc;
	int i;

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MSG ) )
	{
		printf( "Entry stv_protoRegion().\n" );
		fflush( stdout );
	}

	if ( msg == (STV_BUFPTR) NULL )
		return( stvStatusOk );

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MSG ) )
	{
		printf( "Message Received From Task ID=[ 0x%x / %d ]\n",
				msg->src, msg->src );
		fflush( stdout );
	}

	/* Unpack Viewer ID */

	stv_unpack_int( msg->bufid, &vid, 1, 1 );

	if ( (V = stv_find_viewer( vid )) == NULL )
	{
		sprintf( STVERRTXT, "Viewer Not Found for ID=%d - Ignore.",
				vid );

		STV_ERROR( "stv_protoRegion", stvStatusBadParam, STVERRTXT );

		return( stvStatusOk );
	}

	else if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "Viewer ID=%d Found.\n", vid );
		fflush( stdout );
	}

	/* Find View Field Group */

	if ( (VFG = stv_get_view_field_group_id( V->vfglist, V->nvfgs,
			msg->fgid )) == NULL )
	{
		sprintf( STVERRTXT, "View Field Group Not Found for fgid=%d",
				msg->fgid );
		
		STV_ERROR( "stv_protoRegion", stvStatusBadParam, STVERRTXT );

		return( stvStatusBadParam );
	}

	/* Check for Saved Notify... */

	if ( VFG->bail )
	{
		STV_ERROR( "stv_protoRegion", stvStatusBail,
			"Notify Found for View Field Group - Bail" );

		return( stvStatusBail );
	}

	/* Is Task Legit? */

	if ( !stv_check_view_field_group_task( VFG, msg->src ) )
	{
		sprintf( STVERRTXT,
				"Task ID=[ 0x%x / %d ] Not In View Field Group.",
				msg->src, msg->src );
		
		STV_ERROR( "stv_protoRegion", stvStatusBadParam, STVERRTXT );

		return( stvStatusBadParam );
	}

	/* Bump Count, Valid Region Message */

	(VFG->region_cnt)++;

	/* Unpack Region ID */

	stv_unpack_int( msg->bufid, &tmp, 1, 1 );

	if ( tmp != VFG->region_id )
	{
		if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_FLUSH ) )
		{
			printf( "Old Region ID %d.\n", tmp );
			fflush( stdout );
		}

		VFG->frame_ok = stvFalse;
	}

	else if ( VFG->frame_ok )
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
		{
			printf( "\tRegion ID %d O.K.\n", tmp );
			fflush( stdout );
		}

		/* Unpack Task Lock Status */

		waslocked = ( VFG->locked == stvAppLocked )
				? stvTrue : stvFalse;

		stv_unpack_int( msg->bufid, &(VFG->locked), 1, 1 );

		/* Unpack Current Timestamp (Reset Region Time As Needed) */

		stv_unpack_int( msg->bufid, &tmp, 1, 1 );

		if ( VFG->locked == stvAppLocked )
			VFG->region_time = tmp;

		else if ( waslocked )
			VFG->region_time += VFG->eff_freq;

		/* Unpack Field SubRegion Data */

		cc = stv_unpack_elems_SubRegion( msg->bufid, VFG, V );

		if ( cc == stvStatusOk )
		{
			/* Dump Out Updated Field */

			if ( STV_OPT_DUMP_ALL )
			{
				for ( i=0 ; i < VFG->nvfields ; i++ )
				{
					F = (VFG->view_field_list[i])->field;

					if ( F->decomp != NULL )
					{
						printf( "\nField Frame Data for \"%s\":\n\n",
								F->name );
						fflush( stdout );

						stv_dump_vfg_field_array( VFG, i );
					}
				}
			}

			/* Send Steering XON */

			if ( VFG->nvfields == 0 && VFG->region_cnt == VFG->ntids )
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_STEER )
				{
					printf( "\tSending Steer XON Time = " );
					stv_dump_time_str( stdout, VFG->region_time,
							stvTrue );
					fflush( stdout );
				}

				stv_viewer_send_XON( VFG, V );

				VFG->region_cnt = 0;
			}
		}

		else
			VFG->frame_ok = stvFalse;
	}

	return( stvStatusOk );
}


/*** Viewer Update Routines ***/


/* Viewer Vis Region Update Routine */

int
stv_viewer_set_VisRegion( VFG, V, R, C )
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
STV_REGION R;
STV_CELL C;
{
	/* Verify Not Static View Field Group */

	if ( VFG->flag & stvFieldRqstStatic )
	{
		STV_ERROR( "stv_viewer_set_VisRegion", stvStatusBadParam,
			"Attempt to Change Static View Field Group - Ignored." );

		return( stvStatusBadParam );
	}

	/* Copy Over Region & Cell, Submit Request */

	stv_copy_region( VFG->visregion, R );

	stv_copy_cell( VFG->viscell, C );

	return( stv_viewer_send_VisRegion( VFG, V ) );
}


/* Viewer Vis Frequency Update Routine */

void
stv_viewer_set_VisFrequency( VFG, V, freq )
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
int freq;
{
	VFG->freq = freq;

	VFG->eff_freq = stv_calculate_eff_freq( VFG->freq,
			VFG->view_field_list[0]->field->time_scale );

	stv_viewer_send_VisFrequency( VFG, V );
}


/*** Message Routines ***/


/* Send First Initialization Packet INIT0 */

void
stv_viewer_send_Init0( V, dst )
STV_VIEWER V;
int dst;
{
	int sbuf;
	int tmp;

	if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "Sending INIT0 vid=%d seq=%d dst=[0x%x/%d]\n",
			V->id, V->init_seq, dst, dst );
		fflush( stdout );
	}

	stv_init_msg_buf( stvDataDefault, &sbuf );

	tmp = stvMsgOp( stvInit0 );
	stv_pack_int( sbuf, &tmp, 1, 1 );

	/* dummy fgid, for message header consistency */
	tmp = -1;
	stv_pack_int( sbuf, &tmp, 1, 1 );

	stv_pack_int( sbuf, &(V->mytid), 1, 1 );

	stv_pack_int( sbuf, &(V->id), 1, 1 );

	stv_pack_int( sbuf, &(V->init_seq), 1, 1 );

	stv_send_msg( sbuf, dst, V->msgtag, STV_APP_CTX );

	stv_free_msg_buf( sbuf );
}


/* Pack Info on Defined Parameters and Fields */

int
stv_pack_app_info( sbuf, ntasks, P, np, F, nf, mask )
int sbuf;
int ntasks;
STV_PARAMPTR P;
int np;
STV_FIELDPTR F;
int nf;
int mask;
{
	STV_FIELDPTR fptr;

	STV_PARAMPTR pptr;

	int num;
	int tmp;
	int cc;
	int i;

	/* Pack Nprocs */

	stv_pack_int( sbuf, &ntasks, 1, 1 );

	/* Pack # of Fields */

	fptr = F;

	num = 0;

	while ( fptr != NULL )
	{
		if ( fptr->aflag & mask )
			num++;

		fptr = fptr->next;
	}

	stv_pack_int( sbuf, &num, 1, 1 );

	/* Pack Field Info */

	fptr = F;

	while ( fptr != NULL )
	{
		if ( fptr->aflag & mask )
		{
			tmp = stvMarkerField;
			stv_pack_int( sbuf, &tmp, 1, 1 );

			stv_pack_str( sbuf, fptr->name );
			stv_pack_int( sbuf, &(fptr->type), 1, 1 );
			stv_pack_int( sbuf, &(fptr->storage_order), 1, 1 );
			stv_pack_int( sbuf, &(fptr->time_scale), 1, 1 );

			if ( fptr->decomp != NULL )
			{
				tmp = stvMarkerFieldDecomp;
				stv_pack_int( sbuf, &tmp, 1, 1 );

				stv_pack_int( sbuf, &(fptr->decomp->dataDim), 1, 1 );

				stv_pack_int( sbuf,
					fptr->decomp->axisType, fptr->decomp->dataDim, 1 );

				stv_pack_int( sbuf,
					fptr->decomp->axisInfo, fptr->decomp->dataDim, 1 );

				stv_pack_int( sbuf,
					fptr->decomp->nbounds, fptr->decomp->dataDim, 1 );

				for ( i=0 ; i < fptr->decomp->dataDim ; i++ )
				{
					if ( fptr->decomp->nbounds[i] > 0 )
					{
						stv_pack_int( sbuf,
							fptr->decomp->axisLBounds[i],
							fptr->decomp->nbounds[i], 1 );

						stv_pack_int( sbuf,
							fptr->decomp->axisUBounds[i],
							fptr->decomp->nbounds[i], 1 );
					}
				}

				stv_pack_int( sbuf, fptr->decomp->glb,
					fptr->decomp->dataDim, 1 );

				stv_pack_int( sbuf, fptr->decomp->gub,
					fptr->decomp->dataDim, 1 );

				tmp = ( fptr->arrayOffset != NULL )
					? stvTrue : stvFalse;

				stv_pack_int( sbuf, &tmp, 1, 1 );

				if ( fptr->arrayOffset != NULL )
				{
					stv_pack_int( sbuf, fptr->arrayOffset,
						fptr->decomp->dataDim, 1 );
				}

				stv_pack_int( sbuf, fptr->arrayDecl,
					fptr->decomp->dataDim, 1 );

				stv_pack_int( sbuf, &(fptr->decomp->prank), 1, 1 );

				if ( fptr->decomp->prank > 0 )
				{
					stv_pack_int( sbuf, fptr->decomp->pshape,
						fptr->decomp->prank, 1 );

					stv_pack_int( sbuf, fptr->paddr,
						fptr->decomp->prank, 1 );
				}
			}

			else if ( fptr->particle != NULL )
			{
				tmp = stvMarkerFieldParticle;
				stv_pack_int( sbuf, &tmp, 1, 1 );

				stv_pack_str( sbuf, fptr->particle->name );

				stv_pack_int( sbuf, &(fptr->nelems), 1, 1 );

				stv_pack_int( sbuf, &(fptr->particle->dataDim), 1, 1 );

				stv_pack_int( sbuf, fptr->particle->glb,
					fptr->particle->dataDim, 1 );

				stv_pack_int( sbuf, fptr->particle->gub,
					fptr->particle->dataDim, 1 );

				stv_pack_int( sbuf, &(fptr->particle->ntasks), 1, 1 );
			}

			else
			{
				tmp = stvMarkerFieldNone;
				stv_pack_int( sbuf, &tmp, 1, 1 );

				sprintf( STVERRTXT,
					"Error: No Decomposition for Field \"%s\".",
					fptr->name );

				STV_ERROR( "stv_pack_app_info", stvStatusBadParam,
					STVERRTXT );
			}

			tmp = stvMarkerFieldEnd;
			stv_pack_int( sbuf, &tmp, 1, 1 );
		}

		fptr = fptr->next;
	}

	/* Pack # of Parameters */

	pptr = P;

	num = 0;

	while ( pptr != NULL )
	{
		if ( pptr->aflag & mask )
			num++;

		pptr = pptr->next;
	}

	stv_pack_int( sbuf, &num, 1, 1 );

	/* Pack Parameter Info */

	pptr = P;

	while ( pptr != NULL )
	{
		if ( pptr->aflag & mask )
		{
			tmp = stvMarkerParam;
			stv_pack_int( sbuf, &tmp, 1, 1 );

			stv_pack_str( sbuf, pptr->name );

			stv_pack_int( sbuf, &(pptr->num), 1, 1 );

			for ( i=0 ; i < pptr->num ; i++ )
			{
				stv_pack_str( sbuf, pptr->pnames[i] );

				stv_pack_int( sbuf, &(pptr->types[i]), 1, 1 );

				if ( (cc = stv_pack_by_type( sbuf, pptr->vars[i], 1, 1,
						pptr->types[i] & stvTypeMask )) != stvStatusOk )
				{
					STV_ERROR( "stv_pack_app_info", cc,
							"Typed Pack Failed." );

					return( cc );
				}
			}

			tmp = stvMarkerParamEnd;
			stv_pack_int( sbuf, &tmp, 1, 1 );
		}

		pptr = pptr->next;
	}

	return( stvStatusOk );
}


/* Unpack Info on Defined Parameters and Fields */

int
stv_unpack_app_info( rbuf, ntasks, params, nparams, fields, nfields )
int rbuf;
int *ntasks;
STV_PARAMPTR **params;
int *nparams;
STV_FIELDPTR **fields;
int *nfields;
{
	STV_PARTICLEPTR PTtmp;

	STV_DECOMPPTR Dtmp;

	STV_FIELDPTR Ftmp;

	STV_PARAMPTR Ptmp;

	char tmpname[1024];

	int i, j, k;
	int nflds;
	int nprms;
	int type;
	int tmp;
	int cc;

	/* Unpack NPROCS */

	stv_unpack_int( rbuf, &tmp, 1, 1 );

	if ( *ntasks == -1 )
	{
		*ntasks = tmp;

		if ( STV_OPT_VERBOSE
				& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
		{
			printf( "NTASKS is %d.\n", *ntasks );
			fflush( stdout );
		}
	}

	else if ( *ntasks != tmp )
	{
		sprintf( STVERRTXT,
			"Application Error: NPROCS Mismatch %d != %d.",
			tmp, *ntasks );

		STV_ERROR( "stv_unpack_app_info",
			stvStatusBadParam, STVERRTXT );
	}

	/* Unpack Task NFields */

	stv_unpack_int( rbuf, &nflds, 1, 1 );

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
	{
		printf( "Task has %d Fields.\n", nflds );
		fflush( stdout );
	}

	/* Unpack Task Fields */

	if ( nflds > 0 )
	{
		if ( fields != NULL )
		{
			*fields = (STV_FIELDPTR *) MALLOC( (STV_SIZE_T) nflds
				* sizeof(STV_FIELDPTR) );
			STV_MEMCHECK_return( *fields, "Task Field Pointer List" );
		}

		for ( i=0 ; i < nflds ; i++ )
		{
			/* Unpack Field Data (tmp struct) */

			Ftmp = stv_create_field();

			STV_NO_MEM_return( Ftmp, "stv_unpack_app_info" );

			stv_unpack_int( rbuf, &tmp, 1, 1 );

			if ( tmp != stvMarkerField )
			{
				sprintf( STVERRTXT,
					"Parse Error: Expected %s = %d, Got %d",
					"stvMarkerField", stvMarkerField, tmp );

				STV_ERROR( "stv_unpack_app_info",
					stvStatusBadParam, STVERRTXT );
			}

			stv_unpack_str( rbuf, tmpname );

			Ftmp->name = stv_copy_str( tmpname );

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
			{
				printf( "\tUnpacking Field \"%s\":\n", Ftmp->name );
				fflush( stdout );
			}

			stv_unpack_int( rbuf, &(Ftmp->type), 1, 1 );

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
			{
				printf( "\t\ttype = %s\n",
					stv_data_type_str( Ftmp->type ) );
				fflush( stdout );
			}

			stv_unpack_int( rbuf, &(Ftmp->storage_order), 1, 1 );

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
			{
				printf( "\t\tstorage order = %s\n",
					stv_storage_order_str( Ftmp->storage_order ) );
				fflush( stdout );
			}

			stv_unpack_int( rbuf, &(Ftmp->time_scale), 1, 1 );

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
			{
				printf( "\t\ttime scale = " );
				stv_dump_time_str( stdout, Ftmp->time_scale, stvTrue );
				fflush( stdout );
			}

			/* Unpack Decomposition Type Marker */

			stv_unpack_int( rbuf, &tmp, 1, 1 );

			/* Unpack Decomposition Info (tmp struct) */

			if ( tmp == stvMarkerFieldDecomp )
			{
				Ftmp->decomp = Dtmp = stv_create_decomp();

				STV_NO_MEM_return( Dtmp, "stv_unpack_app_info" );

				stv_unpack_int( rbuf, &(Dtmp->dataDim), 1, 1 );

				if ( STV_OPT_VERBOSE
						& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
				{
					printf( "\t\tDECOMP: dataDim = %d\n",
						Dtmp->dataDim );
					fflush( stdout );
				}

				Dtmp->axisType = (int *) MALLOC(
					(STV_SIZE_T) Dtmp->dataDim * sizeof( int ) );
				STV_MEMCHECK_return( Dtmp->axisType,
					"Decomp Axis Type Array" );

				stv_unpack_int( rbuf, Dtmp->axisType,
					Dtmp->dataDim, 1 );

				Dtmp->axisInfo = (int *) MALLOC(
					(STV_SIZE_T) Dtmp->dataDim * sizeof( int ) );
				STV_MEMCHECK_return( Dtmp->axisInfo,
					"Decomp Axis Info Array" );

				stv_unpack_int( rbuf, Dtmp->axisInfo,
					Dtmp->dataDim, 1 );

				Dtmp->nbounds = (int *) MALLOC(
					(STV_SIZE_T) Dtmp->dataDim * sizeof( int ) );
				STV_MEMCHECK_return( Dtmp->nbounds,
					"Decomp Axis Bounds Array Sizes" );

				stv_unpack_int( rbuf, Dtmp->nbounds,
					Dtmp->dataDim, 1 );

				Dtmp->axisLBounds = (int **) MALLOC(
					(STV_SIZE_T) Dtmp->dataDim * sizeof( int * ) );
				STV_MEMCHECK_return( Dtmp->axisLBounds,
					"Decomp Axis Lower Bounds Array" );

				Dtmp->axisUBounds = (int **) MALLOC(
					(STV_SIZE_T) Dtmp->dataDim * sizeof( int * ) );
				STV_MEMCHECK_return( Dtmp->axisUBounds,
					"Decomp Axis Upper Bounds Array" );

				for ( j=0 ; j < Dtmp->dataDim ; j++ )
				{
					if ( Dtmp->nbounds[j] > 0 )
					{
						Dtmp->axisLBounds[j] = (int *) MALLOC(
							(STV_SIZE_T) Dtmp->nbounds[j]
								* sizeof( int ) );
						STV_MEMCHECK_return( Dtmp->axisLBounds[j],
							"Decomp Axis Lower Bounds Sub-Array" );

						Dtmp->axisUBounds[j] = (int *) MALLOC(
							(STV_SIZE_T) Dtmp->nbounds[j]
								* sizeof( int ) );
						STV_MEMCHECK_return( Dtmp->axisUBounds[j],
							"Decomp Axis Upper Bounds Sub-Array" );

						stv_unpack_int( rbuf, Dtmp->axisLBounds[j],
							Dtmp->nbounds[j], 1 );

						stv_unpack_int( rbuf, Dtmp->axisUBounds[j],
							Dtmp->nbounds[j], 1 );
					}

					else
					{
						Dtmp->axisLBounds[j] = (int *) NULL;
						Dtmp->axisUBounds[j] = (int *) NULL;
					}
				}

				Dtmp->glb = (int *) MALLOC(
					(STV_SIZE_T) Dtmp->dataDim * sizeof( int ) );
				STV_MEMCHECK_return( Dtmp->glb,
					"Decomp Global Lower Bound" );

				stv_unpack_int( rbuf,
					Dtmp->glb, Dtmp->dataDim, 1 );

				Dtmp->gub = (int *) MALLOC(
					(STV_SIZE_T) Dtmp->dataDim * sizeof( int ) );
				STV_MEMCHECK_return( Dtmp->gub,
					"Decomp Global Upper Bound" );

				stv_unpack_int( rbuf,
					Dtmp->gub, Dtmp->dataDim, 1 );

				if ( STV_OPT_VERBOSE
						& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
				{
					for ( j=0 ; j < Dtmp->dataDim ; j++ )
					{
						if ( Dtmp->nbounds[j] > 0
							&& Dtmp->axisType[j] == stvExplicit )
						{
							printf( "\t\t\t[%d]: t=%s glb=%d gub=%d\n",
								j, stv_decomp_type_str(
									Dtmp->axisType[j] ),
								Dtmp->glb[j], Dtmp->gub[j] );

							for ( k=0 ; k < Dtmp->nbounds[j] ; k++ )
							{
								printf( "\t\t\t\ti=%d,%d\n",
									Dtmp->axisLBounds[j][k],
									Dtmp->axisUBounds[j][k] );
							}
						}

						else
						{
							printf(
						"\t\t\t[%d]: t=%s info=%d glb=%d gub=%d\n",
								j, stv_decomp_type_str(
									Dtmp->axisType[j] ),
								Dtmp->axisInfo[j],
								Dtmp->glb[j], Dtmp->gub[j] );
						}
					}

					fflush( stdout );
				}

				stv_unpack_int( rbuf, &tmp, 1, 1 );

				if ( tmp )
				{
					Ftmp->arrayOffset = (int *) MALLOC(
						(STV_SIZE_T) Dtmp->dataDim * sizeof( int ) );
					STV_MEMCHECK_return( Ftmp->arrayOffset,
						"Array Offset Array" );

					stv_unpack_int( rbuf,
						Ftmp->arrayOffset, Dtmp->dataDim, 1 );
				}

				else
					Ftmp->arrayOffset = (int *) NULL;

				Ftmp->arrayDecl = (int *) MALLOC(
					(STV_SIZE_T) Dtmp->dataDim * sizeof( int ) );
				STV_MEMCHECK_return( Ftmp->arrayDecl,
					"Array Decl Array" );

				stv_unpack_int( rbuf,
					Ftmp->arrayDecl, Dtmp->dataDim, 1 );

				stv_unpack_int( rbuf, &(Dtmp->prank), 1, 1 );

				if ( STV_OPT_VERBOSE
						& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
				{
					printf( "\t\tprank = %d\n", Dtmp->prank );
					fflush( stdout );
				}

				if ( Dtmp->prank > 0 )
				{
					Dtmp->pshape = (int *) MALLOC(
						(STV_SIZE_T) Dtmp->prank * sizeof( int ) );
					STV_MEMCHECK_return( Dtmp->pshape,
						"Decomp Processor Shape Array" );

					stv_unpack_int( rbuf,
						Dtmp->pshape, Dtmp->prank, 1 );

					if ( STV_OPT_VERBOSE
							& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
					{
						for ( j=0 ; j < Dtmp->prank ; j++ )
						{
							printf( "\t\t\tpshape[%d] = %d\n",
								j, Dtmp->pshape[j] );
						}

						fflush( stdout );
					}

					Ftmp->paddr = (int *) MALLOC(
						(STV_SIZE_T) Dtmp->prank * sizeof( int ) );
					STV_MEMCHECK_return( Ftmp->paddr,
						"Processor Address Array" );

					stv_unpack_int( rbuf, Ftmp->paddr, Dtmp->prank, 1 );

					if ( STV_OPT_VERBOSE
							& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
					{
						for ( j=0 ; j < Dtmp->prank ; j++ )
						{
							printf( "\t\t\tpaddr[%d] = %d\n",
								j, Ftmp->paddr[j] );
						}

						fflush( stdout );
					}
				}
			}

			/* Unpack Particle Information (tmp struct) */

			else if ( tmp == stvMarkerFieldParticle )
			{
				Ftmp->particle = PTtmp = stv_create_particle();

				STV_NO_MEM_return( PTtmp, "stv_unpack_app_info" );

				stv_unpack_str( rbuf, tmpname );

				PTtmp->name = stv_copy_str( tmpname );

				if ( STV_OPT_VERBOSE
						& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
				{
					printf( "\t\tParticle \"%s\":\n", PTtmp->name );
					fflush( stdout );
				}

				stv_unpack_int( rbuf, &(Ftmp->nelems), 1, 1 );

				if ( STV_OPT_VERBOSE
						& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
				{
					printf( "\t\t\tnelems = %d\n", Ftmp->nelems );
					fflush( stdout );
				}

				stv_unpack_int( rbuf, &(PTtmp->dataDim), 1, 1 );

				if ( STV_OPT_VERBOSE
						& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
				{
					printf( "\t\t\tdataDim = %d\n", PTtmp->dataDim );
					fflush( stdout );
				}

				if ( PTtmp->dataDim > 0 )
				{
					PTtmp->glb = (int *) MALLOC(
						(STV_SIZE_T) PTtmp->dataDim * sizeof( int ) );
					STV_MEMCHECK_return( PTtmp->glb,
						"Particle Global Lower Bound" );

					stv_unpack_int( rbuf,
						PTtmp->glb, PTtmp->dataDim, 1 );

					PTtmp->gub = (int *) MALLOC(
						(STV_SIZE_T) PTtmp->dataDim * sizeof( int ) );
					STV_MEMCHECK_return( PTtmp->gub,
						"Particle Global Upper Bound" );

					stv_unpack_int( rbuf,
						PTtmp->gub, PTtmp->dataDim, 1 );

					if ( STV_OPT_VERBOSE
							& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
					{
						for ( j=0 ; j < PTtmp->dataDim ; j++ )
						{
							printf( "\t\t\t\t[%d]: glb=%d gub=%d\n",
								j, PTtmp->glb[j], PTtmp->gub[j] );
						}

						fflush( stdout );
					}
				}

				stv_unpack_int( rbuf, &(PTtmp->ntasks), 1, 1 );

				if ( STV_OPT_VERBOSE
						& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
				{
					printf( "\t\t\tntasks = %d\n", PTtmp->ntasks );
					fflush( stdout );
				}
			}

			/* No Decomposition Information!!! */

			else if ( tmp == stvMarkerFieldNone )
			{
				sprintf( STVERRTXT, "No Decomposition for Field %s.",
					Ftmp->name );

				STV_ERROR( "stv_unpack_app_info",
					stvStatusBadParam, STVERRTXT );
			}

			/* Bogus Decomp Type */

			else
			{
				sprintf( STVERRTXT,
					"Parse Error: Expected Decomp Type, Got %d", tmp );

				STV_ERROR( "stv_unpack_app_info",
					stvStatusBadParam, STVERRTXT );
			}

			/* Get End of Field Safety Marker */

			stv_unpack_int( rbuf, &tmp, 1, 1 );

			if ( tmp != stvMarkerFieldEnd )
			{
				sprintf( STVERRTXT,
					"Parse Error: Expected %s = %d, Got %d",
					"stvMarkerFieldEnd", stvMarkerFieldEnd, tmp );

				STV_ERROR( "stv_unpack_app_info",
					stvStatusBadParam, STVERRTXT );
			}

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
			{
				printf( "\tDone.\n" );
				fflush( stdout );
			}

			/* Set Field Pointer */

			if ( fields != NULL )
				(*fields)[i] = Ftmp;
			
			else
			{
				stv_free_decomp( &(Ftmp->decomp) );
				stv_free_particle( &(Ftmp->particle) );

				FREE( Ftmp->paddr );

				stv_free_field( &Ftmp );
			}
		}
	}

	else
	{
		if ( fields != NULL )
			*fields = (STV_FIELDPTR *) NULL;
	}

	if ( nfields != NULL )
		*nfields = nflds;

	/* Unpack Task NParams */

	stv_unpack_int( rbuf, &nprms, 1, 1 );

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
	{
		printf( "Task has %d Parameters.\n", nprms );
		fflush( stdout );
	}

	/* Unpack Task Parameters */

	if ( nprms > 0 )
	{
		if ( params != NULL )
		{
			*params = (STV_PARAMPTR *) MALLOC(
				(STV_SIZE_T) nprms * sizeof(STV_PARAMPTR) );
			STV_MEMCHECK_return( *params, "Task Parameter Ptr List" );
		}

		for ( i=0 ; i < nprms ; i++ )
		{
			/* Unpack Parameter Data (tmp struct) */

			Ptmp = stv_create_param();

			STV_NO_MEM_return( Ptmp, "stv_unpack_app_info" );

			stv_unpack_int( rbuf, &tmp, 1, 1 );

			if ( tmp != stvMarkerParam )
			{
				sprintf( STVERRTXT,
					"Parse Error: Expected %s = %d, Got %d",
					"stvMarkerParam", stvMarkerParam, tmp );

				STV_ERROR( "stv_unpack_app_info",
					stvStatusBadParam, STVERRTXT );
			}

			stv_unpack_str( rbuf, tmpname );

			Ptmp->name = stv_copy_str( tmpname );

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
			{
				printf( "\tUnpacking Parameter \"%s\":\n", Ptmp->name );
				fflush( stdout );
			}

			stv_unpack_int( rbuf, &(Ptmp->num), 1, 1 );

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
			{
				printf( "\t\t%d parameter vector\n", Ptmp->num );
				fflush( stdout );
			}

			if ( Ptmp->num > 0 )
			{
				Ptmp->pnames = (char **) MALLOC(
					(STV_SIZE_T) Ptmp->num * sizeof( char * ) );
				STV_MEMCHECK_return( Ptmp->pnames,
					"Parameter Vector Names List" );

				Ptmp->vars = (STV_VALUE *) MALLOC(
					(STV_SIZE_T) Ptmp->num * sizeof( STV_VALUE ) );
				STV_MEMCHECK_return( Ptmp->vars,
					"Parameter Vector Values List" );

				Ptmp->types = (int *) MALLOC(
					(STV_SIZE_T) Ptmp->num * sizeof( int ) );
				STV_MEMCHECK_return( Ptmp->types,
					"Parameter Vector Types List" );

				for ( j=0 ; j < Ptmp->num ; j++ )
				{
					stv_unpack_str( rbuf, tmpname );

					Ptmp->pnames[j] = stv_copy_str( tmpname );

					stv_unpack_int( rbuf, &(Ptmp->types[j]), 1, 1 );

					if ( STV_OPT_VERBOSE
							& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
					{
						printf( "\t\t\t%s: type = %s\n",
							Ptmp->pnames[j],
							stv_data_type_str(
								Ptmp->types[j] & stvTypeMask ) );
						fflush( stdout );
					}

					type = Ptmp->types[j] & stvTypeMask;

					Ptmp->vars[j] = (STV_VALUE) MALLOC(
						STV_SIZEOF( type ) );
					STV_MEMCHECK_return( Ptmp->vars[j],
						"Parameter Var Ptr" );

					if ( (cc = stv_unpack_by_type( rbuf,
						Ptmp->vars[j], 1, 1, type )) != stvStatusOk )
					{
						STV_ERROR( "stv_unpack_app_info",
							cc, "Typed Unpack Failed." );
						
						return( cc );
					}
				}
			}

			stv_unpack_int( rbuf, &tmp, 1, 1 );

			if ( tmp != stvMarkerParamEnd )
			{
				sprintf( STVERRTXT,
					"Parse Error: Expected %s = %d, Got %d",
					"stvMarkerParamEnd", stvMarkerParamEnd, tmp );

				STV_ERROR( "stv_unpack_app_info",
					stvStatusBadParam, STVERRTXT );
			}

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER | STV_DEBUG_APPINFO ) )
			{
				printf( "\tDone.\n" );
				fflush( stdout );
			}

			/* Set Parameter Pointer */

			if ( params != NULL )
				(*params)[i] = Ptmp;
			
			else
			{
				for ( j=0 ; j < Ptmp->num ; j++ )
					FREE( Ptmp->vars[j] );

				stv_free_param( &Ptmp );
			}
		}
	}

	else
	{
		if ( params != NULL )
			*params = (STV_PARAMPTR *) NULL;
	}

	if ( nparams != NULL )
		*nparams = nprms;

	return( stvStatusOk );
}


/* Receive and Process INIT1 Application Replies */

int
stv_viewer_recv_Init1( V )
STV_VIEWER V;
{
	int cc;

	if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "%s:  starting nvtasks = %d\n",
				"stv_viewer_recv_Init1()", V->nvtasks );
	}

	do
	{
		if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MSG ) )
		{
			printf( "Blocking for INIT1 Reply Messages.\n" );
			fflush( stdout );
		}

		if ( (cc = stv_handleMessages( stvInit1,
				stvAny, V->msgtag, STV_VIEWER_CTX,
				stvMsgBlockTimeout )) < 0 )
		{
			STV_ERROR( "stv_viewer_recv_Init1", cc,
			"Receiving Init1 Reply From Application." );

			return( cc );
		}

		else if ( !cc )
		{
			STV_ERROR( "stv_viewer_recv_Init1", stvStatusBadParam,
			"Warning: Timed Out - No Init1 Reply From Application." );

			return( stvStatusBadParam );
		}
	}
	while ( V->nvtasks < V->ntasks || V->ntasks == -1 );

	return( stvStatusOk );
}


/* Receive and Process One INIT1 Application Reply */

int
stv_protoInit1( msg )
STV_BUFPTR msg;
{
	STV_VIEW_FIELD VFptr;
	STV_VIEW_FIELD VF;

	STV_VIEW_PARAM VPptr;
	STV_VIEW_PARAM VP;

	STV_VIEW_TASK VT;

	STV_PARTICLEPTR PTtmp;
	STV_PARTICLEPTR PT;

	STV_DECOMPPTR Dtmp;
	STV_DECOMPPTR D;

	STV_FIELDPTR Ftmp;
	STV_FIELDPTR F;

	STV_PARAMPTR Ptmp;
	STV_PARAMPTR P;

	STV_VIEWER V;

	int seqck;
	int i, j;
	int tmp;
	int vid;
	int cc;

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MSG ) )
	{
		printf( "Entry stv_protoInit1().\n" );
		fflush( stdout );
	}

	if ( msg == (STV_BUFPTR) NULL )
		return( stvStatusOk );

	/* Unpack Viewer ID */

	stv_unpack_int( msg->bufid, &vid, 1, 1 );

	if ( (V = stv_find_viewer( vid )) == NULL )
	{
		sprintf( STVERRTXT, "Viewer Not Found for ID=%d", vid );

		STV_ERROR( "stv_protoInit1", stvStatusBadParam, STVERRTXT );

		return( stvStatusBadParam );
	}

	else if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "Viewer ID=%d Found.\n", vid );
		fflush( stdout );
	}

	/* Unpack & Check Init Sequence */

	seqck = 0;

	stv_unpack_int( msg->bufid, &tmp, 1, 1 );

	if ( tmp == V->init_seq )
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
		{
			printf( "Sequence #%d O.K.\n", tmp );
			fflush( stdout );
		}

		seqck++;
	}

	else if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "Old Sequence #%d, Should Be #%d - Ignore.\n",
			tmp, V->init_seq );
		fflush( stdout );

		return( stvStatusOk );
	}

	/* Verify That Everyone is on the Same Page! */
	/* Check the Application Generation (# of Restarts) */

	stv_unpack_int( msg->bufid, &tmp, 1, 1 );

	if ( tmp != V->init_gen )
	{
		/* Not Set Yet */
		if ( V->init_gen == -1 )
			V->init_gen = tmp;

		/* Mismatch */
		else
		{
			seqck = 0;

			if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
			{
				printf( "Generation Mismatch:  %d != %d, Ignore...\n",
					tmp, V->init_gen );
				fflush( stdout );
			}

			return( stvStatusOk );
		}
	}

	if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "Generation #%d O.K.\n", tmp );
		fflush( stdout );
	}

	/* Check for Existing Task Structure */
	/* (From Duplicate Timeout Messages) */

	VT = V->vtasks;

	while ( VT != NULL && seqck )
	{
		if ( VT->tid == msg->src )
		{
			seqck = 0;

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER
							| STV_DEBUG_FLUSH ) )
			{
				printf( "Duplicate Init1 from 0x%x/%d.\n",
					msg->src, msg->src );
				printf( "Dropping...\n" );
			}
		}

		VT = VT->next;
	}

	/* Add New Task */

	if ( seqck )
	{
		/* Create New Task Structure */

		if ( V->vtasks == NULL )
			VT = V->vtasks = stv_create_view_task();

		else
		{
			VT = stv_create_view_task();

			if ( VT != NULL )
				VT->next = V->vtasks;

			V->vtasks = VT;
		}

		STV_NO_MEM_return( VT, "stv_protoInit1" );

		VT->tid = msg->src;

		(V->nvtasks)++;

		stv_unpack_app_info( msg->bufid, &(V->ntasks),
			&(VT->params), &(VT->nparams),
			&(VT->fields), &(VT->nfields) );

		for ( i=0 ; i < VT->nfields ; i++ )
		{
			Ftmp = VT->fields[i];

			Dtmp = Ftmp->decomp;

			PTtmp = Ftmp->particle;

			/* Look For Existing Field */

			F = stv_get_field_name( Ftmp->name,
				V->fields, V->nfields );

			if ( F != NULL )
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
				{
					printf( "\t\tExisting Field \"%s\".\n",
						F->name );
					fflush( stdout );
				}

				/* Check Against Existing Field */

				if ( !stv_match_field( Ftmp, F ) )
				{
					sprintf( STVERRTXT,
						"Field Mismatch \"%s\".", F->name );

					STV_ERROR( "stv_protoInit1",
						stvStatusBadParam, STVERRTXT );
				}

				stv_free_field( &Ftmp );

				/* Check Against Existing Decomp */

				if ( Dtmp != NULL &&
					!stv_match_decomp( Dtmp, F->decomp ) )
				{
					sprintf( STVERRTXT,
						"Decomp Mismatch \"%s\".", F->name );

					STV_ERROR( "stv_protoInit1",
						stvStatusBadParam, STVERRTXT );
				}

				stv_free_decomp( &Dtmp );

				/* Check Against Existing Particle */

				if ( PTtmp != NULL &&
					!stv_match_particle( PTtmp, F->particle ) )
				{
					sprintf( STVERRTXT,
						"Particle Mismatch \"%s\".", F->name );

					STV_ERROR( "stv_protoInit1",
						stvStatusBadParam, STVERRTXT );
				}

				stv_free_particle( &PTtmp );

				/* Add Task ID to View Field List */

				VF = stv_get_view_field_name( F->name,
					V->vfields, V->nvfields );

				VF->tids = stv_add_to_list( VF->tids,
					VF->ntids, VT->tid );

				(VF->ntids)++;

				/* Set View Task Field Pointer */

				VT->fields[i] = F;
			}

			/* Add New Field/Decomp/Particle */

			else
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
				{
					printf( "\t\tNew Field \"%s\".\n",
						Ftmp->name );
					fflush( stdout );
				}

				/* Add New Field */

				if ( V->fields == NULL )
					V->fields = Ftmp;

				else
				{
					Ftmp->next = V->fields;

					V->fields = Ftmp;
				}

				Ftmp->fieldID = (V->nfields)++;

				F = Ftmp;

				/* Add New View Field */

				if ( V->vfields == NULL )
					VF = V->vfields = stv_create_view_field();

				else
				{
					VF = stv_create_view_field();

					VFptr = V->vfields;

					while ( VFptr->next != NULL )
						VFptr = VFptr->next;

					VFptr->next = VF;
				}

				STV_NO_MEM_return( VF, "stv_protoInit1" );

				(V->nvfields)++;

				VF->field = F;

				VF->tids = stv_add_to_list( (int *) NULL, 0,
					VT->tid );

				VF->ntids = 1;

				/* Look for Existing Decomp */

				if ( Dtmp != NULL )
				{
					D = stv_get_decomp( Dtmp,
						V->decomps, V->ndecomps );

					if ( D == NULL )
					{
						if ( STV_OPT_VERBOSE
								& STV_DEBUG_VIEWER )
						{
							printf(
								"\t\t\tAdding New Decomp.\n" );
							fflush( stdout );
						}

						/* Add New Decomp */

						if ( V->decomps == NULL )
							V->decomps = Dtmp;

						else
						{
							Dtmp->next = V->decomps;

							V->decomps = Dtmp;
						}

						(V->ndecomps)++;
					}

					else
					{
						if ( STV_OPT_VERBOSE
								& STV_DEBUG_VIEWER )
						{
							printf(
								"\t\t\tExisting Decomp.\n" );
							fflush( stdout );
						}

						F->decomp = D;

						stv_free_decomp( &Dtmp );
					}
				}

				/* Look for Existing Particle */

				if ( PTtmp != NULL )
				{
					PT = stv_get_particle( PTtmp,
						V->particles, V->nparticles );

					if ( PT == NULL )
					{
						if ( STV_OPT_VERBOSE
								& STV_DEBUG_VIEWER )
						{
							printf(
						"\t\t\tAdding New Particle \"%s\".\n",
								PTtmp->name );
							fflush( stdout );
						}

						/* Add New Particle */

						if ( V->particles == NULL )
							V->particles = PTtmp;

						else
						{
							PTtmp->next = V->particles;

							V->particles = PTtmp;
						}

						(V->nparticles)++;
					}

					else
					{
						if ( STV_OPT_VERBOSE
								& STV_DEBUG_VIEWER )
						{
							printf(
							"\t\t\tExisting Particle \"%s\".\n",
								PT->name );
							fflush( stdout );
						}

						F->particle = PT;

						stv_free_particle( &PTtmp );
					}
				}
			}
		}

		for ( i=0 ; i < VT->nparams ; i++ )
		{
			Ptmp = VT->params[i];

			/* Look For Existing Parameter */

			P = stv_get_param_name( Ptmp->name,
				V->params, V->nparams );

			if ( P != NULL )
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
				{
					printf( "\t\tExisting Parameter \"%s\".\n",
						P->name );
					fflush( stdout );
				}

				/* Check Against Existing Parameters */

				if ( !stv_match_param( Ptmp, P ) )
				{
					sprintf( STVERRTXT, "Parm Mismatch \"%s\".",
						P->name );

					STV_ERROR( "stv_protoInit1",
						stvStatusBadParam, STVERRTXT );
				}

				stv_free_param( &Ptmp );

				/* Add Task ID to View Parm List */

				VP = stv_get_view_param_name( P->name,
					V->vparams, V->nvparams );

				VP->tids = stv_add_to_list( VP->tids,
					VP->ntids, VT->tid );

				(VP->ntids)++;

				/* Set View Task Parameter Pointer */

				VT->params[i] = P;
			}

			/* Add New Parameter */

			else
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
				{
					printf( "\t\tNew Parameter \"%s\".\n",
						Ptmp->name );
					fflush( stdout );
				}

				/* Add New Parameter */

				if ( V->params == NULL )
					V->params = Ptmp;

				else
				{
					Ptmp->next = V->params;

					V->params = Ptmp;
				}

				Ptmp->paramID = (V->nparams)++;

				P = Ptmp;

				/* Add New View Parameter */

				if ( V->vparams == NULL )
					VP = V->vparams = stv_create_view_param();

				else
				{
					VP = stv_create_view_param();

					VPptr = V->vparams;

					while ( VPptr->next != NULL )
						VPptr = VPptr->next;

					VPptr->next = VP;
				}

				STV_NO_MEM_return( VP, "stv_protoInit1" );

				(V->nvparams)++;

				VP->param = P;

				VP->token = stvSteerNot;

				VP->changed = stvFalse;

				VP->tids = stv_add_to_list( (int *) NULL, 0,
					VT->tid );

				VP->ntids = 1;
			}
		}
	}

	return( stvStatusOk );
}


/* Send Field Request to Application */

void
stv_viewer_send_FieldRqst( VFG, V, no_synch, is_static )
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
int no_synch;
int is_static;
{
	STV_VIEW_FIELD VF;

	STV_VIEW_PARAM VP;

	STV_FIELDPTR F;

	STV_PARAMPTR P;

	int sbuf;
	int tmp;
	int i;

	if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "Sending FieldRqst vid=%d fgid=%d region_id=%d\n",
			V->id, VFG->fgid, VFG->region_id );
		fflush( stdout );
	}

	/* Init Message */

	stv_init_msg_buf( stvDataDefault, &sbuf );

	tmp = stvMsgOp( stvFieldRqst );
	stv_pack_int( sbuf, &tmp, 1, 1 );

	/* Pack Field Group ID */

	stv_pack_int( sbuf, &(VFG->fgid), 1, 1 );

	/* Pack Viewer ID */

	stv_pack_int( sbuf, &(V->id), 1, 1 );

	/* Pack Visualization Info */

	stv_pack_int( sbuf, &(VFG->region_id), 1, 1 );

	stv_pack_int( sbuf, VFG->visregion,
		STV_REGION_SIZE( STV_MAX_DIM ), 1 );

	stv_pack_int( sbuf, VFG->viscell,
		STV_CELL_SIZE( STV_MAX_DIM ), 1 );

	stv_pack_int( sbuf, &(VFG->eff_freq), 1, 1 );
	stv_pack_int( sbuf, &(VFG->freq), 1, 1 );

	stv_pack_int( sbuf, &no_synch, 1, 1 );

	stv_pack_int( sbuf, &is_static, 1, 1 );

	/* Pack Fields List */

	stv_pack_int( sbuf, &(VFG->nvfields), 1, 1 );

	for ( i=0 ; i < VFG->nvfields ; i++ )
	{
		VF = VFG->view_field_list[i];

		F = VF->field;

		stv_pack_str( sbuf, F->name );

		stv_pack_int( sbuf, &(F->fieldID), 1, 1 );

		stv_pack_int( sbuf, &(VFG->view_types[i]), 1, 1 );

		stv_pack_int( sbuf, &(VFG->view_storage_orders[i]), 1, 1 );
	}

	/* Pack Parameter Map */

	stv_pack_int( sbuf, &(V->nvparams), 1, 1 );

	VP = V->vparams;

	while ( VP != NULL )
	{
		P = VP->param;

		stv_pack_str( sbuf, P->name );

		stv_pack_int( sbuf, &(P->paramID), 1, 1 );

		VP = VP->next;
	}

	/* Send Request to Task IDs */

	stv_mcast_msg( sbuf, VFG->tid_list, VFG->ntids,
			V->msgtag, STV_APP_CTX );

	stv_free_msg_buf( sbuf );
}


/* Send Viewer a FieldTime Reply */

void
stv_send_viewer_FieldTime( dst, fgid, vid, ftime, ack, tag )
int dst;
int fgid;
int vid;
int ftime;
int tag;
{
	int sbuf;
	int tmp;

	if ( STV_OPT_VERBOSE
			& ( STV_DEBUG_ULIB | STV_DEBUG_TIME | STV_DEBUG_FREQ ) )
	{
		printf(
			"Sending FieldTime vid=%d dst=[0x%x/%d] fgid=%d ftime=%d\n",
			vid, dst, dst, fgid, ftime );
		fflush( stdout );
	}

	stv_init_msg_buf( stvDataDefault, &sbuf );

	tmp = stvMsgOp( stvFieldTime );
	stv_pack_int( sbuf, &tmp, 1, 1 );

	stv_pack_int( sbuf, &fgid, 1, 1 );

	stv_pack_int( sbuf, &vid, 1, 1 );

	stv_pack_int( sbuf, &ftime, 1, 1 );

	stv_pack_int( sbuf, &ack, 1, 1 );

	stv_pack_int( sbuf, &STV_APP_LOCK, 1, 1 );

	stv_send_msg( sbuf, dst, tag, STV_VIEWER_CTX );

	stv_free_msg_buf( sbuf );
}


/* Receive FieldTime Replies from View Field Tasks */

int
stv_viewer_recv_FieldTime( VFG, V )
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
{
	STV_VIEW_TASK VT;

	int *tids;

	int i, j;
	int cc;

	/* Check for Saved Notify */

	if ( VFG->bail )
	{
		STV_ERROR( "stv_viewer_recv_FieldTime", stvStatusBail,
			"Notify Found for View Field Group - Bail..." );

		return( stvStatusBail );
	}

	/* Handle Field Time Replies */

	VFG->locked = stvAppUnlocked;

	VFG->reply_cnt = 0;

	VFG->ack_cnt = 0;

	VFG->timemax = 0;

	do
	{
		if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MSG ) )
		{
			printf( "Blocking for FieldTime Reply Messages.\n" );
			fflush( stdout );
		}

		if ( (cc = stv_handleMessages( stvFieldTime,
				stvAny, V->msgtag, STV_VIEWER_CTX,
				stvMsgBlock )) <= 0 )
		{
			STV_ERROR( "stv_viewer_recv_FieldTime", cc,
					"Receiving FieldTime Reply From Application." );
			
			return( cc );
		}

		/* Check for Saved Notify */

		if ( VFG->bail )
		{
			STV_ERROR( "stv_viewer_recv_FieldTime", stvStatusBail,
				"Notify Found for View Field Group - Bail..." );

			return( stvStatusBail );
		}
	}
	while ( VFG->reply_cnt < VFG->ntids );

	/* Wipe Out Irrelevant Tasks */

	if ( VFG->nvfields > 0 && VFG->ack_cnt < VFG->reply_cnt )
	{
		/* Allocate New Tids List */

		tids = (int *) MALLOC( (STV_SIZE_T) VFG->ack_cnt
				* sizeof(int) );
		STV_MEMCHECK_return( tids, "Static Field Tids List" );

		for ( i=0 ; i < VFG->ack_cnt ; i++ )
			tids[i] = -1;

		/* Extract Ack-ed Tasks from List... */

		for ( i=0, j=0 ; i < VFG->ntids ; i++ )
		{
			VT = stv_get_view_task_tid( VFG->tid_list[i],
				V->vtasks, V->nvtasks );
		
			if ( VT->ack )
				tids[ j++ ] = VFG->tid_list[i];

			VFG->tid_list[i] = -1;
		}

		/* Replace VFG Tids List */

		FREE( VFG->tid_list );

		VFG->tid_list = tids;

		VFG->ntids = VFG->ack_cnt;
	}

	/* Done */

	return( stvStatusOk );
}


/* Receive and Process One FieldTime Reply */

int
stv_protoFieldTime( msg )
STV_BUFPTR msg;
{
	STV_VIEW_FIELD_GROUP VFG;

	STV_VIEW_TASK VT;

	STV_VIEWER V;

	int tmp;
	int vid;
	int cc;

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MSG ) )
	{
		printf( "Entry stv_protoFieldTime().\n" );
		fflush( stdout );
	}

	if ( msg == (STV_BUFPTR) NULL )
		return( stvStatusOk );

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MSG ) )
	{
		printf( "Message Received From Task ID=[ 0x%x / %d ]\n",
				msg->src, msg->src );
		fflush( stdout );
	}

	/* Unpack Viewer ID */

	stv_unpack_int( msg->bufid, &vid, 1, 1 );

	if ( (V = stv_find_viewer( vid )) == NULL )
	{
		sprintf( STVERRTXT, "Viewer Not Found for ID=%d", vid );

		STV_ERROR( "stv_protoFieldTime", stvStatusBadParam, STVERRTXT );

		return( stvStatusBadParam );
	}

	else if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "Viewer ID=%d Found.\n", vid );
		fflush( stdout );
	}

	/* Find View Field Group */

	if ( (VFG = stv_get_view_field_group_id( V->vfglist, V->nvfgs,
			msg->fgid )) == NULL )
	{
		sprintf( STVERRTXT, "View Field Group Not Found for fgid=%d",
			msg->fgid );

		STV_ERROR( "stv_protoFieldTime", stvStatusBadParam, STVERRTXT );

		return( stvStatusBadParam );
	}

	/* Is Task Legit? */

	if ( !stv_check_view_field_group_task( VFG, msg->src ) )
	{
		sprintf( STVERRTXT,
				"Task ID=[ 0x%x / %d ] Not In View Field Group.",
				msg->src, msg->src );

		STV_ERROR( "stv_protoFieldTime", stvStatusBadParam, STVERRTXT );

		return( stvStatusBadParam );
	}

	/* Get Viewer Task Structure */

	if ( (VT = stv_get_view_task_tid( msg->src, V->vtasks, V->nvtasks ))
			== NULL )
	{
		sprintf( STVERRTXT, "Task ID=[ 0x%x / %d ] Not Found.",
				msg->src, msg->src );

		STV_ERROR( "stv_protoFieldTime", stvStatusBadParam, STVERRTXT );

		return( stvStatusBadParam );
	}

	stv_unpack_int( msg->bufid, &(VT->fieldtime), 1, 1 );

	if ( STV_OPT_VERBOSE
			& ( STV_DEBUG_VIEWER | STV_DEBUG_TIME | STV_DEBUG_FREQ ) )
	{
		printf( "\tField Time is %d.\n", VT->fieldtime );
		fflush( stdout );
	}

	stv_unpack_int( msg->bufid, &(VT->ack), 1, 1 );

	if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "\tField Request Ack is %d.\n", VT->ack );
		fflush( stdout );
	}

	if ( VT->ack )
		(VFG->ack_cnt)++;

	if ( ( VT->ack || !(VFG->nvfields) )
			&& VT->fieldtime > VFG->timemax )
		VFG->timemax = VT->fieldtime;

	stv_unpack_int( msg->bufid, &tmp, 1, 1 );

	if ( tmp == stvAppLocked )
		VFG->locked = stvAppLocked;

	(VFG->reply_cnt)++;

	/* Done */

	return( stvStatusOk );
}


/* Send FieldTstart to Tasks */

void
stv_viewer_send_FieldTstart( VFG, V, t0hat )
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
int t0hat;
{
	int sbuf;
	int tmp;

	if ( STV_OPT_VERBOSE
			& ( STV_DEBUG_VIEWER | STV_DEBUG_TIME | STV_DEBUG_FREQ ) )
	{
		printf( "Sending FieldTstart fgid=%d t0hat=%d\n",
			VFG->fgid, t0hat );
		fflush( stdout );
	}

	stv_init_msg_buf( stvDataDefault, &sbuf );

	tmp = stvMsgOp( stvFieldTstart );
	stv_pack_int( sbuf, &tmp, 1, 1 );

	stv_pack_int( sbuf, &(VFG->fgid), 1, 1 );

	stv_pack_int( sbuf, &t0hat, 1, 1 );

	stv_mcast_msg( sbuf, VFG->tid_list, VFG->ntids,
			V->msgtag, STV_APP_CTX );

	stv_free_msg_buf( sbuf );
}


/* Send Viewer a FieldAck Reply */

void
stv_send_viewer_FieldAck( dst, fgid, vid, ack, tag )
int dst;
int fgid;
int vid;
int ack;
int tag;
{
	int sbuf;
	int tmp;

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
	{
		printf(
			"Sending FieldAck vid=%d dst=[0x%x/%d] fgid=%d ack=%d\n",
			vid, dst, dst, fgid, ack );
		fflush( stdout );
	}

	stv_init_msg_buf( stvDataDefault, &sbuf );

	tmp = stvMsgOp( stvFieldAck );
	stv_pack_int( sbuf, &tmp, 1, 1 );

	stv_pack_int( sbuf, &fgid, 1, 1 );

	stv_pack_int( sbuf, &vid, 1, 1 );

	stv_pack_int( sbuf, &ack, 1, 1 );

	stv_pack_int( sbuf, &STV_APP_LOCK, 1, 1 );

	stv_send_msg( sbuf, dst, tag, STV_VIEWER_CTX );

	stv_free_msg_buf( sbuf );
}


/* Receive FieldAck Replies from View Field Tasks */

int
stv_viewer_recv_FieldAcks( VFG, V )
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
{
	STV_VIEW_TASK VT;

	int *tids;

	int i, j;
	int cc;

	/* Check for Saved Notify */

	if ( VFG->bail )
	{
		STV_ERROR( "stv_viewer_recv_FieldAcks", stvStatusBail,
			"Notify Found for View Field Group - Bail..." );

		return( stvStatusBail );
	}

	/* Handle Field Ack Replies */

	VFG->locked = stvAppUnlocked;

	VFG->reply_cnt = 0;

	VFG->ack_cnt = 0;

	do
	{
		if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MSG ) )
		{
			printf( "Blocking for FieldAck Reply Messages.\n" );
			fflush( stdout );
		}

		if ( (cc = stv_handleMessages( stvFieldAck,
				stvAny, V->msgtag, STV_VIEWER_CTX,
				stvMsgBlock )) <= 0 )
		{
			STV_ERROR( "stv_viewer_recv_FieldAcks", cc,
					"Receiving FieldAck Reply From Application." );

			return( cc );
		}

		/* Check for Saved Notify */

		if ( VFG->bail )
		{
			STV_ERROR( "stv_viewer_recv_FieldAcks", stvStatusBail,
				"Notify Found for View Field Group - Bail..." );

			return( stvStatusBail );
		}
	}
	while ( VFG->reply_cnt < VFG->ntids );

	/* Allocate New Tids List */

	if ( VFG->ack_cnt )
	{
		tids = (int *) MALLOC( (STV_SIZE_T) VFG->ack_cnt
				* sizeof(int) );
		STV_MEMCHECK_return( tids, "Static Field Tids List" );

		for ( i=0 ; i < VFG->ack_cnt ; i++ )
			tids[i] = -1;

		/* Extract Ack-ed Tasks from List... */

		for ( i=0, j=0 ; i < VFG->ntids ; i++ )
		{
			VT = stv_get_view_task_tid( VFG->tid_list[i],
				V->vtasks, V->nvtasks );
		
			if ( VT->ack )
				tids[ j++ ] = VFG->tid_list[i];

			VFG->tid_list[i] = -1;
		}

		cc = stvStatusOk;
	}

	else
	{
		STV_ERROR( "stv_viewer_recv_FieldAcks", stvStatusBadParam,
			"No Positive Acks from Appl - No Fields or Unknown?" );

		tids = (int *) NULL;

		cc = stvStatusBadParam;
	}

	/* Replace VFG Tids List */

	FREE( VFG->tid_list );

	VFG->tid_list = tids;

	VFG->ntids = VFG->ack_cnt;

	/* Done */

	return( cc );
}


/* Receive and Process One FieldAck Reply */

int
stv_protoFieldAck( msg )
STV_BUFPTR msg;
{
	STV_VIEW_FIELD_GROUP VFG;

	STV_VIEW_TASK VT;

	STV_VIEWER V;

	int tmp;
	int vid;
	int cc;

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MSG ) )
	{
		printf( "Entry stv_protoFieldAck().\n" );
		fflush( stdout );
	}

	if ( msg == (STV_BUFPTR) NULL )
		return( stvStatusOk );

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MSG ) )
	{
		printf( "Message Received From Task ID=[ 0x%x / %d ]\n",
				msg->src, msg->src );
		fflush( stdout );
	}

	/* Unpack Viewer ID */

	stv_unpack_int( msg->bufid, &vid, 1, 1 );

	if ( (V = stv_find_viewer( vid )) == NULL )
	{
		sprintf( STVERRTXT, "Viewer Not Found for ID=%d", vid );

		STV_ERROR( "stv_protoFieldAck", stvStatusBadParam, STVERRTXT );

		return( stvStatusBadParam );
	}

	else if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "Viewer ID=%d Found.\n", vid );
		fflush( stdout );
	}

	/* Find View Field Group */

	if ( (VFG = stv_get_view_field_group_id( V->vfglist, V->nvfgs,
			msg->fgid )) == NULL )
	{
		sprintf( STVERRTXT, "View Field Group Not Found for fgid=%d",
				msg->fgid );

		STV_ERROR( "stv_protoFieldAck", stvStatusBadParam, STVERRTXT );

		return( stvStatusBadParam );
	}

	/* Is Task Legit? */

	if ( !stv_check_view_field_group_task( VFG, msg->src ) )
	{
		sprintf( STVERRTXT,
				"Task ID=[ 0x%x / %d ] Not In View Field Group.",
				msg->src, msg->src );

		STV_ERROR( "stv_protoFieldAck", stvStatusBadParam, STVERRTXT );

		return( stvStatusBadParam );
	}

	/* Get Viewer Task Structure */

	if ( (VT = stv_get_view_task_tid( msg->src, V->vtasks, V->nvtasks ))
			== NULL )
	{
		sprintf( STVERRTXT, "Task ID=[ 0x%x / %d ] Not Found.",
				msg->src, msg->src );

		STV_ERROR( "stv_protoFieldAck", stvStatusBadParam, STVERRTXT );

		return( stvStatusBadParam );
	}

	stv_unpack_int( msg->bufid, &(VT->ack), 1, 1 );

	if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "\tField Request Ack is %d.\n", VT->ack );
		fflush( stdout );
	}

	if ( VT->ack )
		(VFG->ack_cnt)++;

	stv_unpack_int( msg->bufid, &tmp, 1, 1 );

	if ( tmp == stvAppLocked )
		VFG->locked = stvAppLocked;

	(VFG->reply_cnt)++;

	/* Done */

	return( stvStatusOk );
}


/* Send XON to Tasks */

void
stv_viewer_send_XON( VFG, V )
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
{
	int sbuf;
	int tmp;

	if ( STV_OPT_VERBOSE
			& ( STV_DEBUG_VIEWER | STV_DEBUG_TIME | STV_DEBUG_FREQ ) )
	{
		printf( "Sending XON fgid=%d Region Time = ", VFG->fgid );
		stv_dump_time_str( stdout, VFG->region_time, stvTrue );
		fflush( stdout );
	}

	stv_init_msg_buf( stvDataDefault, &sbuf );

	tmp = stvMsgOp( stvXon );
	stv_pack_int( sbuf, &tmp, 1, 1 );

	stv_pack_int( sbuf, &(VFG->fgid), 1, 1 );

	stv_mcast_msg( sbuf, VFG->tid_list, VFG->ntids,
			V->msgtag, STV_APP_CTX );

	stv_free_msg_buf( sbuf );

	if ( VFG->locked == stvAppUnlocked )
	{
		VFG->region_time += VFG->eff_freq;

		if ( STV_OPT_VERBOSE
				& ( STV_DEBUG_VIEWER | STV_DEBUG_TIME
						| STV_DEBUG_FREQ ) )
		{
			printf( "\tregion_time -> " );
			stv_dump_time_str( stdout, VFG->region_time, stvTrue );
			fflush( stdout );
		}
	}

	else
	{
		if ( STV_OPT_VERBOSE
				& ( STV_DEBUG_VIEWER | STV_DEBUG_TIME
						| STV_DEBUG_FREQ ) )
		{
			printf( "\tregion_time locked at " );
			stv_dump_time_str( stdout, VFG->region_time, stvTrue );
			fflush( stdout );
		}
	}
}


/* Send stvTimeSync Message */

void
stv_viewer_send_TimeSync( VFG, V )
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
{
	int sbuf;
	int tmp;

	/* Send Time Sync Message */

	if ( STV_OPT_VERBOSE
			& ( STV_DEBUG_VIEWER | STV_DEBUG_TIME | STV_DEBUG_FREQ ) )
	{
		printf( "Sending Time Sync Request, vid=%d fgid=%d\n",
			V->id, VFG->fgid );
		fflush( stdout );
	}

	stv_init_msg_buf( stvDataDefault, &sbuf );

	tmp = stvMsgOp( stvTimeSync );
	stv_pack_int( sbuf, &tmp, 1, 1 );

	stv_pack_int( sbuf, &(VFG->fgid), 1, 1 );

	stv_pack_int( sbuf, &(V->id), 1, 1 );

	stv_mcast_msg( sbuf, VFG->tid_list, VFG->ntids,
			V->msgtag, STV_APP_CTX );

	stv_free_msg_buf( sbuf );
}


/* Send AppLock to Tasks */

void
stv_viewer_send_AppLock( VFG, V, no_synch )
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
int no_synch;
{
	STV_VIEW_TASK VT;

	int *tids;
	int ntids;

	int locktime;
	int locked;
	int fgid;
	int sbuf;
	int tmp;
	int cc;
	int i;

	if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		if ( VFG != NULL )
			printf( "Sending AppLock fgid=%d\n", VFG->fgid );
		
		else if ( V != NULL )
			printf( "Sending AppLock to Entire Application.\n" );

		else
		{
			sprintf( STVERRTXT,
				"Error:  Missing View Field Group or Viewer" );

			STV_ERROR( "stv_viewer_send_AppLock", stvStatusBadParam,
				STVERRTXT );

			return;
		}

		fflush( stdout );
	}

	/* Determine Target Task IDs */

	if ( VFG != NULL )
	{
		fgid = VFG->fgid;

		tids = VFG->tid_list;
		ntids = VFG->ntids;

		locktime = VFG->region_time;

		locked = VFG->locked;
	}
	
	else
	{
		/* Assemble Task IDs List */

		tids = (int *) MALLOC( (STV_SIZE_T) V->nvtasks * sizeof(int) );
		STV_MEMCHECK( tids, "AppLock Tids List", return );

		VT = V->vtasks;

		i = 0;

		while ( VT != NULL )
		{
			tids[i++] = VT->tid;

			VT = VT->next;
		}

		ntids = V->nvtasks;

		fgid = -1;

		/* Loosely Synchronize */

		if ( !no_synch )
		{
			/* Set Up Quickie Notify */

			cc = stv_notify_event( stvTaskExit, V->msgtag,
				STV_VIEWER_CTX, ntids, tids );

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER | STV_DEBUG_NOTIFY ) )
			{
				printf( "Set Up Quickie Notify for Tasks, cc=%d\n",
					cc );
				fflush( stdout );
			}

			/* Send Sync Time Request */

			STV_SYNC_VFG = stv_create_view_field_group();

			STV_SYNC_VFG->fgid = STV_NEXT_FGID++;

			STV_SYNC_VFG->tid_list = tids;
			STV_SYNC_VFG->ntids = ntids;

			STV_SYNC_VFG->flag = stvFieldRqstDefault;

			STV_SYNC_VFG->bail = stvFalse;

			stv_viewer_send_TimeSync( STV_SYNC_VFG, V );

			/* Receive Back Field Times */

			cc = stv_viewer_recv_FieldTime( STV_SYNC_VFG, V );

			locktime = STV_SYNC_VFG->timemax;

			locked = STV_SYNC_VFG->locked;

			STV_SYNC_VFG->tid_list = (int *) NULL;

			stv_free_view_field_group( &STV_SYNC_VFG );

			if ( cc != stvStatusOk )
			{
				STV_ERROR( "stv_viewer_TimeSync", cc, "Cannot Lock\n" );

				return;
			}

			if ( locked == stvAppUnlocked )
				locktime += 2;

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER | STV_DEBUG_TIME
							| STV_DEBUG_FREQ ) )
			{
				printf( "All FieldTimes Recvd, Lock Time is %d.\n",
					locktime );
				fflush( stdout );
			}
		}
	}

	if ( locked == stvAppUnlocked )
	{
		/* Send Application Lock Message */

		stv_init_msg_buf( stvDataDefault, &sbuf );

		tmp = stvMsgOp( stvAppLock );
		stv_pack_int( sbuf, &tmp, 1, 1 );

		stv_pack_int( sbuf, &fgid, 1, 1 );

		stv_pack_int( sbuf, &(V->mytid), 1, 1 );

		stv_pack_int( sbuf, &no_synch, 1, 1 );

		if ( !no_synch )
			stv_pack_int( sbuf, &locktime, 1, 1 );

		stv_mcast_msg( sbuf, tids, ntids, V->msgtag, STV_APP_CTX );
	
		stv_free_msg_buf( sbuf );
	}

	if ( VFG == NULL )
		FREE( tids );
}


/* Send AppUnlock to Tasks */

void
stv_viewer_send_AppUnlock( VFG, V )
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
{
	STV_VIEW_TASK VT;

	int *tids;
	int ntids;

	int fgid;
	int sbuf;
	int tmp;
	int i;

	if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		if ( VFG != NULL )
			printf( "Sending AppUnlock fgid=%d\n", VFG->fgid );
		
		else if ( V != NULL )
			printf( "Sending AppUnlock to Entire Application.\n" );

		else
		{
			sprintf( STVERRTXT,
				"Error:  Missing View Field Group or Viewer" );

			STV_ERROR( "stv_viewer_send_AppUnlock", stvStatusBadParam,
				STVERRTXT );

			return;
		}

		fflush( stdout );
	}

	/* Determine Target Task IDs */

	if ( VFG != NULL )
	{
		fgid = VFG->fgid;

		tids = VFG->tid_list;
		ntids = VFG->ntids;
	}
	
	else
	{
		/* Assemble Task IDs List */

		tids = (int *) MALLOC( (STV_SIZE_T) V->nvtasks * sizeof(int) );
		STV_MEMCHECK( tids, "AppLock Tids List", return );

		VT = V->vtasks;

		i = 0;

		while ( VT != NULL )
		{
			tids[i++] = VT->tid;

			VT = VT->next;
		}

		ntids = V->nvtasks;

		fgid = -1;
	}

	stv_init_msg_buf( stvDataDefault, &sbuf );

	tmp = stvMsgOp( stvAppUnlock );
	stv_pack_int( sbuf, &tmp, 1, 1 );

	stv_pack_int( sbuf, &fgid, 1, 1 );

	stv_pack_int( sbuf, &(V->mytid), 1, 1 );

	stv_mcast_msg( sbuf, tids, ntids, V->msgtag, STV_APP_CTX );
	
	stv_free_msg_buf( sbuf );

	if ( VFG == NULL )
		FREE( tids );
}


/* Send stvVisRegion Message */

int
stv_viewer_send_VisRegion( VFG, V )
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
{
	int sbuf;
	int tmp;

	/* Verify Not Static View Field Group */

	if ( VFG->flag & stvFieldRqstStatic )
	{
		STV_ERROR( "stv_viewer_send_VisRegion", stvStatusBadParam,
			"Attempt to Change Static View Field Group - Ignored." );

		return( stvStatusBadParam );
	}

	/* Get New Region ID */

	VFG->region_id = STV_NEXT_REGION_ID++;

	if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "Sending VisRegion fgid=%d region_id=%d\n",
			VFG->fgid, VFG->region_id );
		fflush( stdout );
	}

	/* Send Update Message */

	stv_init_msg_buf( stvDataDefault, &sbuf );

	tmp = stvMsgOp( stvVisRegion );
	stv_pack_int( sbuf, &tmp, 1, 1 );

	stv_pack_int( sbuf, &(VFG->fgid), 1, 1 );

	stv_pack_int( sbuf, &(VFG->region_id), 1, 1 );

	stv_pack_int( sbuf, VFG->visregion,
		STV_REGION_SIZE( STV_MAX_DIM ), 1 );

	stv_pack_int( sbuf, VFG->viscell,
		STV_CELL_SIZE( STV_MAX_DIM ), 1 );

	stv_mcast_msg( sbuf, VFG->tid_list, VFG->ntids,
			V->msgtag, STV_APP_CTX );

	stv_free_msg_buf( sbuf );

	/* Successful Submission */

	return( stvStatusOk );
}


/* Send stvVisFrequency Message */

void
stv_viewer_send_VisFrequency( VFG, V )
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
{
	int sbuf;
	int tmp;

	/* Get New Region ID */

	VFG->region_id = STV_NEXT_REGION_ID++;

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_FREQ ) )
	{
		printf( "%s: fgid=%d, region_id=%d, eff_freq=%d, freq=",
			"Sending VisFrequency",
			VFG->fgid, VFG->region_id, VFG->eff_freq );
		stv_dump_time_str( stdout, VFG->freq, stvFalse );
		printf( ", region_time=%d\n", VFG->region_time );
		fflush( stdout );
	}

	/* Send Update Message */

	stv_init_msg_buf( stvDataDefault, &sbuf );

	tmp = stvMsgOp( stvVisFrequency );
	stv_pack_int( sbuf, &tmp, 1, 1 );

	stv_pack_int( sbuf, &(VFG->fgid), 1, 1 );

	stv_pack_int( sbuf, &(VFG->region_id), 1, 1 );

	stv_pack_int( sbuf, &(VFG->eff_freq), 1, 1 );
	stv_pack_int( sbuf, &(VFG->freq), 1, 1 );

	stv_pack_int( sbuf, &(VFG->region_time), 1, 1 );

	stv_mcast_msg( sbuf, VFG->tid_list, VFG->ntids,
			V->msgtag, STV_APP_CTX );

	stv_free_msg_buf( sbuf );
}


/* Send Set Option Directives Message */

void
stv_viewer_send_Setopt( VFG, V, what, option, value )
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
int what;
int option;
int value;
{
	STV_VIEW_TASK VT;

	int *tids;
	int ntids;

	int sbuf;
	int tmp;
	int i;

	if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "Sending Setopt to Appl: " );
		printf( "what=%s(%d) ", stv_option_str( what ), what );
		printf( "option=%s(%d) ", stv_option_str( option ), option );
		printf( "value=%d\n", value );
		fflush( stdout );
	}

	/* Determine Target Task IDs */

	if ( VFG != NULL )
	{
		tids = VFG->tid_list;
		ntids = VFG->ntids;
	}
	
	else
	{
		/* Assemble Task IDs List */

		tids = (int *) MALLOC( (STV_SIZE_T) V->nvtasks * sizeof(int) );
		STV_MEMCHECK( tids, "Setopt Tids List", return );

		VT = V->vtasks;

		i = 0;

		while ( VT != NULL )
		{
			tids[i++] = VT->tid;

			VT = VT->next;
		}

		ntids = V->nvtasks;
	}

	/* Send Update Message */

	stv_init_msg_buf( stvDataDefault, &sbuf );

	tmp = stvMsgOp( stvSetopt );
	stv_pack_int( sbuf, &tmp, 1, 1 );

	tmp = -1; /* dummy fgid */
	stv_pack_int( sbuf, &tmp, 1, 1 );

	stv_pack_int( sbuf, &what, 1, 1 );

	stv_pack_int( sbuf, &option, 1, 1 );

	stv_pack_int( sbuf, &value, 1, 1 );

	stv_mcast_msg( sbuf, tids, ntids, V->msgtag, STV_APP_CTX );

	stv_free_msg_buf( sbuf );
}


/* Send FieldHaltGroup Message */

void
stv_viewer_send_FieldHaltGroup( VFG, V )
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
{
	int sbuf;
	int tmp;

	if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "Sending FieldHaltGroup fgid=%d\n", VFG->fgid );
		fflush( stdout );
	}

	stv_init_msg_buf( stvDataDefault, &sbuf );

	tmp = stvMsgOp( stvFieldHaltGroup );
	stv_pack_int( sbuf, &tmp, 1, 1 );

	stv_pack_int( sbuf, &(VFG->fgid), 1, 1 );

	stv_mcast_msg( sbuf, VFG->tid_list, VFG->ntids,
			V->msgtag, STV_APP_CTX );

	stv_free_msg_buf( sbuf );

	/* Make Sure Toast */

	VFG->region_id = -1;
}


/* Send FieldHaltGroup Message */

void
stv_viewer_send_FieldHaltAll( VFG, V )
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
{
	int sbuf;
	int tmp;

	if ( STV_OPT_VERBOSE & STV_DEBUG_VIEWER )
	{
		printf( "Sending FieldHaltAll fgid=%d\n", VFG->fgid );
		fflush( stdout );
	}

	stv_init_msg_buf( stvDataDefault, &sbuf );

	tmp = stvMsgOp( stvFieldHaltAll );
	stv_pack_int( sbuf, &tmp, 1, 1 );

	/* dummy fgid, for message header consistency */
	tmp = -1;
	stv_pack_int( sbuf, &tmp, 1, 1 );

	stv_mcast_msg( sbuf, VFG->tid_list, VFG->ntids,
			V->msgtag, STV_APP_CTX );

	stv_free_msg_buf( sbuf );

	/* Make Sure Toast */

	VFG->region_id = -1;
}


/* Handle Notify for Viewer Library */

int
stv_protoViewerNotify( msg )
STV_BUFPTR msg;
{
	STV_VIEWER V;

	int found;

	if ( msg == NULL )
		return( stvStatusOk );

	sprintf( STVERRTXT,
		"Received Task Exit Notify for Task ID=[ 0x%x / %d ]",
		msg->src, msg->src );

	STV_ERROR( "stv_protoViewerNotify", stvStatusBail, STVERRTXT );

	V = STVVIEWERS;

	found = 0;

	while ( V != NULL )
	{
		if ( stv_check_view_field_group_notify( V, msg->src ) )
			found++;

		V = V->next;
	}

	if ( !found )
	{
		STV_ERROR( "stv_protoViewerNotify", stvStatusOk,
			"Task Not in Any View Field Groups - Ignore" );
	}

	return( stvStatusOk );
}


/*** Region Routines ***/


/* Check Vis Region Against Decomp & Cell Sizes */

int
stv_check_region( F, R, C )
STV_FIELDPTR F;
STV_REGION R;
STV_CELL C;
{
	int *glb, *gub;

	int changed;
	int diff;
	int dim;
	int i;

	changed = stvFalse;

	/* Get glb, gub, dim */

	if ( F->decomp != NULL )
	{
		glb = F->decomp->glb;
		gub = F->decomp->gub;

		dim = F->decomp->dataDim;
	}

	else if ( F->particle != NULL )
	{
		glb = F->particle->glb;
		gub = F->particle->gub;

		dim = F->particle->dataDim;
	}

	else
	{
		sprintf( STVERRTXT,
			"Warning: No Decomp or Particle for Field \"%s\"",
			F->name );

		STV_ERROR( "stv_check_region", stvStatusBadParam, STVERRTXT );

		return( changed );
	}

	/* Check Against Decomposition Global Bounds */

	for ( i=0 ; i < dim ; i++ )
	{
		if ( STV_REGION_MIN( R, i ) < glb[i] )
		{
			STV_REGION_MIN( R, i ) = glb[i];

			changed = stvTrue;
		}

		if ( STV_REGION_MIN( R, i ) > gub[i] )
		{
			STV_REGION_MIN( R, i ) = gub[i];

			changed = stvTrue;
		}

		if ( STV_REGION_MAX( R, i ) > gub[i] )
		{
			STV_REGION_MAX( R, i ) = gub[i];

			changed = stvTrue;
		}

		if ( STV_REGION_MAX( R, i ) < glb[i] )
		{
			STV_REGION_MAX( R, i ) = glb[i];

			changed = stvTrue;
		}
	}

	/* Check Divisibility with Cell Size */

	for ( i=0 ; i < dim ; i++ )
	{
		/* First, Make Sure Cell Size is Positive...  D-Oh! */

		if ( STV_CELL_OF_DIM( C, i ) <= 0 )
		{
			STV_CELL_OF_DIM( C, i ) = 1;

			changed = stvTrue;
		}

		/* Leave Min Where It Is, Use Min as Offset for Divisibility */

		/* Bring ( Max - Min ) Down to Last Divisible By Cell */

		diff = STV_REGION_MAX( R, i ) - STV_REGION_MIN( R, i );

		while ( !STV_DIVISIBLE( diff, STV_CELL_OF_DIM( C, i ) )
			&& diff > 0 )
		{
			STV_REGION_MAX( R, i )--;

			changed = stvTrue;

			diff = STV_REGION_MAX( R, i ) - STV_REGION_MIN( R, i );
		}
	}

	/* Check for Valid Region */

	for ( i=0 ; i < dim ; i++ )
	{
		/* Bogus Region, Adjust Bounds to Fix */

		if ( STV_REGION_MIN( R, i ) > STV_REGION_MAX( R, i ) )
		{
			STV_REGION_MAX( R, i ) = STV_REGION_MIN( R, i );

			changed = stvTrue;
		}
	}

	return( changed );
}


/* Allocate Field Frame Array Declaration */

int
stv_allocate_Field_arrayDecl( VFG, index )
STV_VIEW_FIELD_GROUP VFG;
int index;
{
	STV_FIELDPTR F;

	int *glb, *gub, dim;

	int min, max;
	int step;
	int i;

	/* Grab Field */

	F = (VFG->view_field_list[index])->field;

	/* Grab glb, gub, dim */

	if ( F->decomp != NULL )
	{
		glb = F->decomp->glb;
		gub = F->decomp->gub;

		dim = F->decomp->dataDim;
	}

	else if ( F->particle != NULL )
	{
		glb = F->particle->glb;
		gub = F->particle->gub;

		dim = F->particle->dataDim;
	}

	else
	{
		sprintf( STVERRTXT,
			"Warning: No Decomp / Particle for Field \"%s\"",
			F->name );

		STV_ERROR( "stv_allocate_Field_arrayDecl", stvStatusBadParam,
			STVERRTXT );

		return( stvStatusBadParam );
	}

	/* Set Up Array Decl */

	if ( VFG->field_arrayDecls[index] != NULL )
	{
		if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MEM ) )
		{
			printf( "Freeing Previous Array Decl for Field \"%s\"\n",
				F->name );
			fflush( stdout );
		}

		FREE( VFG->field_arrayDecls[index] );
	}

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MEM ) )
	{
		printf( "Allocating Array Decl for Field \"%s\" dim=%d\n",
			F->name, dim );
		fflush( stdout );
	}

	VFG->field_arrayDecls[index] = (int *) MALLOC( (STV_SIZE_T) dim	
		* sizeof(int) );
	STV_MEMCHECK_return( VFG->field_arrayDecls[index],
		"Field Array Decl Vector" );

	for ( i=0 ; i < dim ; i++ )
	{
		if ( STV_OPT_FULL_UNPACK )
		{
			min = glb[i];
			max = gub[i];

			step = 1;
		}

		else
		{
			min = STV_REGION_MIN( VFG->visregion, i );
			max = STV_REGION_MAX( VFG->visregion, i );

			min = min < glb[i] ? glb[i] : min;
			max = max > gub[i] ? gub[i] : max;

			step = STV_CELL_OF_DIM( VFG->viscell, i );
		}

		VFG->field_arrayDecls[index][i] = ( ( max - min ) / step ) + 1;

		if ( VFG->field_arrayDecls[index][i] < 0 )
			VFG->field_arrayDecls[index][i] = 0;

		if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MEM ) )
		{
			printf( "\tarrayDecls[%d] = %d (%d)\n",
				i, VFG->field_arrayDecls[index][i],
				( ( max - min ) / step ) + 1 );
			fflush( stdout );
		}
	}

	return( stvStatusOk );
}


/* Allocate Field Frame Storage */

int
stv_allocate_Field_Data( VFG, index, type )
STV_VIEW_FIELD_GROUP VFG;
int index;
int type;
{
	STV_DECOMPPTR D;

	STV_FIELDPTR F;

	STV_VALUE dst;

	double init;

	int size;
	int cc;
	int i;

	/* Grab Field */

	F = (VFG->view_field_list[index])->field;

	/* Grab Decomp */

	D = F->decomp;

	if ( D == NULL )
	{
		STV_ERROR( "stv_allocate_Field_Data", stvStatusBadParam,
			"Not a Decomposition Field, Particle?" );
		
		return( stvStatusBadParam );
	}

	/* Accumulate Array Size */

	size = 1;

	for ( i=0 ; i < D->dataDim ; i++ )
		size *= VFG->field_arrayDecls[index][i];

	/* Allocate Array */

	if ( VFG->field_values[index] != NULL )
	{
		if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MEM ) )
		{
			printf( "Freeing Previous Data Storage for Field \"%s\"\n",
				F->name );
			fflush( stdout );
		}

		FREE( VFG->field_values[index] );
	}

	if ( size > 0 )
	{
		if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MEM ) )
		{
			printf(
				"Allocating Data Storage for Field \"%s\" size=%d\n",
				F->name, size );
			fflush( stdout );
		}

		VFG->field_values[index] = (STV_VALUE) MALLOC(
			(STV_SIZE_T) size * STV_SIZEOF( type ) );
		STV_MEMCHECK_return( VFG->field_values[index],
			"Field Frame Data" );

		/* Initialize Array Space */

		init = -1.0;

		dst = VFG->field_values[index];

		for ( i=0 ; i < size ; i++ )
		{
			/* Convert Init Value & Copy */

			if ( (cc = stv_copy_converted_data( (STV_VALUE) &init,
				stvDouble, dst, type )) != stvStatusOk )
			{
				STV_ERROR( "stv_allocate_Field_Data", cc,
					"Converted Type Copy Failed." );

				return( cc );
			}

			/* Advance Ptr */

			dst = stv_get_typed_ptr_offset( dst, 1, type );

			if ( dst == NULL )
				return( stvStatusBadParam );
		}
	}

	return( stvStatusOk );
}


/* Free Field Frame Storage */

void
stv_free_field_array( VFG, index )
STV_VIEW_FIELD_GROUP VFG;
int index;
{
	STV_FIELDPTR F;

	STV_VALUE ptr;

	double clear;

	int size;
	int dim;
	int cc;
	int i;

	if ( VFG == NULL )
		return;

	F = (VFG->view_field_list[index])->field;

	if ( F->decomp != NULL )
		dim = F->decomp->dataDim;
	
	else if ( F->particle != NULL )
		dim = F->particle->dataDim;
	
	else
	{
		sprintf( STVERRTXT,
			"Warning: No Decomp / Particle for Field \"%s\"",
			F->name );

		STV_ERROR( "stv_free_field_array", stvStatusBadParam,
			STVERRTXT );

		return;
	}

	size = 1;

	if ( VFG->field_arrayDecls[index] != NULL )
	{
		for ( i=0 ; i < dim ; i++ )
		{
			size *= VFG->field_arrayDecls[index][i];

			VFG->field_arrayDecls[index][i] = -1;
		}

		if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MEM ) )
		{
			printf( "Freeing Array Decls for Field \"%s\"\n",
				F->name );
			fflush( stdout );
		}

		FREE( VFG->field_arrayDecls[index] );

		VFG->field_arrayDecls[index] = (int *) NULL;
	}

	if ( VFG->field_values[index] != NULL )
	{
		if ( size != 1 )
		{
			/* Clear Array Space */

			clear = -2.0;

			ptr = VFG->field_values[index];

			for ( i=0 ; i < size ; i++ )
			{
				/* Convert Clear Value & Copy */

				if ( (cc = stv_copy_converted_data( (STV_VALUE) &clear,
					stvDouble, ptr, F->type )) != stvStatusOk )
				{
					STV_ERROR( "stv_allocate_Field_Data", cc,
						"Converted Type Copy Failed." );

					return;
				}

				/* Advance Ptr */

				ptr = stv_get_typed_ptr_offset( ptr, 1, F->type );

				if ( ptr == NULL )
					return;
			}
		}

		if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MEM ) )
		{
			printf( "Freeing Data Storage for Field \"%s\"\n",
				F->name );
			fflush( stdout );
		}

		FREE( VFG->field_values[index] );

		VFG->field_values[index] = (STV_VALUE) NULL;
	}
}


/* Reset View Field Group Particle Lists */

void
stv_reset_Field_Particles( VFG )
STV_VIEW_FIELD_GROUP VFG;
{
	STV_VIEW_PARTICLE VPT;

	/* Verify Valid View Field Group */

	if ( VFG == NULL )
		return;

	/* Free Any Particle Data */

	VPT = VFG->particle_list;

	while ( VPT != NULL )
	{
		stv_free_view_particle_fields( VPT );

		VPT = VPT->next;
	}

	/* Free Empty List Structure */

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_VIEWER | STV_DEBUG_MEM ) )
	{
		printf( "Freeing Particle List Structure for fgid=%d\n",
			VFG->fgid );
		fflush( stdout );
	}

	stv_free_view_particle_list( &(VFG->particle_list) );
}

