
/* $Id: stv_ulib.c,v 3.109 2000/09/25 18:27:14 kohl Exp $ */

/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* File: stv_ulib.c - Higher level user functions for CUMULVS. */


/* Allows a user program to initialize/define fields and parameters.
* A single call sends new data to the Visualization Front End and
* then returns control to the user.
*
* All the lower-level synch details are handled by the primitives.
*
* User Calls defined: (stv_<xxx>)
*
*	decompDefine()			- (re-)define a decomposition
*       (see also stv_decompAddExplicitBounds() - multi-explicit)
*	fieldDefine()			- name and (re-)define a user field
*	paramDefine()			- name and define a steering parameter
*	vParamDefine()			- name and define a vectored steering param
*	particleDefine()		- name and define a particle 
*	particleFieldDefine()	- name and (re)define a field in a particle
*	sendReadyData()			- send ready fields and params to Vis FE
*	sendToFE()				- send fields and params to Vis FE (orig)
*	isParamChanged			- flag if a steering parameter has changed
*
* Calls for Internal Use: stv_<xxxx>
*
*/


/* C++ Compatibility */

#ifdef __cplusplus
extern "C" {
#endif


/* System Header Files */

#include <stdio.h>
#include <string.h>


/* STV Header Files */

#include "stv.h"
#include "stvmac.h"
#include "stv_glob.h"


/* Bits to Indicate Different ID Types */
/* -- used for parameter checking */

#define DECOMPBIT     ( 1 << 30 )
#define FIELDBIT      ( 1 << 29 )
#define PARAMBIT      ( 1 << 28 )
#define PARTICLEBIT   ( 1 << 27 )


/* List of Attached Viewers to a Particular Field */

struct nlist
{
	int vtid;
	struct nlist *prev, *next;
};

typedef struct nlist NLIST, *NLISTPTR;


/* Private Information Lists' Static Allocation */

static STV_FE_VIEWER VIEWERS;
static STV_DECOMP DECOMPS;
static NLIST VTIDS;


/* Private Information Lists' Size Globals */

static int nVISFIELDS = 0;
static int nVISPARAMS = 0;
static int nDECOMPS = 0;
static int nVIEWERS = 0;


/* Other Private User Library Globals */

static int STVCOORD = -1;


/* Private Error Message Text Storage */

static char errtxt[512];


/* Routines */


/*  -------------  stv_init() -------------- */
/* This initializes the fields and viewers
 *   o  Every task joins the STVSHADOWNAME group.
 *   o  instance 0 of shadow group inserts the msgtag into the name
 *          server
 *   o  instance 0 of the STVSHADOWNAME group joins the STVAPPNAME group
 *   o  Once the msgtag is inserted into name server AND STVAPPNAME
 *          group has a member 0, viewers may attach to the application
 */

int
stv_init( appname, msgtag, ntasks, nodeid )
char *appname;
int msgtag;
int ntasks;
int nodeid;
{
	static int first = stvTrue;

	STV_FIELDPTR f, nf;
	STV_PARAMPTR p, np;
	char *debug_env;
	int cc;

	/* Before You Begin, Check for Debug Flags in Env */

	if ( (debug_env = getenv( "STV_DEBUG" )) != NULL )
	{
		cc = atoi( debug_env );

		if ( cc )
		{
			printf( "Setting STV_DEBUG from environment: " );
			stv_dump_debug_flag( cc, stvTrue );

			stv_setopt( stvOptNull, stvOptVerbose, cc );
		}
	}

	/* And So It Begins... */

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
	{
		if ( first )
		{
			stv_banner();

			first = stvFalse;
		}

		printf( "stv_init:  ntasks=%d, msgtag=%d, nodeid=%d\n",
				ntasks, msgtag, nodeid );
	}

	STV_FULLGROUP = 0;

	STV_DEACTIVATED = stvFalse;

	if ( msgtag >= 0 )  /* will (re)-init */
	{
		/* Save Application Info */

		strcpy( STVAPPNAME, stvPrefix );
		strcpy( STVSHADOWNAME, STVAPPNAME );
		strcat( STVSHADOWNAME, "SHDW_" );

		if ( appname == (char *) NULL || strlen( appname ) == 0 )
		{
			strcpy( STVUSERAPPNAME, stvDefaultClient );
			strcat( STVAPPNAME, stvDefaultClient );
			strcat( STVSHADOWNAME, stvDefaultClient );
		}
		else
		{
			strcpy( STVUSERAPPNAME, appname );
			strcat( STVAPPNAME, appname );
			strcat( STVSHADOWNAME, appname );
		}

		STV_USERTAG = msgtag;

		STV_NTASKS = ntasks;

		STV_MYNODEID = nodeid;

		/* Free the Field & Param Structures, If Any */

		f = &STVFIELDS; nf = f->next;
		while ( nf )
		{
			f = nf;
			nf = nf->next;
			LISTDELETE( f, prev, next );
			FREELIST( f->vlist, prev, next );
			stv_free_field( &f );
		}
		nSTVFIELDS = 0;

		p = &STVPARAMS; np = p->next;
		while ( np )
		{
			p = np;
			np = np->next;
			LISTDELETE( p, prev, next );
			stv_free_param( &p );
		}
		nSTVPARAMS = 0;

		if ( CPDATA == (CPD_DATAPTR) NULL )
		{
			CPDATA = (CPD_DATAPTR) CALLOC( sizeof( CPD_DATA ),
					"stv_init" );
			STV_NO_MEM_return( CPDATA, "stv_init" );

			CPDATA->param = (STV_PARAMPTR) NULL;
			CPDATA->field = (STV_FIELDPTR) NULL;
			CPDATA->next = (CPD_DATAPTR) NULL;
		}

		/* Now start the initialization */
		/*   - don't pass back stvStatusBadParam - pretend it's O.K. */
		/*     (BadParam comes from context creation / check race) */

		if ( (cc = stv_init_system( stvVerbose )) != stvStatusOk )
			return( cc != stvStatusBadParam ? cc : stvStatusOk );

	}

	return( STV_USERTAG );
}


/*  -------------  stv_init_system() -------------- */
/*  Routine to connect to message-passing system, insert
 *  application context into persistent storage, set up
 *  groups, etc.  Extracted into separate routine to allow
 *  delayed execution, in case the message-passing system
 *  isn't up & running yet...
 */

int
stv_init_system( verbose )
int verbose;
{
	STV_FE_VIEWERPTR vs, nvs;

	int c, cc, bufid;
	int systag = -1;
	int index;

	/* Try to Connect to Message Passing System */

	stv_myid( &STV_MYID );

	if ( STV_MYID < 0 )
	{
		if ( verbose )
		{
			sprintf( STVERRTXT, "Error %s from %s %s - %s.",
				"Acquiring Task ID", "Message Passing Substrate",
				STV_MP_NAME, "CUMULVS Functionality Deactivated" );
			STV_ERROR( "stv_init", stvStatusSysFail, STVERRTXT );
		}

		STV_DEACTIVATED = stvTrue;

		return( stvStatusSysFail );
	}

	/* free the viewer structs, if any */

	vs = &VIEWERS; nvs = vs->next;
	while ( nvs )
	{
		vs = nvs;
		nvs = nvs->next;
		if ( STV_GCOORD == 0 )
		{
			stv_init_msg_buf( stvDataDefault, &bufid );
			c = stvMsgOp( stvReconfigure );
			stv_pack_int( bufid, &c, 1, 1 );
			stv_pack_int( bufid, &(vs->vid), 1, 1 );
			stv_send_msg( bufid, vs->vtid, STV_USERTAG,
					STV_VIEWER_CTX );
			stv_free_msg_buf( bufid );
		}
		LISTDELETE( vs, prev, next );
		stv_free_fe_viewer( &vs );
	}
	nVIEWERS = 0;

	/* I haven't already joined the shadowgroup */

	if ( STV_GCOORD < 0 )
	{
		if ( !stv_join_group( STVSHADOWNAME, ntasks, &STV_GCOORD ) )
		{
			sprintf( STVERRTXT, "Joining Shadow Group %s",
					STVSHADOWNAME );
			STV_ERROR( "stv_init", stvStatusSysFail, STVERRTXT );
			return( stvStatusSysFail );
		}
	}

	/* Verify System Contexts */

	if ( (cc = stv_verify_system_contexts(
			( STV_GCOORD == 0 ) ? stvTrue : stvFalse,
			stvFalse )) < 0 )
		return( cc );

	stv_define_message_handlers();

	/* check context & insert message tag into the name server */

	if ( STV_GCOORD == 0 )
	{
		/* Insert App Tag */

		stv_name_insert( STVAPPNAME, 0, &STV_USERTAG, &cc );

		if ( cc < 0 )  /* Delete it. Left over from previous run? */
		{
			sprintf( STVERRTXT,
					"Application Tag Insert Failed (Leftover Tag?), %s",
					"Forcing Re-Insert" );
			STV_ERROR( "stv_init", stvStatusBadParam, STVERRTXT );

			stv_name_delete( STVAPPNAME, 0, &cc );
			stv_name_insert( STVAPPNAME, 0, &STV_USERTAG, &cc );

			if ( cc < 0 )
			{
				sprintf( STVERRTXT,
						"Application Tag Re-Insert Failed - Bailing." );
				STV_ERROR( "stv_init", stvStatusSysFail, STVERRTXT );
				return( stvStatusSysFail );
			}
		}

		if ( cc == 0 )
		{
			systag = -1;
			cc = stv_name_lookup( STVAPPNAME, 0, &systag, &index );
			if ( !cc || index < 0 || systag != STV_USERTAG )
			{
				sprintf( STVERRTXT,
						"%s Inserted Tag %d, Read Back %d (cc=%d)",
						STVAPPNAME, STV_USERTAG, systag, cc );
				STV_ERROR( "stv_init", stvStatusBadParam,
						STVERRTXT );
				return( stvStatusBadParam );
			}
		}

		/* I haven't joined the STVAPPNAME group */

		if ( STVCOORD < 0 )
		{
			if ( !stv_join_group( STVAPPNAME, ntasks, &STVCOORD ) )
			{
				sprintf( STVERRTXT, "Joining App Name Group %s",
						STVAPPNAME );
				STV_ERROR( "stv_init", stvStatusSysFail, STVERRTXT );
				return( stvStatusSysFail );
			}
		}

		if ( STVCOORD != STV_GCOORD )
		{
			sprintf( STVERRTXT,
					"%s Shadow Group Instance (%d) != %s Instance (%d)",
					STVSHADOWNAME, STV_GCOORD,
					STVAPPNAME, STVCOORD );
			STV_ERROR( "stv_init", stvStatusBadParam, STVERRTXT );
			return( stvStatusBadParam );
		}
	}

	return( stvStatusOk );
}


int
stv_define_message_handlers()
{
	/* Add Message Handlers */

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoInit0, stvInit0,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoFieldRqst, stvFieldRqst,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoFieldTstart, stvFieldTstart,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoVisRegion, stvVisRegion,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoVisFrequency, stvVisFrequency,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoNewParams, stvNewParams,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoTimeSync, stvTimeSync,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoAppLock, stvAppLock,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoAppUnlock, stvAppUnlock,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoSetopt, stvSetopt,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoXon, stvXon,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoFieldHaltGroup, stvFieldHaltGroup,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoFieldHaltAll, stvFieldHaltAll,
			stvAny, STV_USERTAG, STV_APP_CTX );

	/* User Library Notify Handlers */

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoNotify, stvNotify,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVNOTIFYHANDLERS, &nSTVNOTIFYHANDLERS,
			stv_protoUlibNotify, stvFieldHaltAll,
			stvAny, STV_USERTAG, STV_APP_CTX );

	/* Add Message Pass-Thrus */

	stv_add_msg_pass_thru( stvNotify );

	stv_add_msg_pass_thru( stvFieldHaltGroup );
	stv_add_msg_pass_thru( stvFieldHaltAll );

	return( stvStatusOk );
}


/*  -------------  stv_reconfigure() -------------- */
/*  Call this routine when the application has reconfigured itself.
This lets the viewers know that the application is redistributing
fields and the viewer should re-negotiate connections

	XXX NOT USED YET. Just a dummy routine.

*/

int
stv_reconfigure()
{
/* Things to take care of on a reconfigure  ---
	1) For STV_GCOORD == 0 (if it still exists)
		- leave the STVAPPNAME group (if 0 died, then group server
			already handles this)
	2) For all members set the STV_USERTAG = -1;
	3) set a reconfigure flag so that everyone can do an stv_init
	4) Free all viewer structs
*/

	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_reconfigure",
			stvSystemInitd | stvSystemActive )) != stvStatusOk )
		return( cc );

	/* Do Your Thing */

	if ( STV_GCOORD == 0 )
		stv_leave_group( STVSHADOWNAME, STV_GCOORD, STV_NTASKS );

	return( stvStatusOk );
}


/*  -------------  stv_findParticle() -------------- */
/* Find a particle description from a descriptor list */

STV_PARTICLEPTR
stv_findParticle( PT, id )
STV_PARTICLEPTR PT;
int id;
{
	STV_PARTICLEPTR ptr;

	if ( ( id & PARTICLEBIT ) == 0
			|| ( ( id & DECOMPBIT ) | ( id & FIELDBIT )
					| ( id & PARAMBIT ) ) )
	{
		sprintf( STVERRTXT, "Invalid Particle ID (%d)\n", id );
		STV_ERROR( "stv_findParticle", stvStatusBadParam, STVERRTXT );
		return( (STV_PARTICLEPTR) NULL );
	}

	ptr = PT;

	while ( ptr->particleID < id && ptr->next )
		ptr = ptr->next;

	if ( ptr->particleID == id )
		return( ptr );

	else
		return( (STV_PARTICLEPTR) NULL );
}


/*  -------------  stv_findDecomp() -------------- */
/* find a decompostion description from decomp list */

STV_DECOMPPTR
stv_findDecomp( head, id )
STV_DECOMPPTR head;
int id;
{
	STV_DECOMPPTR current;

	if ( ( id & DECOMPBIT ) == 0
			|| ( ( id & FIELDBIT ) | ( id & PARAMBIT )
					| ( id & PARTICLEBIT ) ) )
	{
		sprintf( STVERRTXT, "Invalid Decomp ID (%d).", id );
		STV_ERROR( "stv_findDecomp", stvStatusBadParam, STVERRTXT );
		return( (STV_DECOMPPTR) NULL );
	}

	current = head;

	while ( current->decompID < id && current->next )
		current = current->next;

	if ( current->decompID == id )
		return( current );
	else
		return( (STV_DECOMPPTR) NULL );
}


/*  -------------  stv_findField() -------------- */
/* find a field description from a field list  */

STV_FIELDPTR
stv_findField( head, id )
STV_FIELDPTR head;
int id;
{
	STV_FIELDPTR current;

	if ( ( id & FIELDBIT ) == 0
			|| ( ( id & DECOMPBIT ) | ( id & PARAMBIT )
					| ( id & PARTICLEBIT ) ) )
	{
		sprintf( STVERRTXT, "Invalid Field ID (%d).", id );
		STV_ERROR( "stv_findField", stvStatusBadParam, STVERRTXT );
		return( (STV_FIELDPTR) NULL );
	}

	current = head;

	while ( current->fieldID < id && current->next )
		current = current->next;

	if ( current->fieldID == id )
		return( current );
	else
		return( (STV_FIELDPTR) NULL );
}


/*  -------------  stv_findParam() -------------- */
/* find a parameter description from a parameter list  */

STV_PARAMPTR
stv_findParam( head, id )
STV_PARAMPTR head;
int id;
{
STV_PARAMPTR current;

	if ( ( id & PARAMBIT ) == 0
			|| ( ( id & FIELDBIT ) | ( id & DECOMPBIT )
					| ( id & PARTICLEBIT ) ) )
	{
		sprintf( STVERRTXT, "Invalid Parameter ID (%d).", id );
		STV_ERROR( "stv_findParam", stvStatusBadParam, STVERRTXT );
		return( (STV_PARAMPTR) NULL );
	}

	current = head;

	while ( current->paramID < id && current->next )
		current = current->next;

	if ( current->paramID == id )
		return( current );
	else
		return( (STV_PARAMPTR) NULL );
}


/*  -------------  stv_findViewer() -------------- */
/* find a viewer description from a viewer list  */

STV_FE_VIEWERPTR
stv_findViewer( head, id )
STV_FE_VIEWERPTR head;
int id;
{
	while ( head != (STV_FE_VIEWERPTR) NULL )
	{
		if ( head->vtid == id )
			return( head );
		head = head->next;
	}
	return( (STV_FE_VIEWERPTR) NULL );
}


/*  -------------  stv_findFieldGroup() -------------- */
/* find a field Group description from a field group list  */

STV_FIELD_GROUP_PTR
stv_findFieldGroup( fg_list, fgid )
STV_FIELD_GROUP_PTR fg_list;
int fgid;
{
	STV_FIELD_GROUP_PTR fg;

	fg = fg_list;

	while ( fg )
	{
		if ( fg->fgid == fgid )
			return( fg );
		fg = fg->next;
	}

	return( (STV_FIELD_GROUP_PTR) NULL );
}


/*  -------------  stv_recvMessage() -------------- */
/* Receive a message and partially unpack it. A linked list of message
buffers is used so handling of different messages can be localized.

	fgid = -1; wildcard
	op = -1; wildcard
	src = -1; wildcard
	tag = -1; wildcard

This routine will drain the incoming messages and queue their buffers.
It will block on the receive operation if blkType is set to stvMsgBlock
or stvMsgBlockTimeout.

For fault tolerance, if the src is set to a specific (tid, fgid) pair
and the tid fails (indicated by a notify message), then the routine
will return a NULL msg buffer and buffer a message with the op
stvFieldHaltAll
*/

STV_BUFPTR
stv_recvMessage( fgid, op, src, tag, ctxt, blkType )
int fgid;
int op;
int src;
int tag;
STV_CONTEXT ctxt;
int blkType;
{
	STV_MSG_PASS_THRU PT;

	STV_BUFPTR rbuf, bufl;

	int bufid, stvop, msrc, mbytes, mtag;
	int new;

	if ( STV_OPT_VERBOSE & STV_DEBUG_MSG )
	{
		printf( "Entry %s: fgid=%d op=%d src=%x tag=%d blkType=%d\n",
				"stv_recvMessage()", fgid, op, src, tag, blkType );

		if ( nSTVBUFFERS > 0 )
		{
			printf( "%s: nSTVBUFFERS=%d\n",
					"stv_recvMessage()", nSTVBUFFERS );
		}
	}

	/* Drain messages w/tag & ctxt, unpack and search op, src... */
	/* Always unpack the fgid, in case it means something...  :-) */

	rbuf = &STVBUFFERS;

	bufl = (STV_BUFPTR) NULL;

	new = 0;

	while (
		( !new && rbuf && rbuf->next )
		|| (
			( ( blkType == stvMsgNoBlock || blkType == stvMsgProbe ) ?
				stv_recv_msg_noblock( -1, tag, ctxt, &bufid )
				: (
					( blkType == stvMsgBlock
							|| blkType == stvMsgProbeBlock ) ?
						stv_recv_msg( -1, tag, ctxt, &bufid )
						: stv_recv_msg_timeout( -1, tag, ctxt,
							&STV_TIMEOUT, &bufid )
				)
			)
			&& bufid > 0 )
		)
	{
		/* Existing Buffer */

		if ( !new && rbuf && rbuf->next )
			rbuf = rbuf->next;

		/* New Buffer */

		else
		{
			stv_msg_buf_info( bufid, &mbytes, &mtag, &msrc );

			stv_unpack_int( bufid, &stvop, 1, 1 );

			rbuf = stv_create_mbuffer();
			STV_NO_MEM_TYPE_return( rbuf,
					"stv_recvMessage", STV_BUFPTR );

			rbuf->bufid = bufid;

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_ULIB | STV_DEBUG_CPD
						| STV_DEBUG_VIEWER | STV_DEBUG_MSG ) )
			{
				printf( "%s: %s bufid=%d stvop=%s/%d msrc=%x src=%x\n",
					"stv_recvMessage()",
					( blkType == stvMsgNoBlock
							|| blkType == stvMsgBlock
							|| blkType == stvMsgBlockTimeout )
						? "Recv" : "Probe",
					bufid, stv_msg_op_str( stvMsgOp( stvop ) ), stvop,
					msrc, src );
			}

			/* this was a notify, failed tid is in stvop */
			if ( stvop >= 0 )
			{
				if ( STV_OPT_VERBOSE
						& ( STV_DEBUG_ULIB | STV_DEBUG_CPD
							| STV_DEBUG_VIEWER | STV_DEBUG_MSG
							| STV_DEBUG_NOTIFY ) )
				{
					printf( "%s: Tid %d (%x) exited (notify from %x)\n",
						"stv_recvMessage()", stvop, stvop, msrc );
				}

				rbuf->op = stvNotify;

				rbuf->src = stvop;
				rbuf->tag = mtag;
				rbuf->ctxt = ctxt;
			}

			/* else regular message */
			else
			{
				rbuf->op = stvMsgOp( stvop );

				rbuf->src = msrc;
				rbuf->tag = mtag;
				rbuf->ctxt = ctxt;

				stv_unpack_int( bufid, &(rbuf->fgid), 1, 1 );

				if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB
						&& rbuf->fgid >= 0 )
					printf( "\tField Group / Generation = %d\n",
							rbuf->fgid );
			}

			/* first new message */
			if ( !bufl )
				bufl = &STVBUFFERS;

			/* put message at end of list */
			while ( bufl->next ) bufl = bufl->next;
			LISTPUTAFTER( bufl, rbuf, prev, next );
			nSTVBUFFERS++;

			/* new messages have been received */
			/* (existing buffer has been traversed) */
			new++;
		}

		/* Check for matching message */

		if ( STV_CONTEXT_MATCH( ctxt, rbuf->ctxt )
				&& ( src == -1 || src == rbuf->src )
				&& ( fgid == -1 || fgid == rbuf->fgid )
				&& ( op == -1 || op == rbuf->op ) )
		{
			if ( STV_OPT_VERBOSE & STV_DEBUG_MSG )
				printf( "%s: matching buffer found, return.\n",
						"stv_recvMessage()" );

			if ( blkType != stvMsgProbe && blkType != stvMsgProbeBlock )
			{
				LISTDELETE( rbuf, prev, next );
				nSTVBUFFERS--;
			}

			return( rbuf );
		}

		/* check for pass-throughs */

		PT = STVMSGPASSTHRUS;

		while ( PT != NULL )
		{
			if ( rbuf->op == PT->op )
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_MSG )
				{
					printf( "%s: pass-through, message op=%s.\n",
							"stv_recvMessage()",
							stv_msg_op_str( rbuf->op ) );
				}

				LISTDELETE( rbuf, prev, next );
				nSTVBUFFERS--;

				return( rbuf );
			}

			PT = PT->next;
		}
	}

	if ( STV_OPT_VERBOSE & STV_DEBUG_MSG )
		printf( "%s: no matching messages.\n", "stv_recvMessage()" );

	return( (STV_BUFPTR) NULL );
}


/* Define User Particle */

int
stv_particleDefine( name, dataDim, glb, gub, ntasks,
	get_particle, make_particle )
char *name;
int dataDim;
int *glb;
int *gub;
int ntasks;
STV_GET_PARTICLE get_particle;
STV_MAKE_PARTICLE make_particle;
{
	STV_PARTICLEPTR PT;
	STV_PARTICLEPTR ptr;

	int noadd;
	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_particleDefine",
			stvSystemInitd )) != stvStatusOk )
		return( cc );

	/* Check User Function ( O.K. to have make_particle() NULL ) */

	if ( get_particle == NULL )
		return( stvStatusBadParam );

	/* Check for Previous Particle Definition */

	PT = stv_get_particle_name( name,
			STVPARTICLES.next, nSTVPARTICLES );

	if ( PT != NULL )
	{
		/* Existing Particle -- Clear Substructures Associated */

		if ( PT->name != NULL )
			FREE( PT->name );
		if ( PT->glb != NULL )
			FREE( PT->glb );
		if ( PT->gub != NULL )
			FREE( PT->gub );

		noadd = stvTrue;
	}

	/* Create New Particle */

	else
	{
		PT = stv_create_particle();

		STV_NO_MEM_return( PT, "stv_particleDefine" );

		PT->particleID = PARTICLEBIT | nSTVPARTICLES++;

		noadd = stvFalse;
	}

	PT->refcount = 1;

	CPALLOC( PT->name, name, strlen( name ) + 1, char,
			"stv_particleDefine" );
	STV_NO_MEM_return( PT->name, "stv_particleDefine" );

	PT->dataDim = dataDim;

	CPALLOC( PT->glb, glb, dataDim, int, "stv_particleDefine" );
	STV_NO_MEM_return( PT->glb, "stv_particleDefine" );

	CPALLOC( PT->gub, gub, dataDim, int, "stv_particleDefine" );
	STV_NO_MEM_return( PT->gub, "stv_particleDefine" );

	PT->ntasks = ntasks;

	PT->get_particle = get_particle;

	PT->make_particle = make_particle;

	/* Add to Particle List */

	if ( !noadd )
	{
		ptr = &STVPARTICLES;

		while ( ptr->particleID < PT->particleID && ptr->next != NULL )
			ptr = ptr->next;

		LISTPUTAFTER( ptr, PT, prev, next );
	}

	return( PT->particleID );
}


/* ------------- stv_decompDefine() ---------------- */
/* decompID =
 *  stv_decompDefine( dataDim, axisType, axisInfo1, axisInfo2,
 *      glb, gub, prank, pshape )
 *
 *  int dataDim;    # dimension of data distributed on processor array
 *  int *axisType;  # decomposition type for each axis:
 *                  #   ( i = 0 : dataDim ) axisType[i] =
 *                  #       [ stvBlock, stvCyclic, stvCollapse,
 *                  #            stvExplicit ],
 *  int *axisInfo1; # axisInfo1[i] = 1st arg for axisType[i] (if needed)
 *  int *axisInfo2; # axisInfo2[i] = 2nd arg for axisType[i] (if needed)
 *  int *glb;       # dataDim lower bound(s) of global comp region
 *  int *gub;       # dataDim upper bound(s) of global comp region
 *  int prank;      # dimension of processor array
 *  int *pshape;    # number processors in each coordinate direction
 *
 * returns:
 *       decompID  -- identifier handle for this decomposition
 *
 * Discussion:
 *
 *  This routine defines the way indices of a yet-to-be-defined
 *  dataDim-dimensional array will be decomposed onto a
 *  prank-dimensional mesh of processors.
 *
 *  If r of the axisTypes are stvBlock, stvCyclic or stvExplicit,
 *  therefore actually requiring decomposition across some processor
 *  axis, then prank must be greater than or equal to r.
 *
 *  The decompositions specifications are patterned after the HPF
 *  standard.
 *
 * Examples:
 *   3d array on 2d processor mesh, uniform block decomposition
 *       dataDim = 3
 *       axisType = [ stvBlock, stvBlock, stvCollapse ]
 *       axisInfo1 = axisInfo2 = stvstvDefaultInfo
 *       prank = 2, pshape = [ 2, 2 ]
 *   3d array on 2d processor mesh, block in 1st dim, cyclic (10) 2nd
 *       dataDim = 3
 *       axisType = [ stvBlock, stvCyclic, stvCollapse ]
 *       axisInfo1 = [ -1, 10, -1 ], axisInfo2 = stvDefaultInfo
 *       prank = 2, pshape = [ 2, 2 ]
 *   3d array on 1d processor mesh, block (10) in 1st dim
 *       dataDim = 3
 *       axisType = [ stvBlock, stvCollapse, stvCollapse ]
 *       axisInfo1 = [ 10, -1, -1 ], axisInfo2 = stvDefaultInfo
 *       prank = 1, pshape = [4]
 *   3d array on 2d processor mesh, cyclic (3) in 1st dim, explicit 3rd
 *       dataDim = 3
 *       axisType = [ stvCyclic, stvCollapse, stvExplicit ]
 *       axisInfo1 = [ 3, -1, MINZ ], axisInfo2 = [ -1, -1, MAXZ ]
 *           (where [ MINZ, MAXZ ] are the elements along the 3rd axis
 *            which are stored on this specific processor)
 *       prank = 2, pshape = [ 6, 4 ]
 *   3d local array, no decomposition
 *       dataDim = 3
 *       axisType = stvLocalArray
 *       axisInfo1 = axisInfo2 = stvLocalArray
 *       prank = 1, pshape = stvLocalArray
 */

int
stv_decompDefine( dataDim, axisType, axisInfo1, axisInfo2, glb, gub,
		prank, pshape )
int dataDim;
int *axisType;
int *axisInfo1, *axisInfo2;
int *glb;
int *gub;
int prank;
int *pshape;
{
	STV_DECOMPPTR newDecomp, current;

	int local_pshape = 1;
	int i, j;
	int noncollapse = 0;
	int ntasks = 1;
	int thisid;
	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_decompDefine",
			stvSystemInitd )) != stvStatusOk )
		return( cc );

	/* Check Arguments */

	if ( dataDim > STV_MAX_DIM ) {
		sprintf( STVERRTXT,
				"Array dimension (%d) larger than STV_MAX_DIM (%d)",
				dataDim, STV_MAX_DIM ); 
		STV_ERROR( "stv_decompDefine", stvStatusBadParam, STVERRTXT );
		return( stvStatusBadParam );
	}

	if ( axisType == NULL )
		axisType = STV_DUMMY_AXIS_TYPE;
	if ( axisInfo1 == NULL )
		axisInfo1 = STV_DUMMY_AXIS_INFO;
	if ( axisInfo2 == NULL )
		axisInfo2 = STV_DUMMY_AXIS_INFO;

	if ( pshape == NULL )
	{
		pshape = &local_pshape;

		if ( prank != 1 )
		{
			sprintf( STVERRTXT,
				"Warning: Non-Unit Processor Rank (%d) in Local Array.",
				prank );
			STV_ERROR( "stv_decompDefine", stvStatusBadParam,
				STVERRTXT );
			prank = 1;
		}
	}

	/* Check Topology */

	for ( i=0 ; i < prank ; i++ )
		ntasks *= *(pshape + i);

	if ( ntasks <= 0 )
	{
		sprintf( STVERRTXT,
				"Zero or Negatively Sized Axis in Processor Shape." );
		STV_ERROR( "stv_decompDefine", stvStatusBadParam, STVERRTXT );
		return( stvStatusBadParam );
	}

	for ( i=0 ; i < dataDim ; i++ )
	{
		if ( *( axisType + i ) == stvBlock
				|| *( axisType + i ) == stvCyclic
				|| *( axisType + i ) == stvExplicit )
		{
			noncollapse++;
		}
	}

	if ( prank < noncollapse )
	{
		sprintf( STVERRTXT,
				"%d-D Data on %d-D Mesh with %d Collapsed Maps.",
				dataDim, prank, dataDim - noncollapse );
		STV_ERROR( "stv_decompDefine", stvStatusBadParam, STVERRTXT );
		return( stvStatusBadParam );
	}

	thisid = nDECOMPS++;
	thisid = thisid | DECOMPBIT;

	/* allocate memory for DECOMP structure and fill in */

	newDecomp = stv_create_decomp();
	STV_NO_MEM_return( newDecomp, "stv_decompDefine" );

	newDecomp->refcount = 1;
	newDecomp->decompID = thisid;
	newDecomp->dataDim = dataDim;
	newDecomp->prank = prank;
	newDecomp->ntasks = ntasks;

	CPALLOC( newDecomp->axisType, axisType, dataDim, int,
			"stv_decompDefine" );
	STV_NO_MEM_return( newDecomp->axisType, "stv_decompDefine" );

	CPALLOC( newDecomp->axisInfo, axisInfo1, dataDim, int,
			"stv_decompDefine" );
	STV_NO_MEM_return( newDecomp->axisInfo, "stv_decompDefine" );

	newDecomp->axisLBounds = (int **) CALLOC( (STV_SIZE_T) dataDim
			* sizeof( int * ), "stv_decompDefine" );
	STV_NO_MEM_return( newDecomp->axisLBounds, "stv_decompDefine" );

	newDecomp->axisUBounds = (int **) CALLOC( (STV_SIZE_T) dataDim
			* sizeof( int * ), "stv_decompDefine" );
	STV_NO_MEM_return( newDecomp->axisUBounds, "stv_decompDefine" );

	newDecomp->nbounds = (int *) CALLOC( (STV_SIZE_T) dataDim
			* sizeof(int), "stv_decompDefine" );
	STV_NO_MEM_return( newDecomp->nbounds, "stv_decompDefine" );

	CPALLOC( newDecomp->glb, glb, dataDim, int, "stv_decompDefine" );
	STV_NO_MEM_return( newDecomp->glb, "stv_decompDefine" );

	CPALLOC( newDecomp->gub, gub, dataDim, int, "stv_decompDefine" );
	STV_NO_MEM_return( newDecomp->gub, "stv_decompDefine" );

	CPALLOC( newDecomp->pshape, pshape, prank, int,
			"stv_decompDefine" );
	STV_NO_MEM_return( newDecomp->pshape, "stv_decompDefine" );

	/* Convert default values for axisInfo to something we can use */

	j = 0;

	for ( i=0 ; i < dataDim ; i++ )
	{
		/* Translate Old Explicit Args to New Bounds Arrays */

		if ( newDecomp->axisType[i] == stvExplicit )
		{
			newDecomp->axisLBounds[i] = (int *) CALLOC( sizeof(int),
					"stv_decompDefine" );
			STV_NO_MEM_return( newDecomp->axisLBounds[i],
					"stv_decompDefine" );

			newDecomp->axisUBounds[i] = (int *) CALLOC( sizeof(int),
					"stv_decompDefine" );
			STV_NO_MEM_return( newDecomp->axisUBounds[i],
					"stv_decompDefine" );

			newDecomp->axisLBounds[i][0] = axisInfo1[i];
			newDecomp->axisUBounds[i][0] = axisInfo2[i];

			newDecomp->axisInfo[i] = -1;

			newDecomp->nbounds[i] = 1;
		}

		else
		{
			newDecomp->axisLBounds[i] = (int *) NULL;
			newDecomp->axisUBounds[i] = (int *) NULL;

			newDecomp->nbounds[i] = 0;
		}

		/* Check Decomposition */

		stv_check_decomp_axis_info(
				newDecomp->axisType[i], newDecomp->axisInfo + i,
				newDecomp->axisLBounds[i], newDecomp->axisUBounds[i],
				newDecomp->nbounds[i],
				newDecomp->glb[i], newDecomp->gub[i],
				newDecomp->pshape[j] );

		if ( newDecomp->axisType[i] != stvCollapse ) j++;
	}

	/* Insert the data into the decomp list */

	current = &DECOMPS;

	while ( current->decompID < thisid && current->next )
		current = current->next;

	LISTPUTAFTER( current, newDecomp, prev, next );

	return( thisid );
}


/* ------------- stv_decompAddExplicitBounds() ---------------- */
/*
 * cc = stv_decompAddExplicitBounds( decompID, axis,
 *       lowerBound, upperBound )
 *
 * (or vector version:)
 * cc = stv_decompAddExplicitVBounds( decompID, axis,
 *       lowerBounds, upperBounds, nbounds )
 *
 *  int decompID;    # identifier handle for the given decomposition
 *  int axis;        # index of data axis, 0 <= axis < dataDim
 *  int lowerBound;  # lower bound for region, along given axis
 *  int upperBound;  # upper bound for region, along given axis
 *
 *  int lowerBounds; # array of lower bound for region, along given axis
 *  int upperBounds; # array of upper bound for region, along given axis
 *  int nbounds;     # size of lowerBounds & upperBounds arrays
 *
 * returns:
 *       cc -- success (stvStatusOk) or failure (error code) of add.
 *
 * Discussion:
 *  This routine allows the definition of a multi-explicit data
 *  decomposition, by defining additional explicit region ranges
 *  along a particular data axis.
 *
 * Example:
 *
 *   3d array on 2d processor mesh, explicit in 1st & 3rd axes
 *       -> axis #1: [ 10, 20 ] + [ 24, 31 ] + [ 45, 48 ]
 *       -> axis #2: collapse
 *       -> axis #3: [ 5, 15 ] + [ 65, 70 ]
 *
 *   axisType = { stvExplicit, stvCollapse, stvExplicit }
 *   axisInfo1 = { 10, stvDefault, 5 }
 *   axisInfo2 = { 20, stvDefault, 15 }
 *
 *   decompId = stv_decompDefine( 3,
 *        axisType, axisInfo1, axisInfo2, ... );
 *
 *   stv_decompAddExplicitBounds( decompId, 0, 24, 31 );
 *   stv_decompAddExplicitBounds( decompId, 0, 45, 48 );
 *
 *   (or equivalently using vectored routine:)
 *   stv_decompAddExplicitVBounds( decompId, 0,
 *        { 24, 45 }, { 31, 48 }, 2 );
 *
 *   stv_decompAddExplicitBounds( decompId, 2, 65, 70 );
 */

int
stv_decompAddExplicitBounds( decompID, axis, lowerBound, upperBound )
int decompID;
int axis;
int lowerBound;
int upperBound;
{
	return( stv_decompAddExplicitVBounds( decompID, axis,
			&lowerBound, &upperBound, 1 ) );
}

int
stv_decompAddExplicitVBounds( decompID, axis,
		lowerBounds, upperBounds, nbounds )
int decompID;
int axis;
int *lowerBounds;
int *upperBounds;
int nbounds;
{
	STV_DECOMPPTR decomp;

	int *axisLBounds;
	int *axisUBounds;

	int new_nbounds;
	int cc;
	int i;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_decompAddExplicitVBounds",
			stvSystemInitd )) != stvStatusOk )
		return( cc );

	/* Find Decomposition */

	if ( ( decomp = stv_findDecomp( &DECOMPS, decompID ) )
			== (STV_DECOMPPTR) NULL )
	{
		sprintf( STVERRTXT, "Couldn't Find Decompostion (%d).",
				decompID );
		STV_ERROR( "stv_decompAddExplicitVBounds", stvStatusBadParam,
				STVERRTXT );
		return( stvStatusBadParam );
	}

	/* Check Arguments */

	if ( decomp->axisType[ axis ] != stvExplicit )
	{
		STV_ERROR( "stv_decompAddExplicitVBounds", stvStatusBadParam,
				"Axis Type Not Explicit" );
		return( stvStatusBadParam );
	}

	/* Re-Allocate Axis Bounds Arrays */

	new_nbounds = decomp->nbounds[ axis ] + nbounds;

	axisLBounds = (int *) CALLOC( (unsigned) new_nbounds
			* sizeof( int ), "stv_decompAddExplicitVBounds" );
	STV_NO_MEM_return( axisLBounds,
			"stv_decompAddExplicitVBounds" );

	axisUBounds = (int *) CALLOC( (unsigned) new_nbounds
			* sizeof( int ), "stv_decompAddExplicitVBounds" );
	STV_NO_MEM_return( axisUBounds,
			"stv_decompAddExplicitVBounds" );

	/* Copy Over Existing Bounds */

	for ( i=0 ; i < decomp->nbounds[ axis ] ; i++ )
	{
		axisLBounds[i] = decomp->axisLBounds[ axis ][i];
		axisUBounds[i] = decomp->axisUBounds[ axis ][i];

		decomp->axisLBounds[ axis ][i] = -1;
		decomp->axisUBounds[ axis ][i] = -1;
	}

	FREE( decomp->axisLBounds[ axis ] );
	FREE( decomp->axisUBounds[ axis ] );

	/* Add New Explicit Bounds */

	for ( i=0 ; i < nbounds ; i++ )
	{
		axisLBounds[ i + decomp->nbounds[ axis ] ] = lowerBounds[i];
		axisUBounds[ i + decomp->nbounds[ axis ] ] = upperBounds[i];
	}

	decomp->nbounds[ axis ] = new_nbounds;

	decomp->axisLBounds[ axis ] = axisLBounds;
	decomp->axisUBounds[ axis ] = axisUBounds;

	/* Check Decomposition */

	stv_check_decomp_axis_info(
			decomp->axisType[ axis ], decomp->axisInfo + axis,
			decomp->axisLBounds[ axis ], decomp->axisUBounds[ axis ],
			decomp->nbounds[ axis ],
			decomp->glb[ axis ], decomp->gub[ axis ],
			/* decomp->pshape[j] unnecessary for stvExplicit */ -1 );

	return( stvStatusOk );
}


/* ------------- stv_fieldDefine() ---------------- */
/* fieldID =
 *    stv_fieldDefine( var, name, decompID, arrayOffset, arrayDecl,
 *        type, paddr, aflag )
 *
 *    STV_VALUE var;       # pointer to local memory array
 *    char *name;          # user-defined descriptive name
 *    int decompID;        # decompID how the global array is
 *    int *arrayOffset;    # offset to "valid" data in local
 *                               (declared) array data
 *    int *arrayDecl;      # size of the local (declared) dimensions
 *                               of data
 *    int type;            # data type: stvInt, stvFloat, stvDouble
 *    int *paddr;          # address of this processor in the
 *                               processor map.
 *    int aflag;           # action flag, stvVisOnly = View Only,
 *                               stvCpOnly = Checkpoint only,
 *                               stvVisCp = Both.
 *
 * returns:
 *    fieldID  -- identifier for the field
 *
 * Discussion:
 *    This routine defines a field that may be sent to the
 *    visualization front-end.  The field is named by the program
 *    so that the front-end can send the output to the correct module.
 *    A decomposition must be declared using stv_decompDefine().
 *    The field decomposition implicitly defines the mapping from the
 *    local storage pointed to by var to the global indices.  So that
 *    packing is handled correctly, arrayDecl[] must hold the declared
 *    dimensions of the array.  A type descriptor is also needed.
 *    Finally the processor address in the processor array must be
 *    given.  For example, if the processor array is 2-D, then paddr
 *    is a 2-vector.
 *
 * Examples:
 *
 *    Float ( 100, 100, 100 )
 *        - array on 2d ( 10, 5 ) processor mesh,
 *              ( block, cyclic, collapse ) distribution
 *
 *    float localdata[10][20][100];
 *    ldecl = { 10, 20, 100 };
 *    axisType = { stvBlock, stvCyclic, stvCollapse };
 *    glb = { 0, 0, 0 };  gub={ 99, 99, 99 };
 *    pshape = { 10, 5 };  prank = 2;
 *    paddr = { i, j }     # i = 0 : 9, j = 0 : 4
 *
 *    decompID = stv_decompDefine( 3, axisType,
 *                   stvstvDefaultInfo, stvstvDefaultInfo,
 *                   glb, gub, 2, pshape );
 *
 *    fieldID = stv_fieldDefine( localdata, "foo", decompID,
 *                  stvNoArrayOffsets, ldecl, stvFloat,
 *                  paddr, stvVisCp );
 */

int
stv_fieldDefine( var, name, decompID, arrayOffset, arrayDecl, type,
		paddr, aflag )
STV_VALUE var;
char *name;
int decompID;
int *arrayOffset;
int *arrayDecl;
int type;
int *paddr;
int aflag;
{
	STV_FIELDPTR newField, current;

	STV_DECOMPPTR decomp;

	int local_paddr = 0;

	int thisid;
	int noadd;
	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_fieldDefine",
			stvSystemInitd )) != stvStatusOk )
		return( cc );

	/* Check Arguments */

	if ( paddr == NULL )
		paddr = &local_paddr;

	/* allocate memory for FIELD structure and fill in */

	if ( ( decomp = stv_findDecomp( &DECOMPS, decompID ) )
			== (STV_DECOMPPTR) NULL )
	{
		sprintf( STVERRTXT, "Couldn't Find Decompostion (%d).",
				decompID );
		STV_ERROR( "stv_fieldDefine", stvStatusBadParam, STVERRTXT );
		return( stvStatusBadParam );
	}

	/* Existing Field -- clear substructures associated */

	if ( newField = stv_get_field_name( name,
			STVFIELDS.next, nSTVFIELDS ) )
	{
		thisid = newField->fieldID;

		if ( newField->name != NULL )
			FREE( newField->name );
		if ( newField->arrayOffset != NULL )
			FREE( newField->arrayOffset );
		if ( newField->arrayDecl != NULL )
			FREE( newField->arrayDecl );
		if ( newField->paddr != NULL )
			FREE( newField->paddr );
		FREELIST( newField->vlist, prev, next );
		if ( newField->aflag & stvVisOnly ) nVISFIELDS--;
		if ( newField->aflag & stvCpOnly ) nCPFIELDS--;
		noadd = stvTrue;
	}

	/* New Field -- allocate */

	else
	{
		thisid = nSTVFIELDS++;
		thisid = thisid | FIELDBIT;

		newField = stv_create_field();
		STV_NO_MEM_return( newField, "stv_fieldDefine" );

		noadd = stvFalse;
	}

	if ( aflag & stvVisOnly )
		nVISFIELDS++; /* update nVISFIELDS, nCPFIELDS*/

	if ( aflag & stvCpOnly )
		nCPFIELDS++;

	newField->refcount = 1;
	newField->var = var;
	newField->type = type;
	newField->decomp = decomp;
	newField->fieldID = thisid;
	newField->decompID = decompID;
	newField->storage_order = stvRowMajor;
	newField->aflag = aflag;

	/* no default time scale, just use iterations starting at 0 */
	newField->time_scale = stv_time_of( 1, stvTimeNone );
	newField->time_base = 0;
	newField->time_units = stvTimeNone;

	/* default status is ready to send */
	newField->data_ready = stvTrue;

	CPALLOC( newField->name, name, strlen( name ) + 1, char,
			"stv_fieldDefine" );
	STV_NO_MEM_return( newField->name, "stv_fieldDefine" );

	if ( arrayOffset != NULL )
	{
		CPALLOC( newField->arrayOffset, arrayOffset, decomp->dataDim,
				int, "stv_fieldDefine" );
		STV_NO_MEM_return( newField->arrayOffset, "stv_fieldDefine" );
	}

	else
		newField->arrayOffset = (int *) NULL;

	CPALLOC( newField->arrayDecl, arrayDecl, decomp->dataDim, int,
			"stv_fieldDefine" );
	STV_NO_MEM_return( newField->arrayDecl, "stv_fieldDefine" );

	CPALLOC( newField->paddr, paddr, decomp->prank, int,
			"stv_fieldDefine" );
	STV_NO_MEM_return( newField->paddr, "stv_fieldDefine" );

	newField->vlist = (STV_VLISTPTR) CALLOC( sizeof(STV_VLIST),
			"stv_fieldDefine" );
	STV_NO_MEM_return( newField->vlist, "stv_fieldDefine" );

	/* Insert the data into the field list */

	if ( !noadd )
	{
		current = &STVFIELDS;

		while ( current->fieldID < thisid && current->next )
			current = current->next;

		LISTPUTAFTER( current, newField, prev, next );
	}

	return( thisid );
}


/* Original Particle Field Define - For Backwards Compat */

int
stv_particleFieldDefine( name, particleID, type, nelems,
	get_pfield, set_pfield, aflag )
char *name;
int particleID;
int type;
int nelems;
STV_GET_PFIELD get_pfield;
STV_SET_PFIELD set_pfield;
int aflag;
{
	static int old_routine = 0;

	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_particleFieldDefine",
			stvSystemInitd )) != stvStatusOk )
		return( cc );

	/* Obsolescence Check */

	if ( !old_routine++ )
	{
		fprintf( stderr, "Warning: %s is obsolete - use %s\n",
			"stv_particleFieldDefine()", "stv_pfieldDefine()" );
	}

	/* Obsolescence Fix :-) */

	return( stv_pfieldDefine( name, particleID, type, nelems,
		get_pfield, (STV_VALUE) NULL, set_pfield, (STV_VALUE) NULL,
		aflag ) );
}


/* Particle Field Define */

int
stv_pfieldDefine( name, particleID, type, nelems,
	get_pfield, get_pfield_arg, set_pfield, set_pfield_arg, aflag )
char *name;
int particleID;
int type;
int nelems;
STV_GET_PFIELD get_pfield;
STV_VALUE get_pfield_arg;
STV_SET_PFIELD set_pfield;
STV_VALUE set_pfield_arg;
int aflag;
{
	STV_PARTICLEPTR PT;
	STV_FIELDPTR F;
	STV_FIELDPTR ptr;

	int noadd;
	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_pfieldDefine",
			stvSystemInitd )) != stvStatusOk )
		return( cc );

	/* Check User Functions */

	if ( get_pfield == NULL
			|| ( aflag & stvCpOnly && set_pfield == NULL ) )
	{
		return( stvStatusBadParam );
	}

	/* Check for Valid Particle */

	PT = stv_findParticle( &STVPARTICLES, particleID );

	if ( PT == NULL )
	{
		sprintf( STVERRTXT, "Couldn't Find Particle ID = %d\n",
				particleID );
		STV_ERROR( "stv_pfieldDefine", stvStatusBadParam,
			STVERRTXT );
		return( stvStatusBadParam );
	}

	/* Check for Checkpointing Capabilities */

	if ( aflag & stvCpOnly && PT->make_particle == NULL )
	{
		sprintf( STVERRTXT,
				"Particle ID %d - No make_particle() Defined.\n",
				particleID );
		STV_ERROR( "stv_pfieldDefine", stvStatusBadParam,
			STVERRTXT );
		return( stvStatusBadParam );
	}

	/* Create New Field */

	F = stv_get_field_name( name, STVFIELDS.next, nSTVFIELDS );

	if ( F != NULL )
	{
		FREELIST( F->vlist, prev, next );
		if ( F->aflag & stvVisOnly ) nVISFIELDS--;
		if ( F->aflag & stvCpOnly ) nCPFIELDS--;
		noadd = stvTrue;
	}

	else
	{
		F = stv_create_field();
		STV_NO_MEM_return( F, "stv_pfieldDefine" );

		F->fieldID = FIELDBIT | nSTVFIELDS++;

		CPALLOC( F->name, name, strlen( name ) + 1, char,
			"stv_pfieldDefine" );
		STV_NO_MEM_return( F->name, "stv_pfieldDefine" );

		noadd = stvFalse;
	}

	F->refcount = 1;

	F->var = (STV_VALUE) NULL;

	F->type = type;

	F->nelems = nelems;

	F->particle = PT;
	F->particleID = particleID;

	F->get_pfield = get_pfield;
	F->get_pfield_arg = get_pfield_arg;

	F->set_pfield = set_pfield;
	F->set_pfield_arg = set_pfield_arg;

	F->aflag = aflag;

	/* no default time scale, just use iterations starting at 0 */
	F->time_scale = stv_time_of( 1, stvTimeNone );
	F->time_base = 0;
	F->time_units = stvTimeNone;

	/* default status is ready to send */
	F->data_ready = stvTrue;

	F->vlist = (STV_VLISTPTR) CALLOC( sizeof(STV_VLIST),
			"stv_pfieldDefine" );
	STV_NO_MEM_return( F->vlist, "stv_pfieldDefine" );

	if ( aflag & stvVisOnly ) nVISFIELDS++;

	if ( aflag & stvCpOnly ) nCPFIELDS++;

	/* Add to Field List */

	if ( !noadd )
	{
		ptr = &STVFIELDS;

		while ( ptr->fieldID < F->fieldID && ptr->next != NULL )
			ptr = ptr->next;

		LISTPUTAFTER( ptr, F, prev, next );
	}

	return( F->fieldID );
}


/* ------------- stv_paramDefine() ---------------- */
/* scalar version of stv_vparamDefine() */

int
stv_paramDefine( name, var, type, aflag )
char *name;
STV_VALUE var;
int type;
int aflag;
{
	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_paramDefine",
			stvSystemInitd )) != stvStatusOk )
		return( cc );

	return( stv_vparamDefine( name, &var, &name, &type, 1, aflag ) );
}


/* ------------- stv_vparamDefine() ---------------- */
/* paramID =
*       stv_vparamDefine( name, vars, pnames, types, num, aflag )
*
*       char *name;         # name of whole honking parameter
*       STV_VALUE *vars;    # array of pointers to local memory array
*       char **pnames;  # array of user-defined descriptive names
*       int *types;     # array of data types: stvInt, stvFloat, etc.
*       int num;        # number of parameter vector elements
*		int aflag;      # stvVisOnly, stvCpOnly, stvVisCp
*
* returns:
*       paramID  -- identifier for the parameter
*
* Discussion:
*    This routine defines a parameter that may be set by visualization
*    front-end.  It is up to the user to manage any changes in
*    parameters properly.
*/

int
stv_vparamDefine( name, vars, pnames, types, num, aflag )
char *name;
STV_VALUE *vars;
char **pnames;
int *types;
int num;
int aflag;
{
	STV_PARAMPTR newParam, current;
	int thisid;
	int noadd;
	int cc;
	int i;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_vparamDefine",
			stvSystemInitd )) != stvStatusOk )
		return( cc );

	/* Check for Existing Parameter */

	if ( newParam = stv_get_param_name( name,
			STVPARAMS.next, nSTVPARAMS ) )
	{
		thisid = newParam->paramID;

		if ( newParam->aflag & stvVisOnly )
			nVISPARAMS--;

		if ( newParam->aflag & stvCpOnly )
			nCPPARAMS--;

		if ( newParam->name != NULL )
			FREE( newParam->name );

		noadd = stvTrue;
	}

	/* Define New Parameter */

	else
	{
		thisid = nSTVPARAMS++;

		thisid = thisid | PARAMBIT;

		/* allocate memory for PARAM structure and fill in */

		newParam = stv_create_param();

		STV_NO_MEM_return( newParam, "stv_vparamDefine" );

		noadd = stvFalse;
	}

	if ( aflag & stvVisOnly ) nVISPARAMS++;

	if ( aflag & stvCpOnly ) nCPPARAMS++;

	newParam->refcount = 1;

	newParam->paramID = thisid;

	newParam->aflag = aflag;

	CPALLOC( newParam->name, name, strlen( name ) + 1, char,
			"stv_vparamDefine" );
	STV_NO_MEM_return( newParam->name, "stv_vparamDefine" );

	if ( num > 0 )
	{
		newParam->pnames = (char **) CALLOC( (STV_SIZE_T) num
				* sizeof(char *), "stv_vparamDefine" );
		STV_NO_MEM_return( newParam->pnames, "stv_vparamDefine" );

		newParam->vars = (STV_VALUE *) CALLOC( (STV_SIZE_T) num
				* sizeof(STV_VALUE), "stv_vparamDefine" );
		STV_NO_MEM_return( newParam->vars, "stv_vparamDefine" );

		newParam->types = (int *) CALLOC( (STV_SIZE_T) num
				* sizeof(int), "stv_vparamDefine" );
		STV_NO_MEM_return( newParam->types, "stv_vparamDefine" );

		for ( i=0 ; i < num ; i++ )
		{
			CPALLOC( newParam->pnames[i], pnames[i],
					strlen( pnames[i] ) + 1, char, "stv_vparamDefine" );
			STV_NO_MEM_return( newParam->pnames[i],
					"stv_vparamDefine" );

			newParam->vars[i] = vars[i];

			newParam->types[i] = types[i];
		}

		newParam->num = num;
	}

	else
	{
		newParam->pnames = (char **) NULL;

		newParam->vars = (STV_VALUE *) NULL;

		newParam->types = (int *) NULL;

		newParam->num = 0;
	}

	/* Insert the data into the param list */

	if ( !noadd )
	{
		current = &STVPARAMS;

		while ( current->paramID < thisid && current->next )
			current = current->next;

		LISTPUTAFTER( current, newParam, prev, next );
	}

	return( thisid );
}


/* -------------- stv_sendToFE() ----------- */

/*
*   stv_sendToFE()                   - Original Interface Routine.
*   stv_sendReadyData( send_flag )   - New Parameterized Routine.
*
*  Send ready fields to the front-end viewers.
*      -> unless send_flag has the stvSendNoFieldTimeUpdates bit set,
*             bump each individual field's base time counter.
*
*  Returns:
*    < 0     Error
*      0     No parameters updated
*    > 0     Number of parameters updated
*
*/

int
stv_sendToFE()
{
	return( stv_sendReadyData( stvSendDefault ) );
}

int
stv_sendReadyData( send_flag )
int send_flag;
{
	STV_FIELD_GROUP_PTR fg, oldfg;
	STV_FE_VIEWERPTR vs;
	STV_PARAMPTR p;

	int blocked;
	int newParams = 0;
	int quick;
	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_sendReadyData",
			stvSystemInitd | stvSystemActive )) != stvStatusOk )
		return( cc );

	/* Debug Entry Message */

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
	{
		printf( "%s [%d]  nfields=%d, nparams=%d, nviewers=%d\n",
				"stv_sendReadyData", STV_TIMESTAMP,
				nVISFIELDS, nVISPARAMS, nVIEWERS );
	}

	/* Verify Context Obtained */

	if ( (cc = stv_verify_system_contexts( stvFalse, stvTrue )) < 0 )
	{
		printf( "Warning:  Could Not Obtain CUMULVS Context.\n" );
		printf( "\tContinuing...\n\n" );

		if ( !(send_flag & stvSendNoFieldTimeUpdates) )
		{
			stv_bump_field_times( STVFIELDS.next, nSTVFIELDS );

			STV_TIMESTAMP++;
		}

		return( 0 );
	}

	else if ( cc > 0 )
		stv_define_message_handlers();

	/* Determine if we should check for new connections */

	if ( !stv_handleMessages( stvAny,
				stvAny, STV_USERTAG, STV_APP_CTX, stvMsgNoBlock )
			&& nVIEWERS <= 0  ) 
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
			printf( "Quick Return, update timestamps\n" );

		quick = stvTrue;
	}

	else
		quick = stvFalse;

	/* set all the parameters to not updated */

	p = &STVPARAMS;	

	while ( p && p->next )
	{
		p = p->next;
		p->updated = 0;
	}

	/* Loop Until All Clear to Continue with Application */

	blocked = stvFalse;

	do
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
			printf( "Loop\n" );

		/* Make connections, Handle messages till we run out */

		if ( !quick )
		{
			if ( blocked 
					|| STV_APP_LOCK == stvAppLocked
					|| ( STV_APP_LOCK_HOLD > 0
							&& STV_TIMESTAMP >= STV_APP_LOCK_HOLD ) )
			{
				if ( STV_OPT_VERBOSE
						& ( STV_DEBUG_ULIB | STV_DEBUG_MSG ) )
					printf( "Blocking Recv...\n" );

				stv_handleMessages( stvAny,
						stvAny, STV_USERTAG, STV_APP_CTX, stvMsgBlock );
			}
			
			while ( stv_handleMessages( stvAny,
					stvAny, STV_USERTAG, STV_APP_CTX, stvMsgNoBlock )
					> 0 );

			/* Check for Application Lock */

			if ( STV_APP_LOCK_TIME > 0
					&& STV_TIMESTAMP >= STV_APP_LOCK_TIME )
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
					printf( "Application is Locked\n" );

				STV_APP_LOCK = stvAppLocked;

				/* Reset NeedSend flags for OKtoSend viewer fields */

				vs = &VIEWERS;

				while ( vs->next != (STV_FE_VIEWERPTR) NULL )
				{
					vs = vs->next;

					fg = vs->fg_list->next;

					while ( fg )
					{
						if ( fg->OKtoSend == stvTrue )
							fg->NeedSend = stvTrue;

						fg = fg->next;
					}
				}
			}
		}

		blocked = stvFalse;

		/* We've checked for conns, now we can send to all viewers. */
		/* Go through all the viewer structures that we have */

		vs = &VIEWERS;

		while ( vs->next != (STV_FE_VIEWERPTR) NULL )
		{
			vs = vs->next;

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_ULIB | STV_DEBUG_TIME
							| STV_DEBUG_FREQ ) )
			{
				printf( "Processing viewer %x at time %d\n",
					vs->vtid, STV_TIMESTAMP );
			}

			fg = vs->fg_list->next;

			/* process each field group that this viewer holds */

			while ( fg )
			{
				/* ignore this field if it's not ripe...  :-) */

				if ( STV_TIMESTAMP >= fg->toffset )
				{
					/* check for field request Tstart release... */

					if ( fg->vState == stvWaitTstart )
					{
						if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
							printf( "Blocked for Tstart from tid %x\n",
									vs->vtid );

						blocked = stvTrue;
					}

					/* conn has been made, check if we need an XON */

					if ( fg->vState == stvConnected
							&& ( ( STV_TIMESTAMP - fg->toffset )
									% fg->eff_freq ) == 0 )
					{
						if ( STV_OPT_VERBOSE
								& ( STV_DEBUG_ULIB | STV_DEBUG_TIME
										| STV_DEBUG_FREQ ) )
						{
							printf(
								"Ready to send fgid %d to viewer %x\n",
									fg->fgid, vs->vtid );
							printf(
								"t=%d, offset=%d, eff_freq=%d, freq=",
									STV_TIMESTAMP, fg->toffset,
									fg->eff_freq );
							stv_dump_time_str( stdout,
									fg->freq, stvFalse );
							printf( ", (t-off)=%d\n",
									STV_TIMESTAMP - fg->toffset );
						}

						/* Is is time to update our send frequency ? */

						if ( fg->newFreq > 0 && fg->newEffFreq > 0
								&& fg->nfTimeStamp == STV_TIMESTAMP )
						{
							fg->eff_freq = fg->newEffFreq;
							fg->newEffFreq = 0;

							fg->freq = fg->newFreq;
							fg->newFreq = 0;

							/* reset field time offset for new freq */
							fg->toffset = STV_TIMESTAMP;
						}

						/* get any new steering parameters */

						if ( !(send_flag & stvSendNoFieldTimeUpdates) )
							newParams +=
									stv_update_new_params( vs, fg );

						/* do the send thing */

						if ( fg->NeedSend == stvTrue )
						{
							if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
								printf( "Need to Send\n" );

							if ( fg->OKtoSend == stvFalse )
							{
								if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
									printf( "%s fgid=%d, viewer %x\n",
											"Waiting for XON",
											fg->fgid, vs->vtid );

								blocked = stvTrue;
							}

							else
							{
								if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
									printf(
										"Sending to viewer %x id=%d\n",
											vs->vtid, vs->vid );

								stv_pack_SubRegion( vs, fg,
										STV_USERTAG, STV_VIEWER_CTX );

								fg->OKtoSend = stvFalse;

								if ( STV_APP_LOCK != stvAppLocked )
									fg->NeedSend = stvFalse;
							}
						}

						else if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
							printf( "No Need to Send\n" );
					}
				}

				oldfg = fg;
				fg = fg->next;

				/* Handle dead groups and dead Viewers */

				if ( oldfg->vState == stvDeadConnection )
					stv_removeFieldGroup( vs, oldfg );

				if ( vs->nfgs == 0 )
					stv_removeViewer( vs );
			}
		}

		quick = stvFalse;
	}
	while ( blocked || STV_APP_LOCK == stvAppLocked
			|| ( STV_APP_LOCK_HOLD > 0
					&& STV_TIMESTAMP >= STV_APP_LOCK_HOLD ) );

	/* Reset NeedSend flags for viewer field groups */

	vs = &VIEWERS;

	while ( vs->next != (STV_FE_VIEWERPTR) NULL )
	{
		vs = vs->next;

		fg = vs->fg_list->next;

		while ( fg )
		{
			fg->NeedSend = stvTrue;

			fg = fg->next;
		}
	}

	/* Done */

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "Finished all steps...\n\n" );

	if ( !(send_flag & stvSendNoFieldTimeUpdates) )
	{
		stv_bump_field_times( STVFIELDS.next, nSTVFIELDS );

		STV_TIMESTAMP++;
	}

	/* Return # of Changed Params */

	return( newParams );
}


/* ---------------- stv_protoInit0 --------------- */
/* handle init0 message */

int
stv_protoInit0( msg )
STV_BUFPTR msg;
{
	int srcid, bufid;
	int i0seq;
	int tmp;
	int vid;
	int sz;

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "Entry stv_protoInit0()\n" );

	if ( msg == (STV_BUFPTR) NULL )
		return( stvStatusOk );

	/* make sure we're not screwwwwwed in failure recovery. */
	/* if so, drop this Init0 on the floor, and let the */
	/* viewer clean it up */
	if ( STV_CP_RECOVERY_HOLD )
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
			printf(
			"%s: Amidst Failure Recovery, Ignore Viewer Init0 (%x).\n",
				"stv_protoInit0()", msg->src );

		return( stvStatusOk );
	}

	/* forward the init0 packet to the rest of group */
	if ( STV_GCOORD == 0 )
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
			printf( "stv_protoInit0(): forward Init0\n" );

		if ( STV_FULLGROUP != STV_NTASKS )
		{
			stv_group_size( STVSHADOWNAME, &sz );

			if ( sz == STV_NTASKS )
				STV_FULLGROUP = STV_NTASKS;

			/* group not fully formed wait till next time*/
			else
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
					printf( "stv_protoInit0(): group not formed...\n" );

				return( stvStatusOk );
			}
		}

		stv_unpack_int( msg->bufid, &srcid, 1, 1 );
		stv_unpack_int( msg->bufid, &vid, 1, 1 );
		stv_unpack_int( msg->bufid, &i0seq, 1, 1 );

		if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
			printf( "stv_protoInit0(): srcid=%x vid=%d i0seq=%d\n",
					srcid, vid, i0seq );

		if ( STV_NTASKS > 1 )
		{
			stv_init_msg_buf( stvDataDefault, &bufid );

			tmp = stvMsgOp( stvInit0 );
			stv_pack_int( bufid, &tmp, 1, 1 );

			/* pack dummy fgid, for message header consistency */
			tmp = -1;
			stv_pack_int( bufid, &tmp, 1, 1 );

			stv_pack_int( bufid, &srcid, 1, 1 );
			stv_pack_int( bufid, &vid, 1, 1 );
			stv_pack_int( bufid, &i0seq, 1, 1 );

			stv_broadcast( bufid, STVSHADOWNAME, STV_USERTAG,
					STV_APP_CTX );

			stv_free_msg_buf( bufid );
		}
	}

	else
	{
		stv_unpack_int( msg->bufid, &srcid, 1, 1 );
		stv_unpack_int( msg->bufid, &vid, 1, 1 );
		stv_unpack_int( msg->bufid, &i0seq, 1, 1 );

		if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
			printf( "stv_protoInit0(): srcid=%x vid=%d i0seq=%d\n",
					srcid, vid, i0seq );
	}

	/* now pack all the information for **all*** vis fields that have
			been defined */

	stv_init_msg_buf( stvDataDefault, &bufid );

	tmp = stvMsgOp( stvInit1 );
	stv_pack_int( bufid, &tmp, 1, 1 );

	/* dummy fgid, for message header consistency */
	tmp = -1;
	stv_pack_int( bufid, &tmp, 1, 1 );

	stv_pack_int( bufid, &vid, 1, 1 );
	stv_pack_int( bufid, &i0seq, 1, 1 );
	stv_pack_int( bufid, &STV_NRESTARTS, 1, 1 );

	stv_pack_app_info( bufid, STV_NTASKS, STVPARAMS.next, nSTVPARAMS,
		STVFIELDS.next, nSTVFIELDS, stvVisOnly );

	stv_send_msg( bufid, srcid, STV_USERTAG, STV_VIEWER_CTX );

	stv_free_msg_buf( bufid );

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "stv_protoInit0(): reply sent to viewer %x.\n", srcid );

	return( stvStatusOk );
}


/* ---------------- stv_protoFieldRqst --------------- */
/* handle fieldRqst message */

int
stv_protoFieldRqst( msg )
STV_BUFPTR msg;
{
	STV_FIELD_GROUP_PTR fg;

	STV_PARAM_MAP_PTR newMap, pmapTail;

	STV_FIELD_MAP_PTR fmap;

	STV_FE_VIEWERPTR vs;

	STV_FIELDPTR f;

	NLISTPTR nl;

	char fname[256], pname[256];

	int acks;
	int nparams, paramId, i;
	int fgid;

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "Entry stv_protoFieldRqst()\n" );

	if ( msg == (STV_BUFPTR) NULL )
		return( stvStatusOk );

	fgid = msg->fgid;

	/* Find the viewer, else create it */

	if ( (vs = stv_findViewer( &VIEWERS, msg->src )) ==
			(STV_FE_VIEWERPTR) NULL )
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
			printf( "stv_protoFieldRqst(): new viewer %x\n", msg->src );

		vs = stv_create_fe_viewer();
		STV_NO_MEM_return( vs, "stv_protoFieldRqst" );

		vs->refcount = 1;

		vs->fg_list = stv_create_field_group();
		STV_NO_MEM_return( vs->fg_list, "stv_protoFieldRqst" );

		vs->field_map = stv_create_field_map();
		STV_NO_MEM_return( vs->field_map, "stv_protoFieldRqst" );

		vs->param_map = stv_create_param_map();
		STV_NO_MEM_return( vs->param_map, "stv_protoFieldRqst" );

		vs->nfgs = 0;
		vs->vtid = msg->src;
		vs->nNewParams = 0;

		LISTPUTAFTER( &VIEWERS, vs, prev, next );

		nVIEWERS++;
	}

	else if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "stv_protoFieldRqst(): viewer %x found\n", msg->src );

	/* Find the field group, else create it */

	if ( (fg = stv_findFieldGroup( vs->fg_list, fgid )) ==
			(STV_FIELD_GROUP_PTR) NULL )
	{
		fg = stv_create_field_group();
		STV_NO_MEM_return( fg, "stv_protoFieldRqst" );

		fg->fgid = fgid;
		vs->nfgs += 1;

		LISTPUTAFTER( vs->fg_list, fg, prev, next );
	}

	/* Free any previous field group storage */

	if ( fg->field_list )
	{
		for ( i=0; i < fg->nfields; i++ )
		{
			if ( fg->field_list[i] != NULL )
				( fg->field_list[i] )->refcount -= 1;
		}
		FREE( fg->field_list );
	}

	if ( fg->field_types )
		FREE( fg->field_types );
	if ( fg->field_storage_orders )
		FREE( fg->field_storage_orders );
	if ( fg->field_ids )
		FREE( fg->field_ids );

	/* Unpack the visualization information */

	stv_unpack_int( msg->bufid, &(vs->vid), 1, 1 );
	stv_unpack_int( msg->bufid, &(fg->vRegionId), 1, 1 );
	stv_unpack_int( msg->bufid, fg->visRegion,
			STV_REGION_SIZE( STV_MAX_DIM ), 1 );
	stv_unpack_int( msg->bufid, fg->visCell,
			STV_CELL_SIZE( STV_MAX_DIM ), 1 );

	/* Unpack the request info and flags */

	stv_unpack_int( msg->bufid, &(fg->eff_freq), 1, 1 );
	stv_unpack_int( msg->bufid, &(fg->freq), 1, 1 );

	stv_unpack_int( msg->bufid, &(fg->no_synch), 1, 1 );

	stv_unpack_int( msg->bufid, &(fg->is_static), 1, 1 );

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "stv_protoFieldRqst(): no_synch=%d is_static=%d\n",
				fg->no_synch, fg->is_static );

	/* Unpack the list of fields */

	stv_unpack_int( msg->bufid, &(fg->nfields), 1, 1 );

	if ( fg->nfields > 0 )
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
			printf( "stv_protoFieldRqst(): nfields=%d\n", fg->nfields );

		fg->field_list = (STV_FIELDPTR *) CALLOC(
				(STV_SIZE_T) fg->nfields * sizeof(STV_FIELDPTR),
				"stv_protoFieldRqst" );
		STV_NO_MEM_return( fg->field_list, "stv_protoFieldRqst" );

		fg->field_types = (int *) CALLOC( (STV_SIZE_T) fg->nfields
				* sizeof(int), "stv_protoFieldRqst" );
		STV_NO_MEM_return( fg->field_types, "stv_protoFieldRqst" );

		fg->field_storage_orders = (int *) CALLOC(
				(STV_SIZE_T) fg->nfields * sizeof(int),
				"stv_protoFieldRqst" );
		STV_NO_MEM_return( fg->field_storage_orders,
				"stv_protoFieldRqst" );

		fg->field_ids = (int *) CALLOC( (STV_SIZE_T) fg->nfields
				* sizeof(int), "stv_protoFieldRqst" );
		STV_NO_MEM_return( fg->field_ids, "stv_protoFieldRqst" );
	}

	/* Process all the fields that are in this field group */

	acks = 0;

	for ( i=0 ; i < fg->nfields ; i++ )
	{
		stv_unpack_str( msg->bufid, fname );

		stv_unpack_int( msg->bufid, &(fg->field_ids[i]), 1, 1 );
		stv_unpack_int( msg->bufid, &(fg->field_types[i]), 1, 1 );
		stv_unpack_int( msg->bufid, &(fg->field_storage_orders[i]),
			1, 1 );

		f = stv_get_field_name( fname, STVFIELDS.next, nSTVFIELDS );
		fg->field_list[i] = f;

		if ( f != (STV_FIELDPTR) NULL )
		{
			if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
				printf( "stv_protoFieldRqst(): field \"%s\" found.\n",
						fname );

			(f->refcount)++;

			fmap = stv_create_field_map();
			STV_NO_MEM_return( fmap, "stv_protoFieldRqst" );

			fmap->remoteID = fg->field_ids[i];
			fmap->local = f;
			LISTPUTAFTER( vs->field_map, fmap, prev, next );

			/* Verify region overlap for static requests */
			if ( fg->is_static )
			{
				if ( stv_field_region_overlap( f,
						fg->visRegion, fg->visCell ) )
				{
					if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
						printf( "stv_protoFieldRqst(): static ack.\n" );

					acks++;
				}
			}

			else
				acks++;
		}
	}

	if ( fg->no_synch || STV_APP_LOCK == stvAppLocked )
	{
		fg->vState = stvConnected;
		fg->toffset = STV_TIMESTAMP;
		fg->OKtoSend = stvTrue;
		fg->NeedSend = stvTrue;
	}

	else
	{
		fg->vState = stvWaitTstart;
		fg->toffset = STV_TIMESTAMP + 2;
	}

	fg->newEffFreq = 0;
	fg->newFreq = 0;

	/* Unpack the parameter map from the viewer */

	stv_unpack_int( msg->bufid, &nparams, 1, 1 );

	pmapTail = vs->param_map;

	for ( i=0 ; i < nparams ; i++ )
	{
		stv_unpack_str( msg->bufid, pname );

		stv_unpack_int( msg->bufid, &paramId, 1, 1 );

		newMap = stv_create_param_map();
		STV_NO_MEM_return( newMap, "stv_protoFieldRqst" );

		newMap->remoteID = paramId;
		newMap->local = stv_get_param_name( pname,
				STVPARAMS.next, nSTVPARAMS );

		if ( newMap->local )
			(newMap->local->refcount)++;

		LISTPUTAFTER( pmapTail, newMap, prev, next );

		pmapTail = newMap;
	}

	/* See if we need to ask for a notify for the viewer tid */

	nl = VTIDS.next;

	while ( nl )
	{
		if ( nl->vtid == vs->vtid )
			break;
		else
			nl = nl->next;
	}

	/* Add to notify list */

	if ( nl == ( NLISTPTR ) NULL )
	{
		nl = (NLISTPTR) CALLOC( sizeof(NLIST),
			"stv_protoFieldRqst" );
		STV_NO_MEM_return( nl, "stv_protoFieldRqst" );

		nl->vtid = vs->vtid;
		LISTPUTAFTER( &VTIDS, nl, prev, next );

		if ( STV_OPT_VERBOSE & ( STV_DEBUG_ULIB | STV_DEBUG_NOTIFY ) )
		{
			printf( "Asking for a TaskExit Notify for %x\n",
					vs->vtid );
		}

		if ( stv_notify_event( stvTaskExit, STV_USERTAG, STV_APP_CTX,
				1, &(vs->vtid) ) == stvFalse )
		{
			printf( "Notify failed for %x\n", vs->vtid );
		}
	}

	/* If synchronized, respond with an stvFieldTime message */

	if ( !( fg->no_synch ) )
	{
		stv_send_viewer_FieldTime( vs->vtid, fg->fgid, vs->vid,
				STV_TIMESTAMP, ( acks ) ? stvTrue : stvFalse,
				STV_USERTAG );
	}

	/* Else if a static request, send ack / nack */

	else if ( fg->is_static )
	{
		stv_send_viewer_FieldAck( vs->vtid, fg->fgid, vs->vid,
				( acks ) ? stvTrue : stvFalse,
				STV_USERTAG );
	}

	/* If not in static region, or this task doesn't have any fields, */
	/* then mark field group as dead */

	if ( !acks && fg->nfields )
		fg->vState = stvDeadConnection;

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "stv_protoFieldRqst(): reply sent, vState=%d.\n",
				fg->vState );

	return( stvStatusOk );
}


/* ---------------- stv_protoFieldTstart --------------- */
/* handle fieldTstart message */

int
stv_protoFieldTstart( msg )
STV_BUFPTR msg;
{
	STV_FIELD_GROUP_PTR fg;
	STV_FE_VIEWERPTR vs;
	int toff;

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "Entry stv_protoFieldTstart()\n" );

	if ( msg == (STV_BUFPTR) NULL )
		return( stvStatusOk );

	stv_unpack_int( msg->bufid, &toff, 1, 1 );

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "stv_protoFieldTstart(): toff=%d\n", toff );

	if( vs = stv_findViewer( &VIEWERS, msg->src ) )
	{
		if ( ( fg = stv_findFieldGroup( vs->fg_list, msg->fgid ) )
				== (STV_FIELD_GROUP_PTR) NULL )
		{
			sprintf( STVERRTXT,
					"Got fieldTstart for Bad fgid %d - Viewer %x\n",
					msg->fgid, msg->src );
			STV_ERROR( "stv_protoFieldTstart", stvStatusBadParam,
					STVERRTXT );
		}
		else
		{
			if ( fg->vState != stvWaitTstart )
			{
				sprintf( STVERRTXT,
						"Got fieldTstart But Viewer State Was %s (%d)",
						stv_msg_op_str( fg->vState ), fg->vState );
				STV_ERROR( "stv_protoFieldTstart", stvStatusBadParam,
						STVERRTXT );
			}
			else
			{
	 			fg->toffset = toff;
	 			fg->vState = stvConnected;
	 			fg->OKtoSend = stvTrue;
				fg->NeedSend = stvTrue;
			}
		}
	}

	return( stvTrue );
}


/* ---------------- stv_protoVisRegion --------------- */
/* handle fieldTstart message */

int
stv_protoVisRegion( msg )
STV_BUFPTR msg;
{
	STV_FIELD_GROUP_PTR fg;
	STV_FE_VIEWERPTR vs;
	int tmpvRegionId;

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "Entry stv_protoVisRegion()\n" );

	if ( msg == (STV_BUFPTR) NULL )
		return( stvStatusOk );

	if ( (vs = stv_findViewer( &VIEWERS, msg->src ))
			&& (fg = stv_findFieldGroup( vs->fg_list, msg->fgid )) )
	{
		if ( !(fg->is_static) )
		{
			stv_unpack_int( msg->bufid, &tmpvRegionId, 1, 1 );

			if ( tmpvRegionId >= fg->vRegionId )
				fg->vRegionId = tmpvRegionId;

			stv_unpack_int( msg->bufid, fg->visRegion,
				STV_REGION_SIZE( STV_MAX_DIM ), 1 );

			stv_unpack_int( msg->bufid, fg->visCell,
				STV_CELL_SIZE( STV_MAX_DIM ), 1 );

			fg->NeedSend = stvTrue;
		}

		else
		{
			STV_ERROR( "stv_protoVisRegion", stvStatusBadParam,
				"Attempt to Change Static Region" );
		}
	}

	return( stvStatusOk );
}


/* ---------------- stv_protoVisFrequency --------------- */
/* handle fieldTstart message */

int
stv_protoVisFrequency( msg )
STV_BUFPTR msg;
{
	STV_FIELD_GROUP_PTR fg;
	STV_FE_VIEWERPTR vs;
	int tmpvRegionId;

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "Entry stv_protoVisFrequency()\n" );

	if ( msg == (STV_BUFPTR) NULL )
		return( stvStatusOk );

	if( ( vs = stv_findViewer( &VIEWERS, msg->src ) )
			&& ( fg = stv_findFieldGroup( vs->fg_list, msg->fgid ) ) )
	{
		stv_unpack_int( msg->bufid, &tmpvRegionId, 1, 1 );

		if ( tmpvRegionId >= fg->vRegionId )
			fg->vRegionId = tmpvRegionId;

		stv_unpack_int( msg->bufid, &(fg->newEffFreq), 1, 1 );
		stv_unpack_int( msg->bufid, &(fg->newFreq), 1, 1 );

		stv_unpack_int( msg->bufid, &(fg->nfTimeStamp), 1, 1 );

		if ( STV_OPT_VERBOSE
				& ( STV_DEBUG_ULIB | STV_DEBUG_TIME | STV_DEBUG_FREQ ) )
		{
			printf(
			"New freq %d (eff_freq=%d) to be applied at timeStamp %d\n",
				fg->newFreq, fg->newEffFreq, fg->nfTimeStamp );
		}

		fg->NeedSend = stvTrue;
	}

	return( stvStatusOk );
}


/* ---------------- stv_protoNewParams --------------- */
/* handle new parameters message */

int
stv_protoNewParams( msg )
STV_BUFPTR msg;
{
	STV_PARAM_MAP_PTR pmap;
	STV_FE_VIEWERPTR vs;
	STV_VALUE_LIST VL;
	int newTime;
	int remoteID;
	int nParams;
	int i, j;
	int cc;

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "Entry stv_protoNewParams()\n" );

	if ( msg == (STV_BUFPTR) NULL )
		return( stvStatusOk );

	if ( vs = stv_findViewer( &VIEWERS, msg->src ) )
	{
		stv_unpack_int( msg->bufid, &newTime, 1, 1 );

		stv_unpack_int( msg->bufid, &nParams, 1, 1 );

		for ( i=0 ; i < nParams ; i++ )
		{
			stv_unpack_int( msg->bufid, &remoteID, 1, 1 );

			if ( ( pmap = stv_get_param_map( remoteID, vs->param_map ) )
					!= NULL )
			{
				/* Allocate New Value List */

				if ( pmap->newVals == NULL )
					VL = pmap->newVals = stv_create_value_list();

				else
				{
					VL = pmap->newVals;

					while ( VL->next != NULL )
						VL = VL->next;

					VL = VL->next = stv_create_value_list();
				}

				STV_NO_MEM_return( VL, "stv_protoNewParams" );

				/* Allocate New Values */

				VL->values = (STV_VALUE *) CALLOC(
						(STV_SIZE_T) pmap->local->num
							* sizeof(STV_VALUE),
						"stv_protoNewParams" );
				STV_NO_MEM_return( VL->values, "stv_protoNewParams" );

				VL->num = pmap->local->num;

				for ( j=0 ; j < VL->num ; j++ )
				{
					VL->values[j] = (STV_VALUE) CALLOC(
							STV_SIZEOF( pmap->local->types[j]
							& stvTypeMask ), "stv_protoNewParams" );
					STV_NO_MEM_return( VL->values[j],
							"stv_protoNewParams" );
				}

				/* Unpack Values Into List */

				for ( j=0 ; j < VL->num ; j++ )
				{
					if ( (cc = stv_unpack_by_type( msg->bufid,
							VL->values[j], 1, 1,
							pmap->local->types[j] & stvTypeMask ))
						!= stvStatusOk )
					{
						STV_ERROR( "stv_protoNewParams", cc,
							"Typed Unpack Failed." );
						
						return( cc );
					}
				}

				VL->tvalid = newTime;
				VL->update_fgid = msg->fgid;

				(vs->nNewParams)++;
			}

			else
			{
				sprintf( STVERRTXT,
						"paramID (%d) from Viewer %x Not Found",
						remoteID, vs->vtid );
				STV_ERROR( "stv_protoNewParams", stvStatusBadParam,
						STVERRTXT );
			}
		}
	}

	return( stvStatusOk );
}


/* Get Next Set of Indexed Parameter Values ( if any ) */

int
stv_get_remaining_params()
{
	STV_FIELD_GROUP_PTR fg;

	STV_FE_VIEWERPTR vs;

	int moreParams;
	int cc;
	int i;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_get_remaining_params",
			stvSystemInitd | stvSystemActive )) != stvStatusOk )
		return( cc );

	/* Check Each Viewer in Turn */

	vs = &VIEWERS;

	moreParams = 0;

	for ( i=0 ; i < nVIEWERS ; i++ )
	{
		vs = vs->next;

		if ( STV_OPT_VERBOSE & ( STV_DEBUG_ULIB | STV_DEBUG_STEER ) )
		{
			printf( "Checking viewer %x for Remaining Parameters\n",
					vs->vtid );
		}

		fg = vs->fg_list->next;

		/* process each field group that this viewer holds */

		while ( fg )
		{
			if ( fg->vState == stvConnected )
				moreParams += stv_update_new_params( vs, fg );

			fg = fg->next;
		}
	}

	return( moreParams );
}


/* determine if ( any ) new paramters should be updated for the user */

int
stv_update_new_params( vs, fg )
STV_FE_VIEWERPTR vs;
STV_FIELD_GROUP_PTR fg;
{
	STV_PARAM_MAP_PTR pmap;
	STV_VALUE_LIST VLlast;
	STV_VALUE_LIST VL;
	int rval = 0;
	int flag;
	int i;

	if ( vs->param_map )
		pmap = vs->param_map->next;

	else
		return( 0 );       /* hmm don't have any mapped params */

	if ( vs->nNewParams > 0 )
	{
		while( pmap )
		{
			VL = pmap->newVals;

			VLlast = (STV_VALUE_LIST) NULL;

			flag = 0;

			while ( VL != NULL && !flag )
			{
				if ( VL->tvalid != -1
						&& VL->update_fgid == fg->fgid
						&& STV_TIMESTAMP >= VL->tvalid )
				/* D-Oh! */
				{
					if ( STV_OPT_VERBOSE
							& ( STV_DEBUG_ULIB | STV_DEBUG_STEER ) )
					{
						printf( "Updating parameter %s\n",
								pmap->local->name );
					}

					rval++;

					(vs->nNewParams)--;

					for ( i=0 ; i < VL->num ; i++ )
					{
						BCOPY( pmap->local->vars[i], VL->values[i],
								STV_SIZEOF( pmap->local->types[i]
								& stvTypeMask ) );
					}

					pmap->local->updated = stvTrue;

					/* Remove Value Element From List */

					if ( VLlast == NULL )
						pmap->newVals = VL->next;

					else
						VLlast->next = VL->next;

					stv_free_value_list( &VL );

					flag++;
				}

				else
				{
					VLlast = VL;

					VL = VL->next;
				}
			}

			pmap = pmap->next;
		}
	}

	return( rval );
}


/* --------------- stv_protoTimeSync -------------- */
/* handle Time Sync request message */

int
stv_protoTimeSync( msg )
STV_BUFPTR msg;
{
	int lockTime;
	int vid;

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "Entry stv_protoTimeSync()\n" );

	if ( msg == (STV_BUFPTR) NULL ) 
		return( stvStatusOk );

	stv_unpack_int( msg->bufid, &vid, 1, 1 );

	stv_send_viewer_FieldTime( msg->src, msg->fgid, vid,
			STV_TIMESTAMP, stvTrue, STV_USERTAG );

	lockTime = STV_TIMESTAMP + 2;

	if ( STV_APP_LOCK_HOLD < 0 )
		STV_APP_LOCK_HOLD = lockTime;
	
	else
	{
		STV_APP_LOCK_HOLD = ( lockTime < STV_APP_LOCK_HOLD )
			? lockTime : STV_APP_LOCK_HOLD;
	}

	if ( STV_OPT_VERBOSE
			& ( STV_DEBUG_ULIB | STV_DEBUG_TIME | STV_DEBUG_FREQ ) )
	{
		printf( "TimeSync():  app lock in progress, hold at %d\n",
				STV_APP_LOCK_HOLD );
	}

	return( stvStatusOk );
}


/* ---------------- stv_protoAppLock --------------- */
/* handle Application Lock message */

int
stv_protoAppLock( msg )
STV_BUFPTR msg;
{
	int lockTime;
	int lockTid;

	int no_synch;

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "Entry stv_protoAppLock()\n" );

	if ( msg == (STV_BUFPTR) NULL ) 
		return( stvStatusOk );

	stv_unpack_int( msg->bufid, &lockTid, 1, 1 );    /* ignore */

	stv_unpack_int( msg->bufid, &no_synch, 1, 1 );

	if ( !no_synch )
		stv_unpack_int( msg->bufid, &lockTime, 1, 1 );

	STV_APP_LOCK_HOLD = -1;

	if ( STV_APP_LOCK == stvAppUnlocked )
	{
		if ( !no_synch )
		{
			if ( STV_APP_LOCK_TIME < 0 )
				STV_APP_LOCK_TIME = lockTime;

			else
			{
				STV_APP_LOCK_TIME = ( lockTime < STV_APP_LOCK_TIME )
					? lockTime : STV_APP_LOCK_TIME;
			}
		}

		else
			STV_APP_LOCK_TIME = STV_TIMESTAMP;
	}

	if ( STV_OPT_VERBOSE
			& ( STV_DEBUG_ULIB | STV_DEBUG_TIME | STV_DEBUG_FREQ ) )
	{
		printf( "AppLock():  app lock confirmed for %d\n",
			STV_APP_LOCK_TIME );
	}

	return( stvStatusOk );
}


/* ---------------- stv_protoAppUnlock --------------- */
/* handle Application Release message */

int
stv_protoAppUnlock( msg )
STV_BUFPTR msg;
{
	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "Entry stv_protoAppUnlock()\n" );

	if ( msg == (STV_BUFPTR) NULL ) 
		return( stvStatusOk );

	STV_APP_LOCK = stvAppUnlocked;

	STV_APP_LOCK_TIME = -1;
	STV_APP_LOCK_HOLD = -1;

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "AppUnlock():  app unlocked\n" );

	return( stvStatusOk );
}


/* ---------------- stv_protoSetopt --------------- */
/* handle Setopt message */

int
stv_protoSetopt( msg )
STV_BUFPTR msg;
{
	int what;
	int option;
	int value;

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "Entry stv_protoSetopt()\n" );

	if ( msg == (STV_BUFPTR) NULL ) 
		return( stvStatusOk );

	stv_unpack_int( msg->bufid, &what, 1, 1 );
	stv_unpack_int( msg->bufid, &option, 1, 1 );
	stv_unpack_int( msg->bufid, &value, 1, 1 );
	
	stv_setopt( what, option, value );

	return( stvStatusOk );
}


/* ---------------- stv_protoXon --------------- */
/* handle Xon message */

int
stv_protoXon( msg )
STV_BUFPTR msg;
{
	STV_FIELD_GROUP_PTR fg;
	STV_FE_VIEWERPTR vs;

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "Entry stv_protoXon()\n" );

	if ( msg == (STV_BUFPTR) NULL ) 
		return( stvStatusOk );

	if ( (vs = stv_findViewer( &VIEWERS, msg->src ))
			&& (fg = stv_findFieldGroup( vs->fg_list, msg->fgid )) )
	{
		fg->OKtoSend = stvTrue;

		if ( STV_APP_LOCK == stvAppLocked )
			fg->NeedSend = stvTrue;
	}

	return( stvStatusOk );
}


/* ---------------- stv_protoNotify --------------- */
/* handle Notify message */

int
stv_protoNotify( msg )
STV_BUFPTR msg;
{
	STV_HANDLER H;

	int saveop;
	int ret;
	int cc;

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_ULIB | STV_DEBUG_NOTIFY ) )
		printf( "Entry stv_protoNotify()\n" );

	if ( msg == (STV_BUFPTR) NULL ) 
		return( stvStatusOk );

	printf( "Something died...  Damn.\n" );

	H = STVNOTIFYHANDLERS;

	ret = stvStatusOk;

	saveop = msg->op;

	while ( H != NULL )
	{
		msg->op = H->op;

		cc = (H->handler)( msg );

		if ( cc != stvStatusOk )
			ret = cc;

		H = H->next;
	}

	msg->op = saveop;

	return( ret );
}


/* ---------------- stv_protoUlibNotify --------------- */
/* handle Notify message for the User Library */

int
stv_protoUlibNotify( msg )
STV_BUFPTR msg;
{
	int cc;

	if ( STV_OPT_VERBOSE & ( STV_DEBUG_ULIB | STV_DEBUG_NOTIFY ) )
		printf( "Entry stv_protoUlibNotify()\n" );

	if ( msg == (STV_BUFPTR) NULL ) 
		return( stvStatusOk );

	cc = stv_vector_handler( STVHANDLERS, nSTVHANDLERS, msg );

	return( cc );
}


/* ---------------- stv_protoFieldHaltGroup --------------- */
/* handle FieldHaltGroup message
*  - Mark the a field group as dead.
*  - Don't remove the field group
*  - clear previous messages from the tid and this fgid
*/

int
stv_protoFieldHaltGroup( msg )
STV_BUFPTR msg;
{
	STV_FIELD_GROUP_PTR fg;
	STV_FE_VIEWERPTR vs;
	int vtid;

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "Entry stv_protoFieldHaltGroup()\n" );

	if ( msg == (STV_BUFPTR) NULL ) 
		return( stvStatusOk );

	if ( msg->op != stvFieldHaltGroup ) 
		return( stvStatusOk );

	vtid = msg->src;

	if ( ( vs = stv_findViewer( &VIEWERS, vtid ) )
			!= (STV_FE_VIEWERPTR) NULL )
	{
		if ( ( fg = stv_findFieldGroup( vs->fg_list, msg->fgid ) )
				!= (STV_FIELD_GROUP_PTR) NULL )
		{
			fg->vState = stvDeadConnection;
		}
	}

	return( stvStatusOk );
}


/* ---------------- stv_protoFieldHaltAll --------------- */
/* handle FieldHaltAll message
*  - Mark all field groups as dead.
*  - Don't remove the field groups
*  - clear previous messages from the tid and any fgid
*/

int
stv_protoFieldHaltAll( msg )
STV_BUFPTR msg;
{
	STV_FIELD_GROUP_PTR fg;
	STV_FE_VIEWERPTR vs;
	int vtid;

	if ( STV_OPT_VERBOSE & STV_DEBUG_ULIB )
		printf( "Entry stv_protoFieldHaltAll()\n" );

	if ( msg == (STV_BUFPTR) NULL )
		return( stvStatusOk );

	if ( msg->op != stvFieldHaltAll )
		return( stvStatusOk );

	vtid = msg->src;
	if ( ( vs = stv_findViewer( &VIEWERS, vtid ) )
			!= (STV_FE_VIEWERPTR) NULL )
	{
		fg = vs->fg_list->next;
		while( fg )
		{
			fg->vState = stvDeadConnection;
			fg = fg->next;
		}
	}

	return( stvStatusOk );
}


/* ----------- stv_removeFieldGroup ---------- */

void
stv_removeFieldGroup( vs, fg )
STV_FE_VIEWERPTR vs;
STV_FIELD_GROUP_PTR fg;
{
	int i;
	if ( fg )
	{
		if ( fg->field_list )
		{
			for ( i = 0; i < fg->nfields; i++ )
			{
				if ( fg->field_list[i] != NULL )
					( fg->field_list[i] )->refcount -= 1;
			}
			FREE( fg->field_list );
		}

		if ( fg->field_types )
			FREE( fg->field_types );

		if ( fg->field_storage_orders )
			FREE( fg->field_storage_orders );

		if ( fg->field_ids )
			FREE( fg->field_ids );

		LISTDELETE( fg, prev, next );

		if ( vs )
			vs->nfgs -= 1;
	}
}


/* ----------- stv_removeViewer ---------- */

void
stv_removeViewer( vs )
STV_FE_VIEWERPTR vs;
{
	if( vs && vs->nfgs == 0 )
	{
		if ( vs )
		{
			LISTDELETE( vs, prev, next );
			nVIEWERS --;
		}
	}
}


/* ----------- stv_isParamChanged ---------- */
int
stv_isParamChanged(paramId)
int paramId;
{
	STV_PARAMPTR p;

	int rval;
	int cc;

	/* Check System Status */

	if ( (cc = stv_sys_ok( "stv_isParamChanged",
			stvSystemInitd | stvSystemActive )) != stvStatusOk )
		return( cc );

	/* Find & Check Param */

	if ( (p = stv_findParam( &STVPARAMS, paramId ))
			== (STV_PARAMPTR) NULL)
	{
		return( stvStatusBadParam );
	}

	rval = p->updated;

	p->updated = 0;

	return( rval );
}

	
/* ----------- stv_handleMessages ---------- */

/* This handles the different kind of messages that the CUMULVS 
library supports */

int
stv_handleMessages( op, src, tag, ctxt, block )
int op;
int src;
int tag;
STV_CONTEXT ctxt;
int block;
{
	STV_BUFPTR msg;

	int match = stvFalse;
	int cc;

	/* Loop Until Matching Message Returned or Failure */

	do
	{
		/* Receive a Message */

		msg = stv_recvMessage( stvAny, op, src, tag, ctxt, block );

		/* If NULL Message Returned, Something Went Awry or Not Found */

		if ( msg == (STV_BUFPTR) NULL )
			return( stvFalse );

		/* Is Message a Match with Caller Arguments? */

		if ( ( op == -1 || op == msg->op )
				&& ( src == -1 || src == msg->src )
				&& ( tag == -1 || tag == msg->tag )
				&& STV_CONTEXT_MATCH( ctxt, msg->ctxt ) )
		{
			match = stvTrue;
		}

		/* Try to Handle Message */

		cc = stv_vector_handler( STVHANDLERS, nSTVHANDLERS, msg );

		if ( cc == stvStatusOk )
			cc = stvTrue;

		/* Kick Out on Notify */

		if ( msg->op == stvNotify )
		{
			printf( "%s: stvNotify, Kicking Out...\n",
				"stv_handleMessages()" );

			cc = stvFalse;
		}

		/* check for blocking mode failures & bailouts */

		if ( ( block == stvMsgBlock || block == stvMsgBlockTimeout )
				&& msg->src == src && src >= 0 )
		{
			/* waiting for message from task that died */
			if ( msg->op == stvNotify )
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_MSG )
					printf( "%s: blocked, sender died.\n",
							"stv_handleMessages()" );

				cc = stvStatusBail;
			}

			/* blocking receive, single sender failed, halted */
			if ( msg->op == stvFieldHaltAll )
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_MSG )
					printf( "%s: blocked, FieldHaltAll.\n",
							"stv_handleMessages()" );

				cc = stvStatusBail;
			}

			if ( msg->op == stvFieldHaltGroup
					/* XXX BETTER FIX THIS!!! && msg->fgid == fgid */
					)
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_MSG )
					printf( "%s: blocked, %s for fgid=%d.\n",
						"stv_handleMessages()", "FieldHaltGroup",
						msg->fgid );

				cc = stvStatusBail;
			}

			if ( msg->op == stvCpdHalt )
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_MSG )
					printf( "%s: blocked, CpdHalt.\n",
							"stv_handleMessages()" );

				cc = stvStatusBail;
			}

			/* waiting for a cpdXon but got Restart (msg) first */
			if ( msg->op == stvRestart && op == stvCpdXon )
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_MSG )
					printf( "%s: blocked, Restart While Wait CpdXon.\n",
							"stv_handleMessages()" );

				cc = stvStatusBail;
			}

			/* blocking receive, single sender failed, halted */
			if ( msg->op == stvSenderFailed
					|| msg->op == stvCpdPeerFailed )
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_MSG )
					printf( "%s: blocked, %s / %s.\n",
							"stv_handleMessages()",
							"SenderFailed", "CpdPeerFail" );

				cc = stvStatusBail;
			}
		}

		/* free the message buffer and the message */

		stv_free_msg_buf( msg->bufid );

		FREE( msg );
	}
	while ( !match && cc > 0 );

	/* cc == stvTrue means we handled a message */

	return( cc );
}


/* ----------- stv_version ---------- */
char *
stv_version()
{
	char *rcopy;

	/* Little memory leak so that SGI's and others don't yack if
		a calling program tries to scanf the string returned here */

	rcopy = stv_copy_str(STV_VERSION);

	return rcopy;
}


/* ----------- stv_version_name ---------- */
char *
stv_version_name()
{
	char *rcopy;

	/* Little memory leak so that SGI's and others don't yack if
		a calling program tries to scanf the string returned here */

	rcopy = stv_copy_str(STV_VERSION_NAME);

	return rcopy;
}


/* ------------ stv_sys_ok ------------- */
int
stv_sys_ok( caller, flag )
char *caller;
int flag;
{
	static int first = 1;

	int cc;

	if ( ( flag & stvSystemActive ) && STV_DEACTIVATED == stvTrue )
	{
		if ( first )
		{
			sprintf( STVERRTXT,
				"%s - CUMULVS Deactivated Due to Initialization Error.",
				"Function Ignored" );
			STV_ERROR( caller, stvStatusSysFail, STVERRTXT );

			first = 0;
		}

		if ( (cc = stv_init_system( stvTerse )) != stvStatusOk )
			return( cc );

		STV_DEACTIVATED = stvFalse;

		first = 1;
	}

	if ( ( flag & stvSystemInitd ) && STV_MYID == -1 )
	{
		sprintf( STVERRTXT,
			"CUMULVS Not Initialized - Call %s or %s First!",
			"stv_init()", "stvfinit()" );
		STV_ERROR( caller, stvStatusBadParam, STVERRTXT );

		return( stvStatusBadParam );
	}

	return( stvStatusOk );
}


/* C++ Compatibility */

#ifdef __cplusplus
}
#endif

